/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package hybridtree.core;

/**
 * @author Wei LU
 * @version 1.0
 */
import java.util.ArrayList;

import hybridtree.core.perceptron.*;

import java.io.Serializable;

public class Candidate implements Serializable {

	private static final long serialVersionUID = 8098066597240384980L;

	private CandidateSet _canSet; //the set of all Candidates

    private int _rank; // the rank of this candidate in the generative model's output

    private ProdHybridNode _rootNode; // the root node of the hybrid tree

    private ArrayList<FVPair> _featureVector; // the feature vector

    private boolean _isCorrect; // the flag shows whether this Candidate is correct
    
    public static Candidate createCandidate(CandidateSet canSet, ProdHybridNode rootNode, int rank) {
        return new Candidate(canSet, rootNode, rank);
    }

    /**
     * Constructor
     * @param canSet CandidateSet
     * @param rootNode HybridNode
     * @param rank int
     */
    private Candidate(CandidateSet canSet, ProdHybridNode rootNode, int rank) {
        this._canSet = canSet;
        this._rootNode = rootNode;
        this._rank = rank;
        this._isCorrect = false;
        canSet.addCandidate(this);
    }

    public void setCorrect() {
        this._isCorrect = true;
    }

    /**
     * Get the root HybridNode
     * @return HybridNode
     */
    public ProdHybridNode getRootNode() {
        return this._rootNode;
    }

    /**
     * Get the rank of this Candidate
     * @return int
     */
    public int getRank() {
        return this._rank;
    }

    /**
     * Get the probability of this Candidate
     * @return double
     */
    public double getProb() {
        return this.getRootNode().getElem().getProb();
    }

    /**
     * Assume the rank 0 candidate is the correct one, only.
     * @return boolean
     */
    public boolean isCorrect() {
        return this._isCorrect;
    }

    /**
     * compute the cost associated with the whole feature vector
     * @return double
     */
    public double computeFeatureCost() {
        double cost = 0;
        for (int i = 0; i < this._featureVector.size(); i++) {
            cost += this._featureVector.get(i).computeCost();
        }
        return cost;
    }
    
    /**
     * Get the set of all Candidates
     * @return CandidateSet
     */
    public CandidateSet getCandidateSet() {
        return this._canSet;
    }

    /**
     * Get all the features associated with this Candidate, given a Perceptron
     * @param p Perceptron
     * @return Vector
     */
    public ArrayList<FVPair> getFeatures(Perceptron p) {
        if (this._featureVector != null) {
            return this._featureVector;
        }
        this._featureVector = p.extractFeatures(this);
        return this._featureVector;
    }

}
