/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.core;

/**
 * @author LU, Wei
 * @version 1.0
 */

import java.io.Serializable;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import hybridtree.data.GProduction;
import hybridtree.data.InnerFeature;
import hybridtree.data.Model;
import hybridtree.data.Node;
import hybridtree.data.NonterminalNode;
import hybridtree.data.Production;
import hybridtree.data.Rule;
import hybridtree.data.TrainExample;
import hybridtree.data.Word;
import hybridtree.util.KeyGen;


public class Param implements Serializable {
	
	private static final long serialVersionUID = 7074912368311962168L;

	private Model _model;

    /**
     * all the GProductions
     */
    private ArrayList<GProduction> _allGProds;

    /**
     * all the InnerFeatures
     */
    private ArrayList<InnerFeature> _allInnerFeatures;

    /**
     * all the Words
     */
    private ArrayList<Word> _allWords;

    /**
     * The pattern's IDs
     */
    private int[] _patternIDs;

    /**
     * The probability of generating a Word from a Production
     */
    private HashMap<Long, Double> _prob_w1; //(prod-id, word-id)

    private HashMap<Long, Double> _prob_w2; //(prod-id, word1-id, word-id)

    private HashMap<Long, Double> _prob_w3; //(prod-id, word2-id, word1-id, word-id)

    /**
     * The probability of generating a Word from a Production (for new iteration)
     */
    private HashMap<Long, Double> _prob_w1_new;

    private HashMap<Long, Double> _prob_w2_new;

    private HashMap<Long, Double> _prob_w3_new;

    private HashMap<Long, ArrayList<Integer>> _word1_to_word; // (word1-id, word-id)

    private HashMap<Long, ArrayList<Integer>> _word2_word1_to_word; // (word2-id, word1-id, word-id)

    /**
     * The probability of patterns
     */
    private HashMap<Long, Double> _patternTable;

    /**
     * The probability of patterns (new)
     */
    private HashMap<Long, Double> _patternTable_new;

    private boolean _MRL_LM_flag = false; //true if LM already estimated.

    private HashMap<Integer, ArrayList<Integer>[]>_MRLMChildTable;
    private HashMap<Long, Double> _MRLMProbTable;

    private HashMap<Long, Double> hybridRulesTable_unigram; //(GProduction, Word) - count
    private HashMap<Long, Double> hybridRulesTable_bigram; //(GProduction, Word, Word) - count
    private HashMap<Long, Double> hybridRulesTable_trigram; //(GProduction, Word, Word, Word) - count

    private double[] _MRLMProbTable_unigram;

    private HashMap<Long, Double> _gp_word_probs;

    /**
     * Constructor
     * @param examples the Examples
     */
    public Param(Model model) {

        this._model = model;

        this._allWords = this._model.getAllWords();
        this._allGProds = this._model.getAllGProds();
        this._allInnerFeatures = this._model.getAllInnerFeatures();

        this._prob_w1 = new HashMap<Long, Double>();
        this._prob_w1_new = new HashMap<Long, Double>();

        this._prob_w2 = new HashMap<Long, Double>();
        this._prob_w2_new = new HashMap<Long, Double>();

        this._prob_w3 = new HashMap<Long, Double>();
        this._prob_w3_new = new HashMap<Long, Double>();

        this._word1_to_word = new HashMap<Long, ArrayList<Integer>>();
        this._word2_word1_to_word = new HashMap<Long, ArrayList<Integer>>();

        //the pattern rules table
        this._patternTable = new HashMap<Long, Double>();
        this._patternTable_new = new HashMap<Long, Double>();
        this._patternIDs = Rule.getRulePatterns();

        //the hybrid rules table
        this.hybridRulesTable_unigram = new HashMap<Long, Double>();
        this.hybridRulesTable_bigram = new HashMap<Long, Double>();
        this.hybridRulesTable_trigram = new HashMap<Long, Double>();

        this._MRL_LM_flag = false;

        this._gp_word_probs = new HashMap<Long, Double>();
        this._gp_word_probs_w2 = new HashMap<Long, Double>();

    }

    /**
     * Get the model associated with this Param
     * @return Model
     */
    public Model getModel() {
        return this._model;
    }

    public double getPatternProb(GProduction gp, int patternID) {
        long key = KeyGen.toKey(gp.getProdId(), patternID);
        if (this._patternTable.containsKey(key)) {
            double prob = ((Double)this._patternTable.get(key)).doubleValue();
            return prob;
        } else {
            return 0;
        }
    }

    /**
     * Get all Words
     * @return a collection of Words
     */
    public ArrayList<Word> getAllWords() {
        return this._allWords;
    }

    /**
     * Get all GProductions
     * @return a collection of GProductions
     */
    public ArrayList<GProduction> getAllGProds() {
        return this._allGProds;
    }

    public ArrayList<InnerFeature> getAllInnerFeatures() {
        return this._allInnerFeatures;
    }

    /**
     * Normalize the pattern probabilities and swap
     */
    private void normalizeAndSwapPatternProb() {

        /**
         * If the new pattern probability is not filled, skip normalization
         */
        if (this._patternTable_new.isEmpty()) {
            return;
        }

        for (int pIndex = 0; pIndex < this._allGProds.size(); pIndex++) {

            double sum = 0;

            for (int patternID = 0; patternID < this._patternIDs.length;
                                 patternID++) {
                long key = KeyGen.toKey(pIndex, patternID);
                if (this._patternTable_new.containsKey(key)) {
                    double prob = ((Double)this._patternTable_new.get(key)).
                                  doubleValue();
                    sum += prob;
                }
            }

            if (sum != 0) {
                for (int patternID = 0; patternID < this._patternIDs.length;
                                     patternID++) {
                    long key = KeyGen.toKey(pIndex, patternID);
                    if (this._patternTable_new.containsKey(key)) {
                        double prob = ((Double)this._patternTable_new.get(key)).
                                      doubleValue();
                        this._patternTable_new.put(key, new Double(prob / sum));
                    }
                }
            }
        }

        this._patternTable = this._patternTable_new;
        this._patternTable_new = new HashMap<Long, Double>();

    }

    /**
     * Put the pattern counts
     * @param pIndex the Production index
     * @param patternID the pattern Id
     * @param count the patten counts
     */
    public void putPatternCount(GProduction gp, int patternID, double count) {

        if (Double.isNaN(count)) {
            throw new ParamEstimationException("Not a Number!(242)");
        }

        long key = KeyGen.toKey(gp.getProdId(), patternID);
        if (!this._patternTable_new.containsKey(key)) {
            this._patternTable_new.put(key, new Double(count));
        } else {
            double oldValue = ((Double)this._patternTable_new.get(key)).
                              doubleValue();
            this._patternTable_new.put(key, new Double(oldValue + count));
        }

    }


    /**
     * Initialization of the MRL LM probability table
     */
    public void initMRLM() {

        /**
         * Do not do it again if it is already done.
         */
        if (this._MRL_LM_flag) {
            return;
        }

        this._MRLMChildTable = new HashMap<Integer, ArrayList<Integer>[] > ();
        this._MRLMProbTable = new HashMap<Long, Double>();
        this._MRLMProbTable_unigram = new double[this._allGProds.size()];

        for (int i = 0; i < this._allGProds.size(); i++) {
            GProduction gp = (GProduction)this._allGProds.get(i);
            ArrayList<NonterminalNode> rhs_nodes = gp.getRHSNonterminals();

            @SuppressWarnings("unchecked")
			ArrayList<Integer>
                    children[] = (ArrayList<Integer>[]) Array.newInstance(ArrayList.class,
                    rhs_nodes.size());
            for (int j = 0; j < rhs_nodes.size(); j++) {
                children[j] = new ArrayList<Integer>();
            }
            for (int childIndex = 0; childIndex < rhs_nodes.size(); childIndex++) {
                Node rhs_node = rhs_nodes.get(childIndex);
                for (int k = 0; k < this._allGProds.size(); k++) {
                    GProduction gpr = (GProduction)this._allGProds.get(k);
                    if (gpr.getLHS().getName().equals(rhs_node.getName()) &&
                        !gpr.isConstant()) {
                        children[childIndex].add(gpr.getProdId());
                        long key = KeyGen.toKey(gp.getProdId(), childIndex,
                                                gpr.getProdId());
                        this._MRLMProbTable.put(key, new Double(0.00));
                    }
                }
            }
            this._MRLMChildTable.put(gp.getProdId(), children);
        }

    }

    /**
     * Normalization of MRL count
     */
    public void normalizeMRLCount() {

        double unigramSum = 0;
        for (int i = 0; i < this._MRLMProbTable_unigram.length; i++) {
            unigramSum += this._MRLMProbTable_unigram[i];
        }
        for (int i = 0; i < this._MRLMProbTable_unigram.length; i++) {
            this._MRLMProbTable_unigram[i] /= unigramSum;
        }

        for (int i = 0; i < this._allGProds.size(); i++) {
            GProduction gp1 = this._allGProds.get(i);
            ArrayList<NonterminalNode> children = gp1.getRHSNonterminals();
            for (int childIndex = 0; childIndex < children.size(); childIndex++) {
                double alpha = 0;
                double sum = 0;
                for (int j = 0; j < this._allGProds.size(); j++) {
                    GProduction gp2 = this._allGProds.get(j);
                    long key = KeyGen.toKey(gp1.getProdId(), childIndex,
                                            gp2.getProdId());
                    if (this._MRLMProbTable.containsKey(key)) {
                        double bigramCount = this._MRLMProbTable.get(key);
                        if (bigramCount > 0) {
                            double unigramCount = unigramSum *
                                                  this._MRLMProbTable_unigram[
                                                  gp1.getProdId()];
                            double prob = (bigramCount - 0.5) / unigramCount;
                            alpha += prob;
                            this._MRLMProbTable.put(key, prob);
                            if (Double.isNaN(prob)) {
                                System.err.println("(1)");
                                System.err.println(gp1.toLine());
                                System.err.println(childIndex);
                                System.err.println(gp2.toLine());
                                System.err.println(unigramCount);
                                System.err.println(bigramCount);
                                throw new RuntimeException("Error!");
                            }

                        } else {
                            sum += this._MRLMProbTable_unigram[gp2.getProdId()];
                        }
                    }
                }
                alpha = 1 - alpha;
                for (int j = 0; j < this._allGProds.size(); j++) {
                    GProduction gp2 = this._allGProds.get(j);
                    long key = KeyGen.toKey(gp1.getProdId(), childIndex,
                                            gp2.getProdId());
                    if (this._MRLMProbTable.containsKey(key)) {
                        double bigramCount = this._MRLMProbTable.get(key);
                        if (bigramCount == 0) {
                            if (sum == 0) {
                                this._MRLMProbTable.put(key, 0.0);
                            } else {
                                double prob = alpha *
                                              this._MRLMProbTable_unigram[gp2.
                                              getProdId()] / sum;
                                if (alpha == 0) {
                                    this._MRLMProbTable.put(key, 0.0);
                                } else {
                                    this._MRLMProbTable.put(key, prob);
                                }

                                if (Double.isNaN(prob)) {
                                    System.err.println("(2)");
                                    System.err.println(gp1.toLine());
                                    System.err.println(childIndex);
                                    System.err.println(gp2.toLine());
                                    System.err.println(alpha);
                                    System.err.println(this.
                                            _MRLMProbTable_unigram[gp2.
                                            getProdId()]);
                                    System.err.println(sum);
                                    System.exit(1);
                                }
                            }
                        }
                    }
                }
            }
        }

        /**
         * Disable future normalization
         */
        this._MRL_LM_flag = true;

    }

    public double computeMRLLM(Production prod) {
        ArrayList<Production> children = prod.getChildren();
        double prob = 1.0;
        for (int cIndex = 0; cIndex < children.size(); cIndex++) {
            Production child = children.get(cIndex);
            GProduction gp2 = child.getGProduction();
            if (gp2.isConstant()) {
            } else {
                prob *=
                        this.getMRLMProb(prod.getGProduction(), cIndex,
                                         child.getGProduction());
                prob *= computeMRLLM(child);
            }
        }
        return prob;
    }

    /**
     * Get the MRL LM probability
     * @param parent the parent GProduction
     * @param childIndex the child position
     * @param child the child GProduction
     * @return the LM probability value
     */
    public double getMRLMProb(GProduction parent, int childIndex,
                              GProduction child) {

        long key = KeyGen.toKey(parent.getProdId(), childIndex, child.getProdId());
        return ((Double)this._MRLMProbTable.get(key)).doubleValue();

    }

    public void putMRLMCount(GProduction prod, double count) {
        this._MRLMProbTable_unigram[prod.getProdId()] += count;
    }

    /**
     * Put the MRL LM count
     * @param parent the parent GProduction
     * @param childIndex the child position
     * @param child the child GProduction
     * @param count the count value
     */
    public void putMRLMCount(GProduction parent, int childIndex,
                             GProduction child, double count) {

        if (Double.isNaN(count)) {
            throw new ParamEstimationException("Not a Number!(355)");
        }

        /** Check whether MRL LM count requires further recording. **/
        if (this._MRL_LM_flag) {
            return;
        }

        long key2 = KeyGen.toKey(parent.getProdId(), childIndex,
                                 child.getProdId());

        if (this._MRLMProbTable.containsKey(key2)) {
            double oldCount = this._MRLMProbTable.get(key2);
            this._MRLMProbTable.put(key2, oldCount + count);
        } else {
            this._MRLMProbTable.put(key2, count);
        }

    }

    /**
     * Get the probability value given the key
     * @param h HashMap
     * @param key long
     * @return double
     */
    private double getValue(HashMap<Long, Double> h, long key) {

        if (!h.containsKey(key)) {
            return 0.0;
        }
        return ((Double) h.get(key)).doubleValue();

    }

    /**
     * Get the probability, given a GProduction and a Word
     * @param prod GProduction
     * @param word Word
     * @return double
     */
    public double getProb(GProduction prod, Word word) {

        long key = KeyGen.toKey(prod.getProdId(), word.getWordId());
        if (this._gp_word_probs.containsKey(key)) {
            return (Double)this._gp_word_probs.get(key);
        }

        ArrayList<InnerFeature> innerFeatures = prod.getInnerFeatures();

        double prob = 0;
        for (int i = 0; i < innerFeatures.size(); i++) {
            InnerFeature innerFeature = innerFeatures.get(i);
            prob += this.getInnerFeatureProb(innerFeature, word);
        }

        //uniform assumption
        double ratio = 1.0 / (double) innerFeatures.size();
        prob *= ratio;

        this._gp_word_probs.put(key, prob);

        return prob;

    }

    private HashMap<Long, Double> _gp_word_probs_w2;

    public double getProb_mixGram(GProduction prod, Word word1, Word word) {

        long key = KeyGen.toKey(prod.getProdId(), word1.getWordId(),
                                word.getWordId());
        if (this._gp_word_probs_w2.containsKey(key)) {
            return (Double)this._gp_word_probs_w2.get(key);
        }

        ArrayList<InnerFeature> innerFeatures = prod.getInnerFeatures();

        double prob = 0;
        if (this._prob_w2.size() == 0) {
            for (int i = 0; i < innerFeatures.size(); i++) {
                InnerFeature innerFeature = innerFeatures.get(i);
                prob += this.getInnerFeatureProb(innerFeature, word);
            }
        } else {
            for (int i = 0; i < innerFeatures.size(); i++) {
                InnerFeature innerFeature = innerFeatures.get(i);
                prob +=
                        this.getInnerFeatureProb_Bigram(innerFeature, word1,
                        word);
            }
        }

        //uniform assumption
        double ratio = 1.0 / (double) innerFeatures.size();
        prob *= ratio;

        this._gp_word_probs_w2.put(key, prob);

        return prob;

    }

    public double getProb(GProduction prod, Word word1, Word word) {

        long key = KeyGen.toKey(prod.getProdId(), word1.getWordId(),
                                word.getWordId());
        if (this._gp_word_probs_w2.containsKey(key)) {
            return (Double)this._gp_word_probs_w2.get(key);
        }

        ArrayList<InnerFeature> innerFeatures = prod.getInnerFeatures();

        double prob = 0;
        if (this._prob_w2.size() == 0) {
            double value = this.getProb(prod, word);
            return value;
        } else {
            for (int i = 0; i < innerFeatures.size(); i++) {
                InnerFeature innerFeature = innerFeatures.get(i);
                prob += this.getInnerFeatureProb(innerFeature, word1, word);
            }
        }

        //uniform assumption
        double ratio = 1.0 / (double) innerFeatures.size();
        prob *= ratio;

        /*
        //System.err.println('w');

        if (word.getWordStr().equals("0") || word.getWordStr().equals("0.0")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
                if ((w.getWordStr().equals("0.0") || w.getWordStr().equals("0")) &&
                    !w.equals(word)) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
        } else
        if (word.getWordStr().equals("0.4") || word.getWordStr().equals("0.40")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
//                System.err.println(w.getWordStr());
                if ((w.getWordStr().equals("0.4") ||
                     w.getWordStr().equals("0.40")) &&
                    !w.getWordStr().equals(word.getWordStr())) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
            //System.err.print('p');
        } else
        if (word.getWordStr().equals("-10") ||
            word.getWordStr().equals("-10.00")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
                if ((w.getWordStr().equals("-10") ||
                     w.getWordStr().equals("-10.00")) &&
                    !w.equals(word)) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
        } else
        if (word.getWordStr().equals("-13") || word.getWordStr().equals("-13.0")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
                if ((w.getWordStr().equals("-13") ||
                     w.getWordStr().equals("-13.0")) &&
                    !w.equals(word)) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
        } else
        if (word.getWordStr().equals("1.5") || word.getWordStr().equals("1.50")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
                if ((w.getWordStr().equals("1.5") ||
                     w.getWordStr().equals("1.50")) &&
                    !w.equals(word)) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
        } else
        if (word.getWordStr().equals("16") || word.getWordStr().equals("16.0")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
                if ((w.getWordStr().equals("16") ||
                     w.getWordStr().equals("16.0")) &&
                    !w.equals(word)) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
        } else
        if (word.getWordStr().equals("-17") ||
            word.getWordStr().equals("-17.0")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
                if ((w.getWordStr().equals("-17") ||
                     w.getWordStr().equals("-17.0")) &&
                    !w.equals(word)) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
        } else
        if (word.getWordStr().equals("18") || word.getWordStr().equals("18.0")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
                if ((w.getWordStr().equals("18") ||
                     w.getWordStr().equals("18.0")) &&
                    !w.equals(word)) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
        } else
        if (word.getWordStr().equals("2.00") || word.getWordStr().equals("2")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
                if ((w.getWordStr().equals("2.00") || w.getWordStr().equals("2")) &&
                    !w.equals(word)) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
        } else
        if (word.getWordStr().equals("-7") || word.getWordStr().equals("-7.0")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
                if ((w.getWordStr().equals("-7") ||
                     w.getWordStr().equals("-7.0")) &&
                    !w.equals(word)) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
        } else
        if (word.getWordStr().equals("8") || word.getWordStr().equals("8.0")) {
            Word w;
            for (int i = 0; i < this._allWords.size(); i++) {
                w = this._allWords.get(i);
                if ((w.getWordStr().equals("8") || w.getWordStr().equals("8.0")) &&
                    !w.equals(word)) {
                    double subProb = 0;
                    for (int k = 0; k < innerFeatures.size(); k++) {
                        InnerFeature innerFeature = innerFeatures.get(k);
                        subProb +=
                                this.getInnerFeatureProb(innerFeature, word1,
                                w);
                    }
                    //uniform assumption
                    ratio = 1.0 / (double) innerFeatures.size();
                    subProb *= ratio;

                    prob += subProb;
                }
            }
        }
        */


        this._gp_word_probs_w2.put(key, prob);

        return prob;

    }

    public double getProb(GProduction prod, Word word2, Word word1, Word word) {

        ArrayList<InnerFeature> innerFeatures = prod.getInnerFeatures();

        double prob = 0;
        for (int i = 0; i < innerFeatures.size(); i++) {
            InnerFeature innerFeature = innerFeatures.get(i);
            prob += this.getInnerFeatureProb(innerFeature, word2, word1, word);
        }

        //uniform assumption
        double ratio = 1.0 / (double) innerFeatures.size();
        prob *= ratio;

        return prob;

    }

    private double getInnerFeatureProb_Bigram(InnerFeature innerFeature,
                                              Word word1, Word word) {
        return (this.getInnerFeatureProb(innerFeature, word1, word) +
                this.getInnerFeatureProb(innerFeature, word)) / 2.0;
    }

    /**
     * Get the probability, given a InnerFeature and a Word
     * @param innerFeature InnerFeature
     * @param word Word
     * @return double
     */
    public double getInnerFeatureProb(InnerFeature innerFeature, Word word) {
        long key = KeyGen.toKey(innerFeature.getFeatureId(), word.getWordId());
        return this.getValue(this._prob_w1, key);
    }

    public double getInnerFeatureProb(InnerFeature innerFeature, Word word1,
                                      Word word) {

        if (this._prob_w2.size() == 0) {
            return this.getInnerFeatureProb(innerFeature, word);
        }

        long key = KeyGen.toKey(innerFeature.getFeatureId(), word1.getWordId(),
                                word.getWordId());
        return this.getValue(this._prob_w2, key);

    }

    public double getInnerFeatureProb(InnerFeature innerFeature, Word word2,
                                      Word word1, Word word) {

        long key = KeyGen.toKey(innerFeature.getFeatureId(), word2.getWordId(),
                                word1.getWordId(), word.getWordId());
        return this.getValue(this._prob_w3, key);

    }

    public double getHybridRuleProbability(GProduction gp, ArrayList<Integer> hybridWords) {

        int wIndex1 = Word.BEGIN_ID;
        int wIndex2 = Word.BEGIN_ID;

        double prob = 1.0;

        for (int i = 0; i < hybridWords.size(); i++) {
            int wIndex = hybridWords.get(i);
            double subProb = this.getHybridProb(gp, wIndex1, wIndex2, wIndex);
            prob *= subProb;
            
            wIndex1 = wIndex2;
            wIndex2 = wIndex;
            
        }

        prob *= this.getHybridProb(gp, wIndex1, wIndex2, Word.END_ID);

        return prob;

    }

    //simple interpolation-based smoothing method on mixing unigram, bigram and trigram probabilities...
    private double getHybridProb(GProduction gp, int wIndex1, int wIndex2,
                                 int wIndex) {

        double prev_bigram_count = this.getHybridBigramCount(gp, wIndex1,
                wIndex2);
        double prev_unigram_count = this.getHybridUnigramCount(gp, wIndex2);

        double result = 0;

        if (prev_bigram_count > 0) {
            double trigram_prob = this.getHybridTrigramCount(gp, wIndex1,
                    wIndex2, wIndex) /
                                  this.getHybridBigramCount(gp, wIndex1,
                    wIndex2);
            double bigram_prob = this.getHybridBigramCount(gp, wIndex2, wIndex) /
                                 this.getHybridUnigramCount(gp, wIndex2);
            double unigram_prob = this.getHybridUnigramCount(gp, wIndex) /
                                  this.
                                  getHybridUnigramCount(gp, this._allWords.size());

            result = (trigram_prob + bigram_prob + unigram_prob) / 3.0;
        } else if (prev_unigram_count > 0) {
            double bigram_prob = this.getHybridBigramCount(gp, wIndex2, wIndex) /
                                 this.getHybridUnigramCount(gp, wIndex2);
            double unigram_prob = this.getHybridUnigramCount(gp, wIndex) /
                                  this.
                                  getHybridUnigramCount(gp, this._allWords.size());

            result = (bigram_prob + unigram_prob) / 2.0;
        } else {
            double unigram_prob = this.getHybridUnigramCount(gp, wIndex) /
                                  this.
                                  getHybridUnigramCount(gp, this._allWords.size());

            result = unigram_prob;
        }

        return result;
    }

    private double getHybridUnigramCount(GProduction gp, int wIndex) {
        long key_unigram = KeyGen.toKey(gp.getProdId(), wIndex);
        return this.getValue(this.hybridRulesTable_unigram, key_unigram);
    }

    private double getHybridBigramCount(GProduction gp, int wIndex2, int wIndex) {
        long key_bigram = KeyGen.toKey(gp.getProdId(), wIndex2, wIndex);
        return this.getValue(this.hybridRulesTable_bigram, key_bigram);
    }

    private double getHybridTrigramCount(GProduction gp, int wIndex1,
                                         int wIndex2, int wIndex) {
        long key_trigram = KeyGen.toKey(gp.getProdId(), wIndex1, wIndex2,
                                        wIndex);
        return this.getValue(this.hybridRulesTable_trigram, key_trigram);
    }

    public void putHybridRule(GProduction gp, ArrayList<Integer> hybridWords) {

        int pIndex;

        pIndex = gp.getProdId();

        //putting the begin of sentence dummy words...
        Long key_unigram_begin = new Long(KeyGen.toKey(pIndex, Word.BEGIN_ID));
        if (!this.hybridRulesTable_unigram.containsKey(key_unigram_begin)) {
            this.hybridRulesTable_unigram.put(key_unigram_begin, new Double(2.0));
        } else {
            Double old_value = (Double)this.hybridRulesTable_unigram.get(
                    key_unigram_begin);
            this.hybridRulesTable_unigram.put(key_unigram_begin,
                                              new Double(old_value.doubleValue() +
                    2.0));
        }

        Long key_bigram_begin = new Long(KeyGen.toKey(pIndex, Word.BEGIN_ID,
                Word.BEGIN_ID));
        if (!this.hybridRulesTable_bigram.containsKey(key_bigram_begin)) {
            this.hybridRulesTable_bigram.put(key_bigram_begin, new Double(1.0));
        } else {
            Double old_value = (Double)this.hybridRulesTable_bigram.get(
                    key_bigram_begin);
            this.hybridRulesTable_bigram.put(key_bigram_begin,
                                             new Double(old_value.doubleValue() +
                    1.0));
        }

        //beginning of the sentence...
        int wIndex1 = Word.BEGIN_ID;
        int wIndex2 = Word.BEGIN_ID;

        for (int i = 0; i < hybridWords.size(); i++) {
            int wIndex = ((Integer) hybridWords.get(i)).intValue();

            //for the unigrams...
            Long key_unigram = new Long(KeyGen.toKey(pIndex, wIndex));
            if (!this.hybridRulesTable_unigram.containsKey(key_unigram)) {
                this.hybridRulesTable_unigram.put(key_unigram, new Double(1.0));
            } else {
                Double old_value = (Double)this.hybridRulesTable_unigram.get(
                        key_unigram);
                this.hybridRulesTable_unigram.put(key_unigram,
                                                  new Double(old_value.
                        doubleValue() + 1.0));
            }

            //for the denorminator
            //note: this is dummy, this._allWords.size() is not a word and has never been put
            //into hybridrulesTable anywhere else. the purpose is to keep a count of total unigram words.
            Long key_unigram_total = new Long(KeyGen.toKey(pIndex,
                    this._allWords.size()));
            if (!this.hybridRulesTable_unigram.containsKey(key_unigram_total)) {
                this.hybridRulesTable_unigram.put(key_unigram_total,
                                                  new Double(1.0));
            } else {
                Double old_value = (Double)this.hybridRulesTable_unigram.get(
                        key_unigram_total);
                this.hybridRulesTable_unigram.put(key_unigram_total,
                                                  new Double(old_value.
                        doubleValue() + 1.0));
            }

            //for the bigrams...
            Long key_bigram = new Long(KeyGen.toKey(pIndex, wIndex2, wIndex));
            if (!this.hybridRulesTable_bigram.containsKey(key_bigram)) {
                this.hybridRulesTable_bigram.put(key_bigram, new Double(1.0));
            } else {
                Double old_value = (Double)this.hybridRulesTable_bigram.get(
                        key_bigram);
                this.hybridRulesTable_bigram.put(key_bigram,
                                                 new Double(old_value.
                        doubleValue() + 1.0));
            }

            //for the trigrams...
            Long key_trigram = new Long(KeyGen.toKey(pIndex, wIndex1, wIndex2,
                    wIndex));
            if (!hybridRulesTable_trigram.containsKey(key_trigram)) {
                this.hybridRulesTable_trigram.put(key_trigram, new Double(1.0));
            } else {
                Double old_value = (Double)this.hybridRulesTable_trigram.get(
                        key_trigram);
                this.hybridRulesTable_trigram.put(key_trigram,
                                                  new Double(old_value.
                        doubleValue() + 1.0));
            }

            wIndex1 = wIndex2;
            wIndex2 = wIndex;
        }
    }

    /**
     * Put the count for the feature
     * @param f InnerFeature
     * @param word Word
     * @param count double
     */
    public void putCount(InnerFeature f, Word word, double count) {
        if (Double.isNaN(count)) {
            throw new ParamEstimationException("Not a Number!(580)");
        }

        int fIndex = f.getFeatureId();
        int wIndex = word.getWordId();

        long key = KeyGen.toKey(fIndex, wIndex);
        double oldCount = this.getValue(this._prob_w1_new, key);
        this._prob_w1_new.put(key, oldCount + count);

    }

    /**
     * Put the count for the feature
     * @param f InnerFeature
     * @param word1 Word
     * @param word Word
     * @param count double
     */
    public void putCount(InnerFeature f, Word word1, Word word, double count) {
        if (Double.isNaN(count)) {
            throw new ParamEstimationException("Not a Number!(601)");
        }

        int fIndex = f.getFeatureId();
        int wIndex1 = word1.getWordId();
        int wIndex = word.getWordId();

        long key = KeyGen.toKey(fIndex, wIndex1, wIndex);
        double oldCount = this.getValue(this._prob_w2_new, key);
        this._prob_w2_new.put(key, oldCount + count);

        //this.putCount(f, word, count);

        storeBigram(word1, word);
    }

    /**
     * Store the Bigram
     * @param word1 the previous word
     * @param word the current word
     */
    private void storeBigram(Word word1, Word word) {
//        System.err.println("store:\t" + word1.getWordStr() + "\t" + word.getWordStr());
        long key = word1.getWordId();
        if (!this._word1_to_word.containsKey(key)) {
            ArrayList<Integer> words = new ArrayList<Integer>();
            words.add(word.getWordId());
            this._word1_to_word.put(key, words);
        } else {
            ArrayList<Integer>
                    words = (ArrayList<Integer>)this._word1_to_word.get(key);
            if (!words.contains(word.getWordId())) {
                words.add(word.getWordId());
            }
        }
    }

    /**
     * Put the count for the feature
     * @param f InnerFeature
     * @param word2 Word
     * @param word1 Word
     * @param word Word
     * @param count double
     */
    public void putCount(InnerFeature f, Word word2, Word word1, Word word,
                         double count) {
        if (Double.isNaN(count)) {
            throw new ParamEstimationException("Not a Number!(642)");
        }

        int fIndex = f.getFeatureId();
        int wIndex2 = word2.getWordId();
        int wIndex1 = word1.getWordId();
        int wIndex = word.getWordId();

        long key = KeyGen.toKey(fIndex, wIndex2, wIndex1, wIndex);
        double oldCount = this.getValue(this._prob_w3_new, key);
        this._prob_w3_new.put(key, oldCount + count);

        storeTrigram(word2, word1, word);
    }

    /**
     * Store the trigram
     * @param word2 the second previous word
     * @param word1 the first previous word
     * @param word the current word
     */
    private void storeTrigram(Word word2, Word word1, Word word) {
        long key = KeyGen.toKey(word2.getWordId(), word1.getWordId());
        if (!this._word2_word1_to_word.containsKey(key)) {
            ArrayList<Integer> words = new ArrayList<Integer>();
            words.add(word.getWordId());
            this._word2_word1_to_word.put(key, words);
        } else {
            ArrayList<Integer>
                    words = (ArrayList<Integer>)this._word2_word1_to_word.get(key);
            if (!words.contains(word.getWordId())) {
                words.add(word.getWordId());
            }
        }
    }


    /**
     * Normalization of the word probability table and swap
     */
    private void normalizeAndSwapWordProb() {

        /** the unigram model **/
        if (!this._prob_w1_new.isEmpty()) {
            for (int fIndex = 0; fIndex < this._allInnerFeatures.size(); fIndex++) {
                double sum = 0;
                for (int wIndex = 0; wIndex < this._allWords.size(); wIndex++) {
                    long key = KeyGen.toKey(fIndex, wIndex);
                    if (this._prob_w1_new.containsKey(key)) {
                        double prob = ((Double)this._prob_w1_new.get(key)).
                                      doubleValue();
                        sum += prob;
                    }
                }
                if (sum != 0) {
                    for (int wIndex = 0; wIndex < this._allWords.size(); wIndex++) {
                        long key = KeyGen.toKey(fIndex, wIndex);
                        if (this._prob_w1_new.containsKey(key)) {
                            double prob = ((Double)this._prob_w1_new.get(key)).
                                          doubleValue();
                            this._prob_w1_new.put(key, prob / sum);
                        }
                    }
                }
            }

            this._prob_w1 = this._prob_w1_new;
            this._prob_w1_new = new HashMap<Long, Double>();
        }

        /** the bigram model **/
        for (int fIndex = 0; fIndex < this._allInnerFeatures.size(); fIndex++) {
            Iterator<Long> unigramKeys = this._word1_to_word.keySet().iterator();
            
            while (unigramKeys.hasNext()) {
                double sum = 0;

                long unigramkey = unigramKeys.next();
                int wIndex1 = (int) unigramkey;
                ArrayList<Integer>
                        words = (ArrayList<Integer>)this._word1_to_word.
                                get(unigramkey);
                for (int i = 0; i < words.size(); i++) {
                    int wIndex = words.get(i);
                    long key = KeyGen.toKey(fIndex, wIndex1, wIndex);
                    if (this._prob_w2_new.containsKey(key)) {
                        double prob = ((Double)this._prob_w2_new.get(key)).
                                      doubleValue();
                        sum += prob;
                    }
                }
                if (sum != 0) {
                    for (int i = 0; i < words.size(); i++) {
                        int wIndex = words.get(i);
                        long key = KeyGen.toKey(fIndex, wIndex1, wIndex);
                        if (this._prob_w2_new.containsKey(key)) {
                            double prob = ((Double)this._prob_w2_new.get(key)).
                                          doubleValue();
                            this._prob_w2_new.put(key, prob / sum);
                        }
                    }
                }
            }
        }

        this._prob_w2 = this._prob_w2_new;
        this._prob_w2_new = new HashMap<Long, Double>();
        this._word1_to_word = new HashMap<Long, ArrayList<Integer>>();

    }

    /**
     * Normalization of the word probability table and swap
     * Katz Interpolation...
     */
    private void normalizeAndInterpolateWordProb() {

        for (int fIndex = 0; fIndex < this._allInnerFeatures.size(); fIndex++) {
            Iterator<Long> unigramKeys = this._word1_to_word.keySet().iterator();
            while(unigramKeys.hasNext()){
                double sum = 0;

                long unigramkey = unigramKeys.next();
                int wIndex1 = (int) unigramkey;
                ArrayList<Integer>
                        words = (ArrayList<Integer>)this._word1_to_word.
                                get(unigramkey);
                for (int i = 0; i < words.size(); i++) {
                    int wIndex = words.get(i);
                    long key = KeyGen.toKey(fIndex, wIndex1, wIndex);
                    if (this._prob_w2_new.containsKey(key)) {
                        double prob = ((Double)this._prob_w2_new.get(key)).
                                      doubleValue();
                        sum += prob;
                    }
                }
                if (sum != 0) {
                    for (int i = 0; i < words.size(); i++) {
                        int wIndex = words.get(i);
                        long key = KeyGen.toKey(fIndex, wIndex1, wIndex);
                        if (this._prob_w2_new.containsKey(key)) {
                            double prob = ((Double)this._prob_w2_new.get(key)).
                                          doubleValue();
                            this._prob_w2_new.put(key, prob / sum);
                        }
                    }
                }
            }
        }

    }

    /**
     * Perform all the normalizations
     */
    public void normalizeAndSwap() {
        normalizeAndSwapWordProb();
        normalizeAndSwapPatternProb();
        this._gp_word_probs = new HashMap<Long, Double>();
        this._gp_word_probs_w2 = new HashMap<Long, Double>();
    }

    public void normalizeAndInterpolate() {
        normalizeAndInterpolateWordProb();
        normalizeAndSwapPatternProb();
    }

    public void initRand() {

        for (int fIndex = 0; fIndex < this._allInnerFeatures.size(); fIndex++) {
            double sum = 0;
            for (int wIndex = 0; wIndex < this._allWords.size(); wIndex++) {
                Long key1 = new Long(KeyGen.toKey(fIndex, wIndex));
                double prob = 1 - Math.random();
                sum += prob;
                this._prob_w1.put(key1, new Double(prob));
            }
            for (int wIndex = 0; wIndex < this._allWords.size(); wIndex++) {
                Long key1 = new Long(KeyGen.toKey(fIndex, wIndex));
                double prob = ((Double)this._prob_w1.get(key1)).doubleValue();
                this._prob_w1.put(key1, new Double(prob / sum));
            }
        }

        //initialize the pattern probabilities...(uniform)
        for (int pIndex = 0; pIndex < this._allGProds.size(); pIndex++) {
            GProduction prod = this._allGProds.get(pIndex);
            int ruleSize;
            if (prod.countRHSNonterminals() == 0) {
                ruleSize = Rule.totalRules_0;
            } else if (prod.countRHSNonterminals() == 1) {
                ruleSize = Rule.totalRules_1;
            } else {
                ruleSize = Rule.totalRules_2;
            }

            for (int rIndex = 0; rIndex < ruleSize; rIndex++) {
                Long key = new Long(KeyGen.toKey(prod.getProdId(), rIndex));
                this._patternTable.put(key, new Double(1.0 / (double) ruleSize));
            }
        }

    }

    public void initUnif() {

        //w1 probabilities
        for (int fIndex = 0; fIndex < this._allInnerFeatures.size(); fIndex++) {
            double sum = 0;
            for (int wIndex = 0; wIndex < this._allWords.size(); wIndex++) {
                long key = KeyGen.toKey(fIndex, wIndex);
                double prob = 1.0;
                this._prob_w1.put(key, new Double(prob));
                sum += prob;
            }

            for (int wIndex = 0; wIndex < this._allWords.size(); wIndex++) {
                long key = KeyGen.toKey(fIndex, wIndex);
                double prob = ((Double)this._prob_w1.get(key)).doubleValue();
                this._prob_w1.put(key, new Double(prob / sum));
            }
        }

        //initialize the pattern probabilities...
        for (int pIndex = 0; pIndex < this._allGProds.size(); pIndex++) {
            GProduction prod = this._allGProds.get(pIndex);
            int ruleSize;
            if (prod.countRHSNonterminals() == 0) {
                ruleSize = Rule.totalRules_0;
            } else if (prod.countRHSNonterminals() == 1) {
                ruleSize = Rule.totalRules_1;
            } else {
                ruleSize = Rule.totalRules_2;
            }

            for (int rIndex = 0; rIndex < ruleSize; rIndex++) {
                long key = KeyGen.toKey(prod.getProdId(), rIndex);
                this._patternTable.put(key, new Double(1.0 / (double) ruleSize));
            }
        }

    }

    @SuppressWarnings("unchecked")
	private HashMap<Integer, Double>[] counts(TrainExample example) {
        ArrayList<Word> words = example.getWords();
        ArrayList<Production> prods = example.getProductions();

        double c = 1.0 * example.getWeight();

        HashMap<Integer, Double> countTable_1 = new HashMap<Integer, Double>();
        HashMap<Integer, Double> countTable_2 = new HashMap<Integer, Double>();
        HashMap<Integer, Double> countTable_3 = new HashMap<Integer, Double>();

        for (int i = 0; i < words.size(); i++) {
            int word = words.get(i).getWordId();
            if (!countTable_1.containsKey(word)) {
                countTable_1.put(word, c);
            } else {
                double count = (Double) countTable_1.get(word);
                countTable_1.put(word, count + c);
            }

            if (!countTable_2.containsKey(word)) {
                countTable_2.put(word, c);
            } else {
                double count = (Double) countTable_2.get(word);
                countTable_2.put(word, count + c);
            }
        }

        for (int i = 0; i < prods.size(); i++) {
            int gp = prods.get(i).getGProduction().getProdId();
            if (!countTable_3.containsKey(gp)) {
                countTable_3.put(gp, c);
            } else {
                double count = (Double) countTable_3.get(gp);
                countTable_3.put(gp, count + c);
            }

            int word;
            word = Word.BEGIN.getWordId();
            if (!countTable_1.containsKey(word)) {
                countTable_1.put(word, c);
            } else {
                double count = (Double) countTable_1.get(word);
                countTable_1.put(word, count + c);
            }

            word = Word.END.getWordId();
            if (!countTable_2.containsKey(word)) {
                countTable_2.put(word, c);
            } else {
                double count = (Double) countTable_2.get(word);
                countTable_2.put(word, count + c);
            }

            if (prods.get(i).countRHSNonterminals() >= 1) {
                word = Word.CHILD_A.getWordId();
                if (!countTable_1.containsKey(word)) {
                    countTable_1.put(word, c);
                } else {
                    double count = (Double) countTable_1.get(word);
                    countTable_1.put(word, count + c);
                }

                if (!countTable_2.containsKey(word)) {
                    countTable_2.put(word, c);
                } else {
                    double count = (Double) countTable_2.get(word);
                    countTable_2.put(word, count + c);
                }
            }

            if (prods.get(i).countRHSNonterminals() == 2) {
                word = Word.CHILD_B.getWordId();
                if (!countTable_1.containsKey(word)) {
                    countTable_1.put(word, c);
                } else {
                    double count = (Double) countTable_1.get(word);
                    countTable_1.put(word, count + c);
                }

                if (!countTable_2.containsKey(word)) {
                    countTable_2.put(word, c);
                } else {
                    double count = (Double) countTable_2.get(word);
                    countTable_2.put(word, count + c);
                }
            }

        }

        return new HashMap[] {countTable_1, countTable_2, countTable_3};
    }

    private double getCount(HashMap<Integer, Double> countTable, Object word) {
        if (!countTable.containsKey(word)) {
            return 0;
        } else {
            return (Double) countTable.get(word);
        }
    }

    public boolean checkCorrect_new(TrainExample example) {
        HashMap<Integer, Double>[] countTables = this.counts(example);
        HashMap<Integer, Double> countTable_1 = countTables[0];
        HashMap<Integer, Double> countTable_2 = countTables[1];
        HashMap<Integer, Double> countTable_3 = countTables[2];

        for (int w1Index = 0; w1Index < this._allWords.size(); w1Index++) {
            Word word1 = this._allWords.get(w1Index);

            ArrayList<Integer>
                    words = (ArrayList<Integer>)this._word1_to_word.get(new
                    Long(word1.getWordId()));
            double sum = 0;
            if (words != null) {
                for (int wIndex = 0; wIndex < words.size(); wIndex++) {
                    Word word = this._allWords.get((Integer) words.get(wIndex));
                    for (int fIndex = 0; fIndex < this._allInnerFeatures.size();
                                      fIndex++) {
                        InnerFeature feature = this._allInnerFeatures.get(
                                fIndex);
                        double value = this.getInnerFeatureProb_new(feature,
                                word1, word);
                        sum += value;
                    }
                }
            }

            double corrSum = getCount(countTable_1, word1.getWordId()) / 2;
            if (Math.abs(corrSum - sum) > Math.pow(10, -10)) {
                System.err.println("Error(1)!\t" + sum + "\t" + corrSum);
                System.err.println("word1=\t" + word1 + "\t" +
                                   countTable_1.get(word1.getWordId()) +
                                   "\tsize:" + countTable_1.size());

                System.err.println("------------" +
                                   countTable_1.containsKey(word1.getWordId()) +
                                   "\t" +
                                   countTable_1.containsKey(Word.BEGIN.
                        getWordId()) + "\t" + word1.hashCode() + "\t" +
                                   Word.BEGIN.hashCode() + "\t" +
                                   word1.equals(Word.BEGIN));
                Iterator<Integer> keys = countTable_1.keySet().iterator();
                while (keys.hasNext()) {
                    int word = keys.next();
                    System.err.println(word + "\t" + countTable_1.get(word));
                }

                return false;
            }
        }

        for (int wIndex = 0; wIndex < this._allWords.size(); wIndex++) {
            Word word = this._allWords.get(wIndex);
            double sum = 0;
            for (int fIndex = 0; fIndex < this._allInnerFeatures.size(); fIndex++) {
                InnerFeature feature = this._allInnerFeatures.get(fIndex);
                double value = this.getInnerFeatureProb_new(feature, word);
                sum += value;
            }

            for (int w1Index = 0; w1Index < this._allWords.size(); w1Index++) {
                Word word1 = this._allWords.get(w1Index);
                for (int fIndex = 0; fIndex < this._allInnerFeatures.size();
                                  fIndex++) {
                    InnerFeature feature = this._allInnerFeatures.get(fIndex);
                    double value = this.getInnerFeatureProb_new(feature, word1,
                            word);
                    sum += value;
                }
            }

            double corrSum = getCount(countTable_2, word.getWordId());
            if (Math.abs(corrSum - sum) > Math.pow(10, -10)) {
                System.err.println("Error!(2)\t" + sum + "\t" + corrSum);
                System.err.println("word=\t" + word + "\t" +
                                   countTable_2.get(word.getWordId()));
                return false;
            }
        }

        for (int pIndex = 0; pIndex < this._allGProds.size(); pIndex++) {
            GProduction gp = this._allGProds.get(pIndex);
            double sum = 0;
            for (int ptIndex = 0; ptIndex < this._patternIDs.length; ptIndex++) {
                long key = KeyGen.toKey(gp.getProdId(), ptIndex);
                if (this._patternTable_new.containsKey(key)) {
                    double prob = (Double)this._patternTable_new.get(key);
                    sum += prob;
                }
            }

            double corrSum = getCount(countTable_3, gp.getProdId());
            if (Math.abs(corrSum - sum) > Math.pow(10, -10)) {
                System.err.println();
                System.err.println("Example " + example.getExampleId());
                System.err.println("Error!(3)\t" + sum + "\t" + corrSum);
                System.err.println("gp \t" + gp.toLine() + "\tshould be\t" +
                                   countTable_3.get(gp.getProdId()));
                for (int ptIndex = 0; ptIndex < this._patternIDs.length;
                                   ptIndex++) {
                    long key = KeyGen.toKey(gp.getProdId(), ptIndex);
                    if (this._patternTable_new.containsKey(key)) {
                        double prob = (Double)this._patternTable_new.get(key);
                        sum += prob;
                        System.err.println(gp.toLine() + "\t" + ptIndex + "\t" +
                                           prob);
                    }
                }

                return false;
            }
        }

        /**/

        return true;

    }

    public void reset_new() {
        this._prob_w1_new = new HashMap<Long, Double>();
        this._prob_w2_new = new HashMap<Long, Double>();
        this._patternTable_new = new HashMap<Long, Double>();
    }

    public String toString_new() {
    	
        System.err.print("*");
        this._gp_word_probs_w2 = new HashMap<Long, Double>();
        this._gp_word_probs = new HashMap<Long, Double>();
        
        return null;

        /*
                 StringBuffer sb = new StringBuffer();

                 sb.append("=Feature Parameter=");

                 sb.append('\n');
                 sb.append("---- First Word ----");
                 sb.append('\n');
         for (int w1Index = 0; w1Index < this._allWords.size(); w1Index++) {
            Word word1 = this._allWords.get(w1Index);
            ArrayList<Integer> words = (ArrayList<Integer>)this._word1_to_word.get(new Long(word1.getWordId()));
            if (words != null) {
                double sum = 0;
                for (int wIndex = 0; wIndex < words.size(); wIndex++) {
         Word word = this._allWords.get((Integer) words.get(wIndex));
         for (int fIndex = 0; fIndex < this._allInnerFeatures.size(); fIndex++) {
         InnerFeature feature = this._allInnerFeatures.get(fIndex);
         double value = (this.getInnerFeatureProb_new(feature, word1, word));
                        sum += value;
                    }
                }
                if (sum > 0) {
                    sb.append(word1.toString() + "\t" + sum);
                    sb.append('\n');
                }
            }
                 }

                 sb.append('\n');
                 sb.append("---- Second Word ----");
                 sb.append('\n');
         for (int wIndex = 0; wIndex < this._allWords.size(); wIndex++) {
            Word word = this._allWords.get(wIndex);
            double sum = 0;
         for (int fIndex = 0; fIndex < this._allInnerFeatures.size(); fIndex++) {
                InnerFeature feature = this._allInnerFeatures.get(fIndex);
                double value = this.getInnerFeatureProb_new(feature, word);
                sum += value;
            }

            for (int w1Index = 0; w1Index < this._allWords.size(); w1Index++) {
                Word word1 = this._allWords.get(w1Index);
         for (int fIndex = 0; fIndex < this._allInnerFeatures.size(); fIndex++) {
                    InnerFeature feature = this._allInnerFeatures.get(fIndex);
         double value = this.getInnerFeatureProb_new(feature, word1, word);
                    sum += value;
                }
            }
            if (sum > 0) {
                sb.append(word.toString() + "\t" + sum);
                sb.append('\n');
            }
                 }

                 sb.append('\n');
                 sb.append("---- Rule Pattern ----");
                 sb.append('\n');

         for (int pIndex = 0; pIndex < this._allGProds.size(); pIndex++) {
            GProduction gp = this._allGProds.get(pIndex);
            double sum = 0;
         for (int ptIndex = 0; ptIndex < this._patternIDs.length; ptIndex++) {
                long key = KeyGen.toKey(gp.getProdId(), ptIndex);
                if (this._patternTable_new.containsKey(key)) {
                    double prob = (Double)this._patternTable_new.get(key);
                    sum += prob;
                }
            }
            if (sum > 0) {
                sb.append(gp.toLine() + "\t" + sum);
                sb.append('\n');
            }
                 }

                 return sb.toString();
         */
    }


    public double getInnerFeatureProb_new(InnerFeature innerFeature, Word word1,
                                          Word word) {
        long key = KeyGen.toKey(innerFeature.getFeatureId(), word1.getWordId(),
                                word.getWordId());
        return this.getValue(this._prob_w2_new, key);
    }

    public double getInnerFeatureProb_new(InnerFeature innerFeature, Word word) {
        long key = KeyGen.toKey(innerFeature.getFeatureId(), word.getWordId());
        return this.getValue(this._prob_w1_new, key);
    }


    public String toString() {

        StringBuffer sb = new StringBuffer();

        sb.append("=Feature Parameter=");
        sb.append('\n');

        for (int fIndex = 0; fIndex < this._allInnerFeatures.size(); fIndex++) {
            InnerFeature feature = this._allInnerFeatures.get(fIndex);
            sb.append("-" + feature.toString() + "-");
            sb.append('\n');
            for (int wIndex = 0; wIndex < this._allWords.size(); wIndex++) {
                Word word = this._allWords.get(wIndex);
                double value = this.getInnerFeatureProb(feature, word);
                if (value > Math.pow(10, -4)) {
                    sb.append(feature.toString() + "\t" + word.toString() +
                              "\t" + value);
                    sb.append("\n");
                }
            }
            sb.append("\n");
        }

        sb.append("Feature Size:\t" + this._allInnerFeatures.size());
        sb.append('\n');
        sb.append("GProduction Size:\t" + this._allGProds.size());
        sb.append('\n');
        sb.append("Word Size:\t" + this._allWords.size());
        sb.append('\n');

        return sb.toString();

    }

    private class ParamEstimationException extends RuntimeException {
		private static final long serialVersionUID = -7920244619286549143L;

		public ParamEstimationException(String msg) {
            super(msg);
        }
    }
}
