/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.core;

/**
 *
 * @author Wei LU
 * @version 1.0
 */
import hybridtree.data.GProduction;
import hybridtree.data.Production;

import java.io.Serializable;
import java.util.ArrayList;

public class ResElem implements Comparable<ResElem>, Serializable {
	
	private static final long serialVersionUID = -1004891753153702482L;
	private GProduction _gp;
    private int _bIndex;
    private int _eIndex;
    private ResElem[] _children;
    private int _rulePattern = -1;
    private double _prob = Double.NEGATIVE_INFINITY;
    
    private Production _prodTree; // convert this hybridNode to a Production tree
    
    public static ResElem _dummy = new ResElem(null, -1, -1, Double.NEGATIVE_INFINITY);

    public ResElem(GProduction gp, int bIndex, int eIndex, double prob) {
        this._gp = gp;
        this._bIndex = bIndex;
        this._eIndex = eIndex;
        this._prob = prob;
        this._children = new ResElem[0];
    }

    public ResElem(GProduction gp, int bIndex, int eIndex, double prob,
                   int rulePattern, ResElem[] children) {
        this._gp = gp;
        this._bIndex = bIndex;
        this._eIndex = eIndex;
        this._prob = prob;
        this._rulePattern = rulePattern;
        this._children = children;
    }

    public boolean equals(Object o) {

        if (o instanceof ResElem) {
            ResElem res = (ResElem) o;

            if (this._prob != res._prob) {
                return false;
            }
            if (this._bIndex != res._bIndex) {
                return false;
            }
            if (this._eIndex != res._eIndex) {
                return false;
            }
            if (this._children.length != res._children.length) {
                return false;
            }

            if (this._gp != null) {
                if (!this._gp.equals(res._gp)) {
                    return false;
                }
            } else {
                if (res._gp != null) {
                    return false;
                }
            }

            for (int i = 0; i < this._children.length; i++) {
                if (this._children[i] != res._children[i]) {
                    return false;
                }
            }

            return true;
        }

        return false;
    }

    public int compareTo(ResElem o) {
        ResElem res = (ResElem) o;
        double diff = this._prob - res._prob;
        if (diff == 0) {
            if (this == o) {
                return 0;
            } else {
                if (this._rulePattern > res._rulePattern) {
                    return +1;
                } else if (this._rulePattern < res._rulePattern) {
                    return -1;
                } else {
                    if (this._children.length > res._children.length) {
                        return +1;
                    } else if (this._children.length < res._children.length) {
                        return -1;
                    } else {
                        for (int i = 0; i < this._children.length; i++) {
                            ResElem child1 = this._children[i];
                            ResElem child2 = res._children[i];
                            int subRes = child1.compareTo(child2);
                            if (subRes != 0) {
                                return subRes;
                            }
                        }
                        return 0;
                    }
                }
            }
        } else if (diff < 0) {
            return -1;
        } else {
            return +1;
        }
    }

    public double getProb() {
        return this._prob;
    }

    public GProduction getGP() {
        return this._gp;
    }

    public int getBIndex() {
        return this._bIndex;
    }

    public int getEIndex() {
        return this._eIndex;
    }

    public int getRulePattern() {
        return this._rulePattern;
    }

    public void setRulePattern(int rulePattern) {
        this._rulePattern = rulePattern;
    }

    public ResElem[] getChildren() {
        return this._children;
    }

    public ResElem getFirstChild() {
        if (this._children.length > 0) {
            return this._children[0];
        }
        return null;
    }

    public ResElem getChildAt(int cIndex) {
        return this._children[cIndex];
    }

    public void setChild(int cIndex, ResElem child) {
        this._children[cIndex] = child;
    }

    public String toString() {
        return toString(0);
    }

    /**
     * Converting the ResElem to a Production tree
     * @return Production
     */
    public Production getProdTree() {
        if (this._prodTree != null) {
            return this._prodTree;
        }
        Production root = new Production(this._gp);
        if (this._children.length != 0) {
            ArrayList<Production> cProds = new ArrayList<Production>(this._children.length);
            for (int i = 0; i < this._children.length; i++) {
                Production cProd = this._children[i].getProdTree();
                cProd.setParent(root);
                cProds.add(cProd);
                root.setChildren(cProds);
            }
        }
        this._prodTree = root;

        return root;
    }


    public String toString(int indent) {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < indent; i++) {
            sb.append(' ');
        }
        if (this._gp != null) {
            sb.append(this._gp.toLine());
        } else {
            sb.append("DUMMY_GPRODUCTION");
        }
        sb.append('[');
        sb.append(this._bIndex);
        sb.append(',');
        sb.append(this._eIndex);
        sb.append(']');
        sb.append(this._prob);
        sb.append('<');
        sb.append(this._rulePattern);
        sb.append('>');
        sb.append('\n');

        for (int i = 0; i < this._children.length; i++) {
            ResElem child = this._children[i];
            sb.append(child.toString(indent + 1));
        }
        return sb.toString();
    }
}
