/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.core.learn;

/**
 * @author LU, Wei
 * @version 1.0
 */

import hybridtree.core.*;
import hybridtree.data.*;

import java.util.*;


public class InitLearner {

    private int _ITERATIONS = 100;
    
    private Model _model;

    private double _loglikelihood = Double.NEGATIVE_INFINITY;

    private Param _param;

    public InitLearner(Param _param, Model model) {
        this._model = model;
        this._param = _param;
    }

    public void learn() {
        initUnif();

        computeMRLM();

        for (int iteration = 0; iteration < this._ITERATIONS; iteration++) {
            System.err.print("\rIteration " + iteration + "/" + this._ITERATIONS);
            em_step();
//            this.getJointLikelihood();
        }

        System.err.println("\rOK!                                                                          ");
    }
    
    @SuppressWarnings("unused")
	private void initRand() {
        this._param.initRand();
    }

    private void initUnif() {
        this._param.initUnif();
    }

    private void computeMRLM() {
        System.err.println("Learning MRL Language Model");
        this._param.initMRLM();

        ArrayList<Example> train_examples = this._model.getTrainExamples();

        for (int i = 0; i < train_examples.size(); i++) {
            TrainExample example = (TrainExample) train_examples.get(i);
            if (example.getExampleId() < 0) {
                continue;
            }
            ArrayList<Production> prods = example.getProductions();
            for (int j = 0; j < prods.size(); j++) {
                Production prod = (Production) prods.get(j);
                this._param.putMRLMCount(prod.getGProduction(), 1.0);
                ArrayList<Production> children = prod.getChildren();
                for (int childIndex = 0; childIndex < prod.countChildren(); childIndex++) {
                    Production child = children.get(childIndex);
                    GProduction g2 = child.getGProduction();
                    if (g2.isConstant()) {
                    } else {
                        this._param.putMRLMCount(prod.getGProduction(), childIndex, child.getGProduction(), 1.0);
                    }
                }
            }
        }

        this._param.normalizeMRLCount();

        System.err.println("MRL Language Model Parameters Learned!");

        /*
                 train_examples = this._model.getTestExamples();

                 for (int i = 0; i < train_examples.size(); i++) {
            TrainExample example = (TrainExample) train_examples.get(i);
            if (example.getExampleId() < 0) {
                continue;
            }
            double prob = this._param.computeMRLLM(example.getRootProduction());
            if (prob < Math.pow(10, -10) || prob == 1) {
                System.err.println("example:" + example.getExampleId() + "=" + prob);
                System.err.println(example.getNL());
                ArrayList<Production> prods = example.getProductions();
                for (int j = 0; j < prods.size(); j++) {
                    System.err.println(prods.get(j).toLine());
                }
            }
                 }
                 System.exit(1);
         */

    }

    private void em_step() {

        ArrayList<Example> train_examples = this._model.getTrainExamples();

        //e step + m step
        for (int i = 0; i < train_examples.size(); i++) {

            TrainExample example = (TrainExample) train_examples.get(i);

            ArrayList<Word> words = example.getWords();
            ArrayList<InnerFeature> features = example.getInnerFeatures();
            ArrayList<Production> prods = example.getProductions();

            double weight = example.getWeight();

            double[][] h = new double[words.size()][features.size()]; //membership matrix

            for (int wPos = 0; wPos < words.size(); wPos++) {

                Word word = words.get(wPos);
                double total = 0;

                int featureIndex = 0;

                for (int pPos = 0; pPos < prods.size(); pPos++) {
                    Production prod = prods.get(pPos);
                    double count = this._param.getProb(prod.getGProduction(), word);

                    ArrayList<InnerFeature> fs = prod.getGProduction().getInnerFeatures();
                    double ratio = 1.0 / (double) fs.size();
                    for (int fPos = 0; fPos < fs.size(); fPos++) {
                        h[wPos][featureIndex] = count * ratio;
                        total += h[wPos][featureIndex];
                        featureIndex++;
                    }
                }

                for (int fPos = 0; fPos < features.size(); fPos++) {
                    h[wPos][fPos] /= total;
                }

                for (int fPos = 0; fPos < features.size(); fPos++) {
                    InnerFeature feature = features.get(fPos);
                    double count = h[wPos][fPos] * weight;
                    this._param.putCount(feature, word, count);
                }

            }

            for (int pPos = 0; pPos < prods.size(); pPos++) {
                Production prod = prods.get(pPos);

                ArrayList<InnerFeature> fs = prod.getGProduction().getInnerFeatures();
                double count = 1.0 / (double) fs.size();

                count *= weight;

                int childrenSize = prod.countChildren();
                for (int fPos = 0; fPos < fs.size(); fPos++) {
                    InnerFeature feature = fs.get(fPos);
                    if (childrenSize == 0) {
                        this._param.putCount(feature, Word.END, count);
                    } else if (childrenSize == 1) {
                        this._param.putCount(feature, Word.CHILD_A, count);
                        this._param.putCount(feature, Word.END, count);
                    } else {
                        this._param.putCount(feature, Word.CHILD_A, count);
                        this._param.putCount(feature, Word.CHILD_B, count);
                        this._param.putCount(feature, Word.END, count);
                    }
                }
            }

        }

        this._param.normalizeAndSwap();

    }

    public Param getParam() {
        return this._param;
    }

    @SuppressWarnings("unused")
	private void getJointLikelihood() {

        ArrayList<Example> examples = this._model.getTrainExamples();

        double loglikelihood = 0;
        for (int i = 0; i < examples.size(); i++) {

            TrainExample example = (TrainExample) examples.get(i);
            double weight = example.getWeight();

            ArrayList<Word> words = example.getWords();
            ArrayList<InnerFeature> features = example.getFirstInnerFeatures();
            ArrayList<Production> prods = example.getProductions();

            double logProb = 0;
            for (int wPos = 0; wPos < words.size(); wPos++) {
                Word word = words.get(wPos);
                double subProb = 0;

                for (int fPos = 0; fPos < features.size(); fPos++) {
                    InnerFeature feature = features.get(fPos);
                    subProb += this._param.getInnerFeatureProb(feature, word);
                }

                logProb += Math.log(subProb);
            }

            for (int fPos = 0; fPos < features.size(); fPos++) {
                InnerFeature feature = features.get(fPos);
                double subProb = this._param.getInnerFeatureProb(feature, Word.END);
                logProb += Math.log(subProb);
            }

            for (int fPos = 0; fPos < features.size(); fPos++) {
                InnerFeature feature = features.get(fPos);

                Production prod = prods.get(fPos);
                int childrenSize = prod.countChildren();
                if (childrenSize == 0) {
                    logProb += Math.log(this._param.getInnerFeatureProb(feature, Word.END));
                } else if (childrenSize == 1) {
                    logProb += Math.log(this._param.getInnerFeatureProb(feature, Word.CHILD_A));
                    logProb += Math.log(this._param.getInnerFeatureProb(feature, Word.END));
                } else {
                    logProb += Math.log(this._param.getInnerFeatureProb(feature, Word.CHILD_A));
                    logProb += Math.log(this._param.getInnerFeatureProb(feature, Word.CHILD_B));
                    logProb += Math.log(this._param.getInnerFeatureProb(feature, Word.END));
                }
            }

            loglikelihood += logProb * weight;

        }

        double improvement = loglikelihood - this._loglikelihood;
        System.err.println("The loglikelihood: " + loglikelihood);
        if (improvement < 0) {
            System.err.println("Error!");
            System.exit(1);
        }

        this._loglikelihood = loglikelihood;
    }


}
