/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.core.learn;

/**
 * @author LU, Wei
 * @version 1.0
 */

import java.util.ArrayList;

import hybridtree.core.*;
import hybridtree.data.Production;
import hybridtree.data.TrainExample;


public abstract class TrainTable extends DeriveTable {

    /**
     * the Productions
     */
    protected ArrayList<Production> _productions;

    /**
     * the beta values
     */
    protected double[][][] _beta;

    /**
     * the alpha values
     */
    protected double[][][] _alpha;

    /**
     * the weight of this example
     */
    protected double _weight = 1.0;

    /**
     * whether this example should be ignored.
     */
    protected boolean _ignoreFlag = false;

    /**
     * Constructor
     * @param example the Example
     * @param parameter the parameter
     */
    protected TrainTable(TrainExample example, Param parameter) {

        super(example, parameter);

        this._productions = example.getProductions();
        this._weight = example.getWeight();

    }

    /**
     * Set the collection of Productions
     * @param prods a collection of Productions
     */
    public void setProductions(ArrayList<Production> prods) {
        this._productions = prods;
    }

    /**
     * Get the global inside probability
     * @return the inside probability
     */
    public double getGlobalBeta() {
        return this.getBeta(this._productions.get(0), 0, this._N); // the root.
    }

    /**
     * Get the collection of Productions
     * @return the collection of Productions
     */
    public ArrayList<Production> getProductionArrayLists() {
        return this._productions;
    }

    /**
     * Get the Loglikelihood
     * @return the log-likelhood value
     */
    public double getLogLikelihood() {
        return Math.log(getLikelihood());
    }

    /**
     * Get the likelihood
     * @return the likelihood value
     */
    public double getLikelihood() {
        return this._beta[0][this._N][0];
    }

    /**
     * Check whether this parse instance should be ignored.
     * @return ignore flag
     */
    public boolean ignore() {
        return this._ignoreFlag;
    }


    protected double getBeta(Production prod, int bIndex, int eIndex) {
        return this._beta[bIndex][eIndex - bIndex][prod.getProdSeq()];
    }

    protected double getAlpha(Production prod, int bIndex, int eIndex) {
        return this._alpha[bIndex][eIndex - bIndex][prod.getProdSeq()];
    }

    public abstract void computeAlpha();

    public abstract void computeBeta();

    public static UGTrainTable createUGTable(TrainExample example, Param parameter) {
        return new UGTrainTable(example, parameter);
    }

    public static BGTrainTable createBGTable(TrainExample example, Param parameter) {
        return new BGTrainTable(example, parameter);
    }

    public static MBGTrainTable createMBGTable(TrainExample example, Param parameter) {
        return new MBGTrainTable(example, parameter);
    }

    public static NMBGTrainTable createNMBGTable(TrainExample example, Param parameter) {
        return new NMBGTrainTable(example, parameter);
    }

    public static class ParamUpdateException extends RuntimeException {
		private static final long serialVersionUID = -6249560436505334731L;

		ParamUpdateException(String msg) {
            super(msg);
        }
    }

}



