/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.core.learn;

/**
 * @author LU, Wei
 * @version 1.0
 */

import java.util.ArrayList;

import hybridtree.core.*;
import hybridtree.data.InnerFeature;
import hybridtree.data.Production;
import hybridtree.data.Rule;
import hybridtree.data.TrainExample;
import hybridtree.data.Word;


public class UBGTrainTable extends TrainTable {

    private double beta_w[][][];
    private double beta_Xw[][][][];
    private double beta_wX[][][][];
    private double beta_w1Xw2[][][][];
    private double beta_X[][][][];

    private double alpha_w[][][];
    private double alpha_Xw[][][][];
    private double alpha_wX[][][][];
    private double alpha_w1Xw2[][][][];

    /**
     * Constructor
     * @param example the Example
     * @param parameter the parameter
     */
    protected UBGTrainTable(TrainExample example, Param parameter) {

        super(example, parameter);

        this._productions = example.getProductions();
        this._weight = example.getWeight();

    }

    /****************************************************************/
    /****************** End Of Public Methods ***********************/
    /****************************************************************/


    private double getBeta_w(Production prod, int bIndex, int eIndex) {
        if (bIndex == eIndex) {
            return 0;
        }
        return this.beta_w[prod.getProdSeq()][bIndex][eIndex - bIndex];
    }

    private double getBeta_X(Production prod, int cIndex, int bIndex, int eIndex) {
        if (bIndex == eIndex) {
            return 0;
        }
        return this.beta_X[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private double getBeta_Xw(Production prod, int cIndex, int bIndex, int eIndex) {
        if (bIndex == eIndex) {
            return 0;
        }
        return this.beta_Xw[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private double getBeta_wX(Production prod, int cIndex, int bIndex, int eIndex) {
        if (bIndex == eIndex) {
            return 0;
        }
        return this.beta_wX[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private double getBeta_w1Xw2(Production prod, int cIndex, int bIndex, int eIndex) {
        if (bIndex == eIndex) {
            return 0;
        }
        return this.beta_w1Xw2[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    /**
     * Get Alpha Temporary Table Methods.
     */
    private double getAlpha_w(Production prod, int bIndex, int eIndex) {
        return this.alpha_w[prod.getProdSeq()][bIndex][eIndex - bIndex];
    }

    private double getAlpha_Xw(Production prod, int cIndex, int bIndex, int eIndex) {
        return this.alpha_Xw[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private double getAlpha_wX(Production prod, int cIndex, int bIndex, int eIndex) {
        return this.alpha_wX[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private double getAlpha_w1Xw2(Production prod, int cIndex, int bIndex, int eIndex) {
        return this.alpha_w1Xw2[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private void updateRuleProb(Production prod, int patternID, double count) {
        count *= this._weight;
        this._parameter.putPatternCount(prod.getGProduction(), patternID, count);
    }


    /****************************************************************/
    /****************** End Of Update Methods ***********************/
    /****************************************************************/

    /**
     * Reset the Outside Probability Tables
     */
    private void resetAlpha() {

        this._alpha = new double[this._N][][]; //bIndex, length, prod

        for (int i = 0; i < this._alpha.length; i++) {
            this._alpha[i] = new double[this._N - i + 1][];
            for (int j = 1; j < this._alpha[i].length; j++) {
                this._alpha[i][j] = new double[this._productions.size()];
            }
        }

        this.alpha_w = new double[this._productions.size()][][];
        this.alpha_Xw = new double[this._productions.size()][][][];
        this.alpha_wX = new double[this._productions.size()][][][];
        this.alpha_w1Xw2 = new double[this._productions.size()][][][];

        for (int prod_seq = 0; prod_seq < this._productions.size(); prod_seq++) {
            Production prod = (Production)this._productions.get(prod_seq);
            int childrensize = prod.countChildren();

            this.alpha_w[prod_seq] = new double[this._N][];
            this.alpha_Xw[prod_seq] = new double[childrensize][this._N][];
            this.alpha_wX[prod_seq] = new double[childrensize][this._N][];
            this.alpha_w1Xw2[prod_seq] = new double[childrensize][this._N][];

            for (int bIndex = 0; bIndex <= this._N - 1; bIndex++) {
                this.alpha_w[prod_seq][bIndex] = new double[this._N - bIndex + 1];
                for (int cIndex = 0; cIndex < childrensize; cIndex++) {
                    this.alpha_Xw[prod_seq][cIndex][bIndex] = new double[this._N - bIndex + 1];
                    this.alpha_wX[prod_seq][cIndex][bIndex] = new double[this._N - bIndex + 1];
                    this.alpha_w1Xw2[prod_seq][cIndex][bIndex] = new double[this._N - bIndex + 1];
                }
            }
        }

    }

    /**
     * Reset the Inside Probability Tables
     */
    private void resetBeta() {

        this._beta = new double[this._N][][]; //bIndex, length, prod
        for (int i = 0; i < this._beta.length; i++) {
            this._beta[i] = new double[this._N - i + 1][];
            for (int j = 1; j < this._beta[i].length; j++) {
                this._beta[i][j] = new double[this._productions.size()];
            }
        }

        this.beta_w = new double[this._productions.size()][][];
        this.beta_X = new double[this._productions.size()][][][];
        this.beta_Xw = new double[this._productions.size()][][][];
        this.beta_wX = new double[this._productions.size()][][][];
        this.beta_w1Xw2 = new double[this._productions.size()][][][];

        for (int prod_seq = this._productions.size() - 1; prod_seq >= 0; prod_seq--) {
            Production prod = this._productions.get(prod_seq);
            int childrensize = prod.countChildren();

            this.beta_w[prod_seq] = new double[this._N][];
            this.beta_X[prod_seq] = new double[childrensize][this._N][];
            this.beta_Xw[prod_seq] = new double[childrensize][this._N][];
            this.beta_wX[prod_seq] = new double[childrensize][this._N][];
            this.beta_w1Xw2[prod_seq] = new double[childrensize][this._N][];

            for (int bIndex = 0; bIndex <= this._N - 1; bIndex++) {
                this.beta_w[prod_seq][bIndex] = new double[this._N - bIndex + 1];
                for (int cIndex = 0; cIndex < childrensize; cIndex++) {
                    this.beta_X[prod_seq][cIndex][bIndex] = new double[this._N - bIndex + 1];
                    this.beta_Xw[prod_seq][cIndex][bIndex] = new double[this._N - bIndex + 1];
                    this.beta_wX[prod_seq][cIndex][bIndex] = new double[this._N - bIndex + 1];
                    this.beta_w1Xw2[prod_seq][cIndex][bIndex] = new double[this._N - bIndex + 1];
                }
            }
        }
    }

    /****************************************************************/
    /******************* End Of Reset Methods ***********************/
    /****************************************************************/

    /**
     * Computing the outside Probability
     */
    public void computeAlpha() {

        Production root = this._productions.get(0);
        double global_beta = getBeta(root, 0, this._N);

        if (global_beta == 0) {
            System.err.println("[Instance Ignored...] Beta Probability for this instance is zero! \n\t" + this._sentence);
            _ignoreFlag = true;
            return;
        }

        this.resetAlpha();

        setAlpha(root, 0, this._N, 1.0);
        for (int k = 1; k < this._N; k++) {
            setAlpha(root, 0, k, 0.0);
        }

        for (int prod_seq = 0; prod_seq < this._productions.size(); prod_seq++) {

            Production prod = this._productions.get(prod_seq);

            ArrayList<Production> children = prod.getChildren();

            int childrensize = children.size();

            for (int len = this._N; len >= 1; len--) {
                for (int bIndex = 0; bIndex <= this._N - len; bIndex++) {
                    int eIndex = bIndex + len;

                    //for the contiguous case..
                    if (childrensize == 0) {
                        double subProb;
                        double count;
                        Word bWord = this.getWord(bIndex);
                        Word eWord = this.getWord(eIndex - 1);

                        subProb = this.getAlpha(prod, bIndex, eIndex) * this.computeRuleTypeProb(prod, Rule.RULE_w) *
                                  this.computeOneWordProb(prod, Word.BEGIN, bWord) * this.computeOneWordProb(prod, eWord, Word.END);

                        this.putAlpha_w(prod, bIndex, eIndex, subProb);

                        //update the rule probabilities..
                        {
                            double ruleCount = Math.log(this.getAlpha(prod, bIndex, eIndex)) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_w)) +
                                               Math.log(this.computeOneWordProb(prod, Word.BEGIN, bWord)) +
                                               Math.log(this.getBeta_w(prod, bIndex, eIndex)) +
                                               Math.log(this.computeOneWordProb(prod, eWord, Word.END)) - Math.log(global_beta);
                            ruleCount = Math.exp(ruleCount);

                            this.updateRuleProb(prod, Rule.RULE_w, ruleCount);
                        }

                        if (len == 1) {
                            //update...
                            count = Math.log(this.getAlpha_w(prod, bIndex, eIndex)) + Math.log(this.getBeta_w(prod, bIndex, eIndex)) -
                                    Math.log(global_beta);
                            count = Math.exp(count);

                            this.updateWordParam(prod, Word.BEGIN, eWord, count);
                        } else {
                            Word eWord1 = this.getWord(eIndex - 2);
                            subProb = this.getAlpha_w(prod, bIndex, eIndex) * this.computeOneWordProb(prod, eWord1, eWord);

                            this.putAlpha_w(prod, bIndex, eIndex - 1, subProb);

                            //update...
                            count = Math.log(this.getAlpha_w(prod, bIndex, eIndex)) + Math.log(this.getBeta_w(prod, bIndex, eIndex)) -
                                    Math.log(global_beta);
                            count = Math.exp(count);

                            this.updateWordParam(prod, eWord1, eWord, count);
                        }

                        //for the END
                        {
                            count = Math.log(this.getAlpha(prod, bIndex, eIndex)) + Math.log(this.getBeta(prod, bIndex, eIndex)) -
                                    Math.log(global_beta);
                            count = Math.exp(count);

                            this.updateWordParam(prod, eWord, Word.END, count);
                        }

                    } else if (childrensize == 1) {

                        int cIndex = 0;
                        Production child = children.get(cIndex);

                        Word cWord = Word.CHILD_A;

                        double subProb;
                        double tCount;

                        subProb = this.getAlpha(prod, bIndex, eIndex) * this.computeRuleTypeProb(prod, Rule.RULE_w1Xw2) *
                                  this.computeOneWordProb(prod, Word.BEGIN, this.getWord(bIndex)) *
                                  this.computeOneWordProb(prod, this.getWord(eIndex - 1), Word.END);
                        this.putAlpha_w1Xw2(prod, cIndex, bIndex, eIndex, subProb);

                        //update...
                        tCount = Math.log(subProb) + Math.log(this.getBeta_w1Xw2(prod, cIndex, bIndex, eIndex)) - Math.log(global_beta);
                        tCount = Math.exp(tCount);
                        this.updateWordParam(prod, Word.BEGIN, this.getWord(bIndex), tCount);
                        this.updateWordParam(prod, this.getWord(eIndex - 1), Word.END, tCount);

                        subProb = this.getAlpha(prod, bIndex, eIndex) * this.computeRuleTypeProb(prod, Rule.RULE_wX) *
                                  this.computeOneWordProb(prod, Word.BEGIN, this.getWord(bIndex)) * this.computeOneWordProb(prod, cWord, Word.END);
                        this.putAlpha_wX(prod, cIndex, bIndex, eIndex, subProb);

                        //update...
                        tCount = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex, eIndex)) - Math.log(global_beta);
                        tCount = Math.exp(tCount);
                        this.updateWordParam(prod, Word.BEGIN, this.getWord(bIndex), tCount);
                        this.updateWordParam(prod, cWord, Word.END, tCount);

                        subProb = this.getAlpha(prod, bIndex, eIndex) * this.computeRuleTypeProb(prod, Rule.RULE_Xw) *
                                  this.computeOneWordProb(prod, Word.BEGIN, cWord) * this.computeOneWordProb(prod, this.getWord(eIndex - 1), Word.END);
                        this.putAlpha_Xw(prod, cIndex, bIndex, eIndex, subProb);

                        //update...
                        tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, bIndex, eIndex)) - Math.log(global_beta);
                        tCount = Math.exp(tCount);
                        this.updateWordParam(prod, Word.BEGIN, cWord, tCount);
                        this.updateWordParam(prod, this.getWord(eIndex - 1), Word.END, tCount);

                        //update rule pattern probabilities...
                        {
                            Word bWord = this.getWord(bIndex);
                            Word eWord = this.getWord(eIndex - 1);

                            double ruleCount;

                            ruleCount = Math.log(this.getAlpha(prod, bIndex, eIndex)) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_w1Xw2)) +
                                        Math.log(this.computeOneWordProb(prod, Word.BEGIN, bWord)) +
                                        Math.log(this.getBeta_w1Xw2(prod, cIndex, bIndex, eIndex)) +
                                        Math.log(this.computeOneWordProb(prod, eWord, Word.END)) - Math.log(global_beta);
                            ruleCount = Math.exp(ruleCount);

                            this.updateRuleProb(prod, Rule.RULE_w1Xw2, ruleCount);

                            ruleCount = Math.log(this.getAlpha(prod, bIndex, eIndex)) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_wX)) +
                                        Math.log(this.computeOneWordProb(prod, Word.BEGIN, bWord)) +
                                        Math.log(this.getBeta_wX(prod, cIndex, bIndex, eIndex)) +
                                        Math.log(this.computeOneWordProb(prod, cWord, Word.END)) - Math.log(global_beta);
                            ruleCount = Math.exp(ruleCount);

                            this.updateRuleProb(prod, Rule.RULE_wX, ruleCount);

                            ruleCount = Math.log(this.getAlpha(prod, bIndex, eIndex)) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_Xw)) +
                                        Math.log(this.computeOneWordProb(prod, Word.BEGIN, cWord)) +
                                        Math.log(this.getBeta_Xw(prod, cIndex, bIndex, eIndex)) +
                                        Math.log(this.computeOneWordProb(prod, eWord, Word.END)) - Math.log(global_beta);
                            ruleCount = Math.exp(ruleCount);

                            this.updateRuleProb(prod, Rule.RULE_Xw, ruleCount);
                        }

                        /**...pass to its child...**/
                        {
                            subProb = this.getAlpha(prod, bIndex, eIndex) * this.computeRuleTypeProb(prod, Rule.RULE_X) *
                                      this.computeOneWordProb(prod, Word.BEGIN, cWord) * this.computeOneWordProb(prod, cWord, Word.END);
                            this.addAlpha(child, bIndex, eIndex, subProb);

                            //update rule pattern probabilities...
                            double ruleCount = Math.log(subProb) + Math.log(this.getBeta(child, bIndex, eIndex)) - Math.log(global_beta);
                            ruleCount = Math.exp(ruleCount);

                            this.updateRuleProb(prod, Rule.RULE_X, ruleCount);

                            //update word probabilities...
                            double count = ruleCount;
                            this.updateWordParam(prod, Word.BEGIN, cWord, count);
                            this.updateWordParam(prod, cWord, Word.END, count);
                        }

                        /**...enclosed by w1Xw2...**/
                        if (len >= 3) {
                            Word eWord = this.getWord(eIndex - 1);

                            if (len >= 4) {
                                Word eWord1 = this.getWord(eIndex - 2);
                                subProb = this.getAlpha_w1Xw2(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, eWord1, eWord);
                                this.putAlpha_w1Xw2(prod, cIndex, bIndex, eIndex - 1, subProb);

                                //update:
                                double count = Math.log(subProb) + Math.log(this.getBeta_w1Xw2(prod, cIndex, bIndex, eIndex - 1)) -
                                               Math.log(global_beta);
                                count = Math.exp(count);

                                this.updateWordParam(prod, eWord1, eWord, count);
                            }

                            subProb = this.getAlpha_w1Xw2(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, cWord, eWord);
                            this.putAlpha_wX(prod, cIndex, bIndex, eIndex - 1, subProb);

                            //update:
                            double count = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex, eIndex - 1)) - Math.log(global_beta);
                            count = Math.exp(count);

                            this.updateWordParam(prod, cWord, eWord, count);
                        }

                        /**enclosed by Xw**/
                        if (len >= 2) {
                            Word eWord = this.getWord(eIndex - 1);

                            if (len >= 3) {
                                Word eWord1 = this.getWord(eIndex - 2);
                                subProb = this.getAlpha_Xw(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, eWord1, eWord);
                                this.putAlpha_Xw(prod, cIndex, bIndex, eIndex - 1, subProb);

                                //update:
                                double count = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, bIndex, eIndex - 1)) - Math.log(global_beta);
                                count = Math.exp(count);

                                this.updateWordParam(prod, eWord1, eWord, count);
                            }

                            subProb = this.getAlpha_Xw(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, cWord, eWord);

                            //propagate
                            double subAlpha = subProb;
                            this.addAlpha(child, bIndex, eIndex - 1, subAlpha);

                            //update:
                            double count = Math.log(subProb) + Math.log(this.getBeta_X(prod, cIndex, bIndex, eIndex - 1)) - Math.log(global_beta);
                            count = Math.exp(count);

                            this.updateWordParam(prod, cWord, eWord, count);
                        }

                        /**enclosed by wX**/
                        if (len >= 2) {
                            Word bWord = this.getWord(bIndex);

                            if (len >= 3) {
                                Word bWord_1 = this.getWord(bIndex + 1);
                                subProb = this.getAlpha_wX(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, bWord, bWord_1);
                                this.putAlpha_wX(prod, cIndex, bIndex + 1, eIndex, subProb);

                                //update:
                                double count = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex + 1, eIndex)) - Math.log(global_beta);
                                count = Math.exp(count);

                                this.updateWordParam(prod, bWord, bWord_1, count);
                            }

                            subProb = this.getAlpha_wX(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, bWord, cWord);

                            //propagate
                            double subAlpha = subProb;
                            this.addAlpha(child, bIndex + 1, eIndex, subAlpha);

                            //update:
                            double count = Math.log(subProb) + Math.log(this.getBeta_X(prod, cIndex, bIndex + 1, eIndex)) - Math.log(global_beta);
                            count = Math.exp(count);

                            this.updateWordParam(prod, bWord, cWord, count);
                        }

                    } else if (childrensize == 2) {

                        double prob = this.getAlpha(prod, bIndex, eIndex);
                        double subProb;

                        //update rule probabilities...
                        {

                            //CASE OF FIVE PARTS...
                            if (len >= 5) {
                                // w1 X w2 | Y w3
                                // w1 Y w2 | X w3
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    subProb = 0;
                                    for (int sIndex = bIndex + 3; sIndex <= eIndex - 2; sIndex++) {
                                        subProb += this.getBeta_w1Xw2(prod, cIndex, bIndex, sIndex) *
                                                this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex)) *
                                                this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex);
                                    }
                                    subProb *= this.computeOneWordProb(prod, Word.BEGIN, this.getWord(bIndex)) *
                                            this.computeOneWordProb(prod, this.getWord(eIndex - 1), Word.END);

                                    if (cIndex == 0) {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_w1Xw2Yw3)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_w1Xw2Yw3, ruleCount);
                                    } else {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_w1Yw2Xw3)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_w1Yw2Xw3, ruleCount);
                                    }
                                }
                            }

                            //CASE OF FOUR PARTS..
                            if (len >= 4) {
                                // w1 X | w2 Y
                                // w1 Y | w2 X
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    subProb = 0;
                                    for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                                        subProb += this.getBeta_wX(prod, cIndex, bIndex, sIndex) *
                                                this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getWord(sIndex)) *
                                                this.getBeta_wX(prod, 1 - cIndex, sIndex, eIndex);
                                    }
                                    subProb *= this.computeOneWordProb(prod, Word.BEGIN, this.getWord(bIndex)) *
                                            this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);

                                    if (cIndex == 0) {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_w1Xw2Y)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_w1Xw2Y, ruleCount);
                                    } else {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_w1Yw2X)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_w1Yw2X, ruleCount);
                                    }
                                }

                                // X w1 | Y w2
                                // Y w1 | X w2
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    subProb = 0;
                                    for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                                        subProb += this.getBeta_Xw(prod, cIndex, bIndex, sIndex) *
                                                this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex)) *
                                                this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex);
                                    }
                                    subProb *= this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                            this.computeOneWordProb(prod, this.getWord(eIndex - 1), Word.END);

                                    if (cIndex == 0) {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_Xw1Yw2)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        //System.err.println("Xw1Yw2="+ruleCount);
                                        this.updateRuleProb(prod, Rule.RULE_Xw1Yw2, ruleCount);
                                    } else {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_Yw1Xw2)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        //System.err.println("\tYw1Xw2="+ruleCount);
                                        this.updateRuleProb(prod, Rule.RULE_Yw1Xw2, ruleCount);
                                    }
                                }

                                // w1 X | Y w2
                                // w1 Y | X w2
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    subProb = 0;
                                    for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                                        subProb += this.getBeta_wX(prod, cIndex, bIndex, sIndex) *
                                                this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                                this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex);
                                    }
                                    subProb *= this.computeOneWordProb(prod, Word.BEGIN, this.getWord(bIndex)) *
                                            this.computeOneWordProb(prod, this.getWord(eIndex - 1), Word.END);

                                    if (cIndex == 0) {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_w1XYw2)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_w1XYw2, ruleCount);
                                    } else {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_w1YXw2)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_w1YXw2, ruleCount);
                                    }
                                }
                            }

                            //CASE OF THREE PARTS.
                            if (len >= 3) {
                                // w X | Y
                                // w Y | X
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    subProb = 0;
                                    for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                                        subProb += this.getBeta_wX(prod, cIndex, bIndex, sIndex) *
                                                this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                                this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex);
                                    }
                                    subProb *= this.computeOneWordProb(prod, Word.BEGIN, this.getWord(bIndex)) *
                                            this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);

                                    if (cIndex == 0) {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_wXY)) +
                                                Math.log(subProb) + -Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_wXY, ruleCount);
                                    } else {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_wYX)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_wYX, ruleCount);
                                    }
                                }

                                // X w | Y
                                // Y w | X
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    subProb = 0;
                                    for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                                        subProb += this.getBeta_Xw(prod, cIndex, bIndex, sIndex) *
                                                this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex)) *
                                                this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex);
                                    }
                                    subProb *= this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                            this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);

                                    if (cIndex == 0) {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_XwY)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_XwY, ruleCount);
                                    } else {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_YwX)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_YwX, ruleCount);
                                    }
                                }

                                // X | Y w
                                // Y | X w
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    subProb = 0;
                                    for (int sIndex = bIndex + 1; sIndex <= eIndex - 2; sIndex++) {
                                        subProb += this.getBeta_X(prod, cIndex, bIndex, sIndex) *
                                                this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                                this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex);
                                    }
                                    subProb *= this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                            this.computeOneWordProb(prod, this.getWord(eIndex - 1), Word.END);

                                    if (cIndex == 0) {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_XYw)) +
                                                Math.log(subProb) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_XYw, ruleCount);
                                    } else {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_YXw)) +
                                                Math.log(subProb) + -Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_YXw, ruleCount);
                                    }
                                }
                            }

                            //CASE OF TWO PARTS.
                            if (len >= 2) {
                                // X | Y
                                // Y | X
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    subProb = 0;
                                    for (int sIndex = bIndex + 1; sIndex <= eIndex - 1; sIndex++) {
                                        subProb += this.getBeta_X(prod, cIndex, bIndex, sIndex) *
                                                this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                                this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex);
                                    }
                                    subProb *= this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                            this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);

                                    if (cIndex == 0) {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_XY)) + Math.log(subProb) -
                                                Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_XY, ruleCount);
                                    } else {
                                        double ruleCount = Math.log(prob) + Math.log(this.computeRuleTypeProb(prod, Rule.RULE_YX)) + Math.log(subProb) -
                                                Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_YX, ruleCount);
                                    }
                                }
                            }
                        }

                        //update the word probabilities...
                        //stage 1: decomposition phase
                        //CASE OF FIVE PARTS.
                        if (len >= 5) {
                            double tCount;
                            double tProb = prob * this.computeOneWordProb(prod, Word.BEGIN, this.getWord(bIndex)) *
                                           this.computeOneWordProb(prod, this.getWord(eIndex - 1), Word.END);

                            //case : w1 X w2 | Y w3 / w1 Y w2 | X w3
                            for (int sIndex = bIndex + 3; sIndex <= eIndex - 2; sIndex++) {
                                int cIndex;

                                cIndex = 0;

                                // w1 Y w2 | * X w3 *
                                subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1Yw2Xw3) *
                                          this.getBeta_w1Xw2(prod, 1 - cIndex, bIndex, sIndex) *
                                          this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex));
                                this.putAlpha_Xw(prod, cIndex, sIndex, eIndex, subProb);

                                tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, sIndex, eIndex)) - Math.log(global_beta);
                                tCount = Math.exp(tCount);

                                this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex), tCount);
                                this.updateWordParam(prod, this.getWord(eIndex - 1), Word.END, tCount);

                                // * w1 X w2 * | Y w3
                                subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1Xw2Yw3) *
                                          this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) *
                                          this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex));
                                this.putAlpha_w1Xw2(prod, cIndex, bIndex, sIndex, subProb);

                                tCount = Math.log(subProb) + Math.log(this.getBeta_w1Xw2(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                tCount = Math.exp(tCount);

                                this.updateWordParam(prod, Word.BEGIN, this.getWord(bIndex), tCount);
                                this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex), tCount);

                                cIndex = 1;

                                // w1 X w2 | * Y w3 *
                                subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1Xw2Yw3) *
                                          this.getBeta_w1Xw2(prod, 1 - cIndex, bIndex, sIndex) *
                                          this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex));

                                this.putAlpha_Xw(prod, cIndex, sIndex, eIndex, subProb);

                                tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, sIndex, eIndex)) - Math.log(global_beta);
                                tCount = Math.exp(tCount);

//                                this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex), tCount);
                                this.updateWordParam(prod, this.getWord(eIndex - 1), Word.END, tCount);

                                // * w1 Y w2 * | X w3
                                subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1Yw2Xw3) *
                                          this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) *
                                          this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex));
                                this.putAlpha_w1Xw2(prod, cIndex, bIndex, sIndex, subProb);

                                tCount = Math.log(subProb) + Math.log(this.getBeta_w1Xw2(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                tCount = Math.exp(tCount);

                                this.updateWordParam(prod, Word.BEGIN, this.getWord(bIndex), tCount);
//                                this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex), tCount);
                            }
                        }

                        //CASE OF FOUR PARTS.
                        if (len >= 4) {
                            double tCount;
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                                //case : w1 X | w2 Y / w1 Y | w2 X
                                {
                                    double tProb = prob * this.computeOneWordProb(prod, Word.BEGIN, this.getWord(bIndex));

                                    int cIndex;

                                    cIndex = 0;
                                    // w1 Y | * w2 X *
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1Yw2X) *
                                              this.getBeta_wX(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), this.getWord(sIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), Word.END);
                                    this.putAlpha_wX(prod, cIndex, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, this.getChildWord(1 - cIndex), this.getWord(sIndex), tCount);
                                    this.updateWordParam(prod, this.getChildWord(cIndex), Word.END, tCount);

                                    // * w1 X * | w2 Y
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1Xw2Y) *
                                              this.getBeta_wX(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getWord(sIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);
                                    this.putAlpha_wX(prod, cIndex, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getWord(bIndex), tCount);
                                    this.updateWordParam(prod, this.getChildWord(cIndex), this.getWord(sIndex), tCount);

                                    cIndex = 1;
                                    // w1 X | * w2 Y *
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1Xw2Y) *
                                              this.getBeta_wX(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), this.getWord(sIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), Word.END);
                                    this.putAlpha_wX(prod, cIndex, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

//                                    this.updateWordParam(prod, this.getChildWord(1 - cIndex), this.getWord(sIndex), tCount);
                                    this.updateWordParam(prod, this.getChildWord(cIndex), Word.END, tCount);

                                    // * w1 Y * | w2 X
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1Yw2X) *
                                              this.getBeta_wX(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getWord(sIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);
                                    this.putAlpha_wX(prod, cIndex, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getWord(bIndex), tCount);
//                                    this.updateWordParam(prod, this.getChildWord(cIndex), this.getWord(sIndex), tCount);
                                }

                                //case : X w1 | Y w2 / Y w1 | X w2
                                {
                                    double tProb = prob * this.computeOneWordProb(prod, this.getWord(eIndex - 1), Word.END);

                                    int cIndex;

                                    cIndex = 0;

                                    // * X w1 * | Y w2
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_Xw1Yw2) *
                                              this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                              this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex));
                                    this.putAlpha_Xw(prod, cIndex, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex), tCount);

                                    // Y w1 | * X w2 *
                                    subProb = tProb * this.getBeta_Xw(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeRuleTypeProb(prod, Rule.RULE_Yw1Xw2) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(1 - cIndex)) *
                                              this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex));
                                    this.putAlpha_Xw(prod, cIndex, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getWord(eIndex - 1), Word.END, tCount);

                                    cIndex = 1;

                                    // * Y w1 * | X w2
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_Yw1Xw2) *
                                              this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                              this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex));
                                    this.putAlpha_Xw(prod, cIndex, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getChildWord(cIndex), tCount);
//                                    this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex), tCount);

                                    // X w1 | * Y w2 *
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_Xw1Yw2) *
                                              this.getBeta_Xw(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(1 - cIndex)) *
                                              this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex));
                                    this.putAlpha_Xw(prod, cIndex, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

//                                    this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getWord(eIndex - 1), Word.END, tCount);
                                }

                                //case : w1 X | Y w2 / w1 Y | X w2
                                {
                                    double tProb = prob * this.computeOneWordProb(prod, Word.BEGIN, this.getWord(bIndex)) *
                                            this.computeOneWordProb(prod, this.getWord(eIndex - 1), Word.END);

                                    int cIndex;

                                    cIndex = 0;
                                    // * w1 X * | Y w2
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1XYw2) *
                                              this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex));
                                    this.putAlpha_wX(prod, cIndex, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getWord(bIndex), tCount);
                                    this.updateWordParam(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);

                                    // w1 Y | * X w2 *
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1YXw2) *
                                              this.getBeta_wX(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex));
                                    this.putAlpha_Xw(prod, cIndex, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getWord(eIndex - 1), Word.END, tCount);

                                    cIndex = 1;
                                    // * w1 Y * | X w2
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1YXw2) *
                                              this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex));
                                    this.putAlpha_wX(prod, cIndex, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getWord(bIndex), tCount);
//                                    this.updateWordParam(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);

                                    // w1 X | * Y w2 *
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_w1XYw2) *
                                              this.getBeta_wX(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex));
                                    this.putAlpha_Xw(prod, cIndex, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

//                                    this.updateWordParam(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getWord(eIndex - 1), Word.END, tCount);
                                }
                            }
                        }

                        //CASE OF THREE PARTS.
                        if (len >= 3) {
                            double tCount;
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                                //case : w X | Y / w Y | X
                                {
                                    double tProb = prob * this.computeOneWordProb(prod, Word.BEGIN, this.getWord(bIndex));

                                    int cIndex;
                                    Production child;

                                    cIndex = 0;
                                    child = children.get(cIndex);

                                    // * w X * | Y
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_wXY) * this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);

                                    this.putAlpha_wX(prod, cIndex, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getWord(bIndex), tCount);
                                    this.updateWordParam(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);

                                    // w Y | * X *
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_wYX) *
                                              this.getBeta_wX(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), Word.END);

                                    //propagate
                                    this.addAlpha(child, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta(child, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getChildWord(cIndex), Word.END, tCount);

                                    cIndex = 1;
                                    child = children.get(cIndex);

                                    // * w Y * | X
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_wYX) * this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);

                                    this.putAlpha_wX(prod, cIndex, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getWord(bIndex), tCount);
//                                    this.updateWordParam(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);

                                    // w X | * Y *
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_wXY) *
                                              this.getBeta_wX(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), Word.END);

                                    //propagate
                                    this.addAlpha(child, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta(child, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

//                                    this.updateWordParam(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getChildWord(cIndex), Word.END, tCount);
                                }

                                //case : X w | Y / Y w | X
                                {
                                    int cIndex;
                                    Production child;

                                    cIndex = 0;
                                    child = children.get(cIndex);

                                    // * X w * | Y
                                    subProb = prob * this.computeRuleTypeProb(prod, Rule.RULE_XwY) * this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                              this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);
                                    this.putAlpha_Xw(prod, cIndex, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex), tCount);

                                    // Y w | * X *
                                    subProb = prob * this.computeRuleTypeProb(prod, Rule.RULE_YwX) * this.getBeta_Xw(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(1 - cIndex)) *
                                              this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), Word.END);

                                    //propagate
                                    this.addAlpha(child, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta(child, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getChildWord(cIndex), Word.END, tCount);

                                    cIndex = 1;
                                    child = children.get(cIndex);

                                    // * Y w * | X
                                    subProb = prob * this.computeRuleTypeProb(prod, Rule.RULE_YwX) * this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                              this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);
                                    this.putAlpha_Xw(prod, cIndex, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getChildWord(cIndex), tCount);
//                                    this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(1 - cIndex), tCount);

                                    // X w | * Y *
                                    subProb = prob * this.computeRuleTypeProb(prod, Rule.RULE_XwY) * this.getBeta_Xw(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(1 - cIndex)) *
                                              this.computeOneWordProb(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), Word.END);

                                    //propagate
                                    this.addAlpha(child, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta(child, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

//                                    this.updateWordParam(prod, this.getWord(sIndex - 1), this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getChildWord(cIndex), Word.END, tCount);
                                }
                            }

                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 2; sIndex++) {
                                //case : X | Y w / Y | X w
                                {
                                    double tProb = prob * this.computeOneWordProb(prod, this.getWord(eIndex - 1), Word.END);

                                    int cIndex;
                                    Production child;

                                    cIndex = 0;
                                    child = children.get(cIndex);

                                    // * X * | Y w
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_XYw) *
                                              this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex));

                                    //propagate
                                    this.addAlpha(child, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta(child, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);

                                    // Y | * X w *
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_YXw) * this.getBeta_X(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(1 - cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex));

                                    this.putAlpha_Xw(prod, cIndex, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getWord(eIndex - 1), Word.END, tCount);

                                    cIndex = 1;
                                    child = children.get(cIndex);

                                    // * Y * | X w
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_YXw) *
                                              this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex));

                                    //propagate
                                    this.addAlpha(child, bIndex, sIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta(child, bIndex, sIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

                                    this.updateWordParam(prod, Word.BEGIN, this.getChildWord(cIndex), tCount);
//                                    this.updateWordParam(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);

                                    // X | * Y w *
                                    subProb = tProb * this.computeRuleTypeProb(prod, Rule.RULE_XYw) * this.getBeta_X(prod, 1 - cIndex, bIndex, sIndex) *
                                              this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(1 - cIndex)) *
                                              this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex));

                                    this.putAlpha_Xw(prod, cIndex, sIndex, eIndex, subProb);

                                    tCount = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, sIndex, eIndex)) - Math.log(global_beta);
                                    tCount = Math.exp(tCount);

//                                    this.updateWordParam(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                                    this.updateWordParam(prod, this.getWord(eIndex - 1), Word.END, tCount);
                                }
                            }
                        }

                        //CASE OF TWO PARTS..
                        //case : X | Y / Y | X
                        if (len >= 2) {
                            double tCount;
                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 1; sIndex++) {
                                int cIndex;
                                Production child;

                                cIndex = 0;
                                child = children.get(cIndex);

                                // * X * | Y
                                subProb = prob * this.computeRuleTypeProb(prod, Rule.RULE_XY) * this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex) *
                                          this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                          this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                          this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);

                                //propagate
                                this.addAlpha(child, bIndex, sIndex, subProb);

                                tCount = Math.log(subProb) + Math.log(this.getBeta(child, bIndex, sIndex)) - Math.log(global_beta);
                                tCount = Math.exp(tCount);

                                this.updateWordParam(prod, Word.BEGIN, this.getChildWord(cIndex), tCount);
                                this.updateWordParam(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);

                                // Y | * X *
                                subProb = prob * this.computeRuleTypeProb(prod, Rule.RULE_YX) * this.getBeta_X(prod, 1 - cIndex, bIndex, sIndex) *
                                          this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(1 - cIndex)) *
                                          this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex)) *
                                          this.computeOneWordProb(prod, this.getChildWord(cIndex), Word.END);

                                //propagate
                                this.addAlpha(child, sIndex, eIndex, subProb);

                                tCount = Math.log(subProb) + Math.log(this.getBeta(child, sIndex, eIndex)) - Math.log(global_beta);
                                tCount = Math.exp(tCount);

                                this.updateWordParam(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                                this.updateWordParam(prod, this.getChildWord(cIndex), Word.END, tCount);

                                cIndex = 1;
                                child = children.get(cIndex);

                                // * Y * | X
                                subProb = prob * this.computeRuleTypeProb(prod, Rule.RULE_YX) * this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex) *
                                          this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(cIndex)) *
                                          this.computeOneWordProb(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex)) *
                                          this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), Word.END);

                                //propagate
                                this.addAlpha(child, bIndex, sIndex, subProb);

                                tCount = Math.log(subProb) + Math.log(this.getBeta(child, bIndex, sIndex)) - Math.log(global_beta);
                                tCount = Math.exp(tCount);

                                this.updateWordParam(prod, Word.BEGIN, this.getChildWord(cIndex), tCount);
//                                this.updateWordParam(prod, this.getChildWord(cIndex), this.getChildWord(1 - cIndex), tCount);

                                // X | * Y *
                                subProb = prob * this.computeRuleTypeProb(prod, Rule.RULE_XY) * this.getBeta_X(prod, 1 - cIndex, bIndex, sIndex) *
                                          this.computeOneWordProb(prod, Word.BEGIN, this.getChildWord(1 - cIndex)) *
                                          this.computeOneWordProb(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex)) *
                                          this.computeOneWordProb(prod, this.getChildWord(cIndex), Word.END);

                                //propagate
                                this.addAlpha(child, sIndex, eIndex, subProb);

                                tCount = Math.log(subProb) + Math.log(this.getBeta(child, sIndex, eIndex)) - Math.log(global_beta);
                                tCount = Math.exp(tCount);

//                                this.updateWordParam(prod, this.getChildWord(1 - cIndex), this.getChildWord(cIndex), tCount);
                                this.updateWordParam(prod, this.getChildWord(cIndex), Word.END, tCount);
                            }
                        }

                        /**POST PROCESSING BEGINS**/
                        {
                            double count;

                            for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                Production child = children.get(cIndex);
                                Word cWord = this.getChildWord(cIndex);
                                // case : w1 X w2
                                if (len >= 3) {
                                    Word eWord = this.getWord(eIndex - 1);

                                    // w1 X w2 | w
                                    if (len >= 4) {
                                        Word eWord1 = this.getWord(eIndex - 2);
                                        subProb = this.getAlpha_w1Xw2(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, eWord1, eWord);
                                        this.putAlpha_w1Xw2(prod, cIndex, bIndex, eIndex - 1, subProb);

                                        //update...
                                        count = Math.log(subProb) + Math.log(this.getBeta_w1Xw2(prod, cIndex, bIndex, eIndex - 1)) -
                                                Math.log(global_beta);
                                        count = Math.exp(count);

                                        this.updateWordParam(prod, eWord1, eWord, count);
                                    }

                                    // w1 X | w
                                    subProb = this.getAlpha_w1Xw2(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, cWord, eWord);
                                    this.putAlpha_wX(prod, cIndex, bIndex, eIndex - 1, subProb);

                                    //update...
                                    count = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex, eIndex - 1)) - Math.log(global_beta);
                                    count = Math.exp(count);

                                    this.updateWordParam(prod, cWord, eWord, count);
                                }

                                // case : X w
                                if (len >= 2) {
                                    Word eWord = this.getWord(eIndex - 1);

                                    // X w | w
                                    if (len >= 3) {
                                        Word eWord1 = this.getWord(eIndex - 2);
                                        subProb = this.getAlpha_Xw(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, eWord1, eWord);
                                        this.putAlpha_Xw(prod, cIndex, bIndex, eIndex - 1, subProb);

                                        //update...
                                        count = Math.log(subProb) + Math.log(this.getBeta_Xw(prod, cIndex, bIndex, eIndex - 1)) -
                                                Math.log(global_beta);
                                        count = Math.exp(count);

                                        this.updateWordParam(prod, eWord1, eWord, count);
                                    }

                                    // X | w
                                    subProb = this.getAlpha_Xw(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, cWord, eWord);

                                    //propagate
                                    double subAlpha = subProb;
                                    this.addAlpha(child, bIndex, eIndex - 1, subAlpha);

                                    //update...
                                    count = Math.log(subProb) + Math.log(this.getBeta_X(prod, cIndex, bIndex, eIndex - 1)) - Math.log(global_beta);
                                    count = Math.exp(count);

                                    this.updateWordParam(prod, cWord, eWord, count);
                                }

                                // case : w X
                                if (len >= 2) {
                                    Word bWord = this.getWord(bIndex);

                                    // w | w X
                                    if (len >= 3) {
                                        Word bWord_1 = this.getWord(bIndex + 1);
                                        subProb = this.getAlpha_wX(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, bWord, bWord_1);
                                        this.putAlpha_wX(prod, cIndex, bIndex + 1, eIndex, subProb);

                                        //update..
                                        count = Math.log(subProb) + Math.log(this.getBeta_wX(prod, cIndex, bIndex + 1, eIndex)) -
                                                Math.log(global_beta);
                                        count = Math.exp(count);

                                        this.updateWordParam(prod, bWord, bWord_1, count);
                                    }

                                    // w | X
                                    subProb = this.getAlpha_wX(prod, cIndex, bIndex, eIndex) * this.computeOneWordProb(prod, bWord, cWord);

                                    //propagete
                                    double subAlpha = subProb;
                                    this.addAlpha(child, bIndex + 1, eIndex, subAlpha);

                                    //update..
                                    count = Math.log(subProb) + Math.log(this.getBeta_X(prod, cIndex, bIndex + 1, eIndex)) - Math.log(global_beta);
                                    count = Math.exp(count);

                                    this.updateWordParam(prod, bWord, cWord, count);
                                }
                            }
                        }

                        /**POST PROCESSING ENDS**/

                    } else {
                        System.out.println("Invalid number of children:"+prod.toLine());
                        System.err.println("Currently only support at most two children.");
                        System.exit(1);
                    }
                }
            }
        }

        double oldSum = Double.NEGATIVE_INFINITY;
        for (int k = 0; k < this._productions.size(); k++) {
            Production prod = this._productions.get(k);
            double sum = 0;
            for (int i = 0; i <= this._N; i++) {
                for (int j = i + 1; j <= this._N; j++) {
                    sum += this.getAlpha(prod, i, j) * this.getBeta(prod, i, j);
                    if(_debugMode){
                        this.viewAlpha(prod, i, j);
                        this.viewBeta(prod, i, j);
                    }
                }
            }

            if (oldSum == Double.NEGATIVE_INFINITY) {
                oldSum = sum;
            } else {
                if (Math.abs(oldSum - sum) > Math.pow(10, -10) * sum) {
                    System.err.println("Sum=" + sum);
                    System.err.println(this._productions);
                    this.viewBeta(this._productions.get(0), 0, this._N);
                    System.exit(1);
                } else {
                }
            }
        }
//        System.exit(1);
    }

    /**
     * Computing the Inside Probability
     */
    public void computeBeta() {

        this.resetBeta();

        /** consider each production from bottom up... **/
        for (int prod_seq = this._productions.size() - 1; prod_seq >= 0; prod_seq--) {
            Production prod = this._productions.get(prod_seq);
            int childrensize = prod.countChildren();

            /** varying length with increasing order: 1 to N **/
            for (int len = 1; len <= this._N; len++) {

                /** the begin index of the span **/
                for (int bIndex = 0; bIndex <= this._N - len; bIndex++) {
                    int eIndex = bIndex + len;

                    /** compute the Beta_w **/
                    {
                        double wProb;
                        if (len == 1) {
                            wProb = 1.0;
                        } else {
                            Word word1 = this.getWord(eIndex - 2);
                            Word word = this.getWord(eIndex - 1);
                            wProb = this.getBeta_w(prod, bIndex, eIndex - 1) * this.computeOneWordProb(prod, word1, word);
                        }
                        this.putBeta_w(prod, bIndex, eIndex, wProb);
                    }

                    double prob = 0;
                    if (childrensize == 0) {
                        //case 1: w
                        Word bWord = this.getWord(bIndex);
                        Word eWord = this.getWord(eIndex - 1);
                        prob = this.computeRuleTypeProb(prod, Rule.RULE_w) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                               this.getBeta_w(prod, bIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);

                        this.setBeta(prod, bIndex, eIndex, prob);
                        if (_debugMode) {
                            viewBeta(prod, bIndex, eIndex);
                        }
                    } else if (childrensize == 1) {
                        int cIndex = 0;
                        Production child = prod.getChildren().get(cIndex);
                        double subProb;

                        Word cWord = Word.CHILD_A;

                        //case of ONE part
                        {
                            //case : X
                            subProb = this.getBeta(child, bIndex, eIndex);
                            this.putBeta_X(prod, cIndex, bIndex, eIndex, subProb);

                            Word bWord = Word.CHILD_A;
                            Word eWord = Word.CHILD_A;
                            prob += this.computeRuleTypeProb(prod, Rule.RULE_X) * this.computeOneWordProb(prod, Word.BEGIN, bWord) * subProb *
                                    this.computeOneWordProb(prod, eWord, Word.END);
                        }

                        //case of TWO parts
                        if (len >= 2) {
                            //case : w X
                            {
                                Word bWord = this.getWord(bIndex);
                                Word eWord = cWord;
                                subProb = this.computeOneWordProb(prod, bWord, eWord) * this.getBeta_X(prod, cIndex, bIndex + 1, eIndex);
                                if (len >= 3) {
                                    Word word_1 = this.getWord(bIndex + 1);
                                    subProb += this.computeOneWordProb(prod, bWord, word_1) * this.getBeta_wX(prod, cIndex, bIndex + 1, eIndex);
                                }
                                this.putBeta_wX(prod, cIndex, bIndex, eIndex, subProb);

                                prob += this.computeRuleTypeProb(prod, Rule.RULE_wX) * this.computeOneWordProb(prod, Word.BEGIN, bWord) * subProb *
                                        this.computeOneWordProb(prod, eWord, Word.END);
                            }

                            //case : X w
                            {
                                Word bWord = cWord;
                                Word eWord = this.getWord(eIndex - 1);
                                subProb = this.getBeta_X(prod, cIndex, bIndex, eIndex - 1) * this.computeOneWordProb(prod, bWord, eWord);
                                if (len >= 3) {
                                    Word word1 = this.getWord(eIndex - 2);
                                    subProb += this.getBeta_Xw(prod, cIndex, bIndex, eIndex - 1) * this.computeOneWordProb(prod, word1, eWord);
                                }
                                this.putBeta_Xw(prod, cIndex, bIndex, eIndex, subProb);

                                prob += this.computeRuleTypeProb(prod, Rule.RULE_Xw) * this.computeOneWordProb(prod, Word.BEGIN, bWord) * subProb *
                                        this.computeOneWordProb(prod, eWord, Word.END);
                            }
                        }

                        //case of THREE parts
                        if (len >= 3) {
                            //case : w1 X w2
                            Word bWord = this.getWord(bIndex);
                            Word eWord = this.getWord(eIndex - 1);

                            subProb = this.getBeta_wX(prod, cIndex, bIndex, eIndex - 1) * this.computeOneWordProb(prod, cWord, eWord);
                            if (len >= 4) {
                                Word word1 = this.getWord(eIndex - 2);
                                subProb += this.getBeta_w1Xw2(prod, cIndex, bIndex, eIndex - 1) * this.computeOneWordProb(prod, word1, eWord);
                            }
                            this.putBeta_w1Xw2(prod, cIndex, bIndex, eIndex, subProb);

                            prob += this.computeRuleTypeProb(prod, Rule.RULE_w1Xw2) * this.computeOneWordProb(prod, Word.BEGIN, bWord) * subProb *
                                    this.computeOneWordProb(prod, eWord, Word.END);
                        }

                        this.setBeta(prod, bIndex, eIndex, prob);
                        if (_debugMode) {
                            viewBeta(prod, bIndex, eIndex);
                        }

                    } else if (childrensize == 2) {
                        Production child;
                        double subProb;

                        /**
                         * BEGIN OF PREPROCESS
                         */
                        for (int cIndex = 0; cIndex <= 1; cIndex++) {

                            child = prod.getChildren().get(cIndex);
                            Word cWord;
                            if (cIndex == 0) {
                                cWord = Word.CHILD_A;
                            } else {
                                cWord = Word.CHILD_B;
                            }

                            //case of ONE part
                            {
                                //case 1-(a): X
                                subProb = this.getBeta(child, bIndex, eIndex);
                                this.putBeta_X(prod, cIndex, bIndex, eIndex, subProb);
                            }

                            //case of TWO parts
                            if (len >= 2) {
                                //case : w X
                                {
                                    Word bWord = this.getWord(bIndex);
                                    subProb = this.computeOneWordProb(prod, bWord, cWord) * this.getBeta_X(prod, cIndex, bIndex + 1, eIndex);
                                    if (len >= 3) {
                                        Word word_1 = this.getWord(bIndex + 1);
                                        subProb += this.computeOneWordProb(prod, bWord, word_1) * this.getBeta_wX(prod, cIndex, bIndex + 1, eIndex);
                                    }
                                    this.putBeta_wX(prod, cIndex, bIndex, eIndex, subProb);
                                }

                                //case : X w
                                {
                                    Word eWord = this.getWord(eIndex - 1);
                                    subProb = this.getBeta_X(prod, cIndex, bIndex, eIndex - 1) * this.computeOneWordProb(prod, cWord, eWord);
                                    if (len >= 3) {
                                        Word word1 = this.getWord(eIndex - 2);
                                        subProb += this.getBeta_Xw(prod, cIndex, bIndex, eIndex - 1) * this.computeOneWordProb(prod, word1, eWord);
                                    }
                                    this.putBeta_Xw(prod, cIndex, bIndex, eIndex, subProb);
                                }
                            }

                            //case of THREE parts
                            if (len >= 3) {
                                //case : w1 X w2
                                Word eWord = this.getWord(eIndex - 1);
                                subProb = this.getBeta_wX(prod, cIndex, bIndex, eIndex - 1) * this.computeOneWordProb(prod, cWord, eWord);
                                if (len >= 4) {
                                    Word word1 = this.getWord(eIndex - 2);
                                    subProb += this.getBeta_w1Xw2(prod, cIndex, bIndex, eIndex - 1) * this.computeOneWordProb(prod, word1, eWord);
                                }
                                this.putBeta_w1Xw2(prod, cIndex, bIndex, eIndex, subProb);
                            }
                        }
                        /**
                         * END OF PREPROCESS
                         */

                        //case of TWO parts.
                        if (len >= 2) {
                            //case 2-a, 2-b : X | Y / Y | X
                            subProb = 0;
                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 1; sIndex++) {
                                int cIndex;
                                int rulePattern;
                                Word bWord;
                                Word eWord;

                                cIndex = 0;
                                bWord = Word.CHILD_A;
                                eWord = Word.CHILD_B;
                                rulePattern = Rule.RULE_XY;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_X(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, bWord, eWord) *
                                        this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);

                                cIndex = 1;
                                bWord = Word.CHILD_B;
                                eWord = Word.CHILD_A;
                                rulePattern = Rule.RULE_YX;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_X(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, bWord, eWord) *
                                        this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);
                            }
                            prob += subProb;
                        }

                        //case of THREE parts..
                        if (len >= 3) {
                            //case : X | w Y / Y | w X
                            subProb = 0;
                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 2; sIndex++) {
                                int cIndex;
                                int rulePattern;

                                Word bWord;
                                Word word = this.getWord(sIndex);
                                Word eWord;

                                cIndex = 0;
                                bWord = Word.CHILD_A;
                                eWord = Word.CHILD_B;
                                rulePattern = Rule.RULE_XwY;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_X(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, bWord, word) *
                                        this.getBeta_wX(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);

                                cIndex = 1;
                                bWord = Word.CHILD_B;
                                eWord = Word.CHILD_A;
                                rulePattern = Rule.RULE_YwX;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_X(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, bWord, word) *
                                        this.getBeta_wX(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);
                            }

                            //case : X | Y w / Y | X w
                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 2; sIndex++) {
                                int cIndex;
                                int rulePattern;

                                Word bWord;
                                Word word;
                                Word eWord = this.getWord(eIndex - 1);

                                cIndex = 0;
                                bWord = Word.CHILD_A;
                                word = Word.CHILD_B;
                                rulePattern = Rule.RULE_XYw;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_X(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, bWord, word) *
                                        this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);

                                cIndex = 1;
                                bWord = Word.CHILD_B;
                                word = Word.CHILD_A;
                                rulePattern = Rule.RULE_YXw;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_X(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, bWord, word) *
                                        this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);
                            }

                            //case : w X | Y / w Y | X
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                                int cIndex;
                                int rulePattern;

                                Word bWord = this.getWord(bIndex);
                                Word word;
                                Word eWord;

                                cIndex = 0;
                                word = Word.CHILD_A;
                                eWord = Word.CHILD_B;
                                rulePattern = Rule.RULE_wXY;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_wX(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, word, eWord) *
                                        this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);

                                cIndex = 1;
                                word = Word.CHILD_B;
                                eWord = Word.CHILD_A;
                                rulePattern = Rule.RULE_wYX;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_wX(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, word, eWord) *
                                        this.getBeta_X(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);
                            }
                            prob += subProb;
                        }

                        //case of FOUR parts
                        if (len >= 4) {
                            //case : X w1 | Y w2 / Y w1 | X w2
                            subProb = 0;
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                                int cIndex;
                                int rulePattern;

                                Word bWord;
                                Word word1 = this.getWord(sIndex - 1);
                                Word word2;
                                Word eWord = this.getWord(eIndex - 1);

                                cIndex = 0;
                                bWord = Word.CHILD_A;
                                word2 = Word.CHILD_B;
                                rulePattern = Rule.RULE_Xw1Yw2;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_Xw(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, word1, word2) *
                                        this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);

                                cIndex = 1;
                                bWord = Word.CHILD_B;
                                word2 = Word.CHILD_A;
                                rulePattern = Rule.RULE_Yw1Xw2;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_Xw(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, word1, word2) *
                                        this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);
                            }

                            //case : w1 X | w2 Y / w1 Y | w2 X
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                                int cIndex;
                                int rulePattern;

                                Word bWord = this.getWord(bIndex);
                                Word word1;
                                Word word2 = this.getWord(sIndex);
                                Word eWord;

                                cIndex = 0;
                                word1 = Word.CHILD_A;
                                eWord = Word.CHILD_B;
                                rulePattern = Rule.RULE_w1Xw2Y;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_wX(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, word1, word2) *
                                        this.getBeta_wX(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);

                                cIndex = 1;
                                word1 = Word.CHILD_B;
                                eWord = Word.CHILD_A;
                                rulePattern = Rule.RULE_w1Yw2X;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_wX(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, word1, word2) *
                                        this.getBeta_wX(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);
                            }

                            //case : w1 X | Y w2 / w1 Y | X w2
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                                int cIndex;
                                int rulePattern;

                                Word bWord = this.getWord(bIndex);
                                Word word1;
                                Word word2;
                                Word eWord = this.getWord(eIndex - 1);

                                cIndex = 0;
                                word1 = Word.CHILD_A;
                                word2 = Word.CHILD_B;
                                rulePattern = Rule.RULE_w1XYw2;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_wX(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, word1, word2) *
                                        this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);

                                cIndex = 1;
                                word1 = Word.CHILD_B;
                                word2 = Word.CHILD_A;
                                rulePattern = Rule.RULE_w1YXw2;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_wX(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, word1, word2) *
                                        this.getBeta_Xw(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);
                            }
                            prob += subProb;
                        }

                        //case of FIVE parts
                        if (len >= 5) {
                            subProb = 0;
                            //case : w1 X | w2 Y w3 / w1 Y | w2 X w3
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 3; sIndex++) {
                                int cIndex;
                                int rulePattern;

                                Word bWord = this.getWord(bIndex);
                                Word word1;
                                Word word2 = this.getWord(sIndex);
                                Word eWord = this.getWord(eIndex - 1);

                                cIndex = 0;
                                word1 = Word.CHILD_A;
                                rulePattern = Rule.RULE_w1Xw2Yw3;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_wX(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, word1, word2) *
                                        this.getBeta_w1Xw2(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);

                                cIndex = 1;
                                word1 = Word.CHILD_B;
                                rulePattern = Rule.RULE_w1Yw2Xw3;
                                subProb += this.computeRuleTypeProb(prod, rulePattern) * this.computeOneWordProb(prod, Word.BEGIN, bWord) *
                                        this.getBeta_wX(prod, cIndex, bIndex, sIndex) * this.computeOneWordProb(prod, word1, word2) *
                                        this.getBeta_w1Xw2(prod, 1 - cIndex, sIndex, eIndex) * this.computeOneWordProb(prod, eWord, Word.END);
                            }
                            prob += subProb;
                        }

                        this.setBeta(prod, bIndex, eIndex, prob);
                        if (_debugMode) {
                            viewBeta(prod, bIndex, eIndex);
                        }
                    } else {
                        System.out.println("Invalid number of children:"+prod.toLine());
                        System.err.println("Currently only support two children for a production.");
                        System.exit(1);
                        //error!
                    }
                }
            }
        }
    }

    /****************************************************************/
    /*************** End Of Inside/Outside Methods ******************/
    /****************************************************************/

    private void viewAlpha(Production prod, int bIndex, int eIndex) {
        double alpha_value = this._alpha[bIndex][eIndex - bIndex][prod.getProdSeq()];

        StringBuffer part = new StringBuffer();
        for (int i = bIndex; i < eIndex; i++) {
            part.append(this._words.get(i).getWordStr() + " ");
        }
        System.out.println("Alfa [" + prod.toLine() + "] : [" + part + "] " + alpha_value);
    }

    private void viewBeta(Production prod, int bIndex, int eIndex) {
        double beta_value = this._beta[bIndex][eIndex - bIndex][prod.getProdSeq()];

        StringBuffer part = new StringBuffer();
        for (int i = bIndex; i < eIndex; i++) {
            part.append(this._words.get(i).getWordStr() + " ");
        }
        System.out.println("Beta [" + prod.toLine() + "] : [" + part + "] " + beta_value);
    }

    private void setBeta(Production prod, int bIndex, int eIndex, double value) {
        this._beta[bIndex][eIndex - bIndex][prod.getProdSeq()] = value;
    }

    private void setAlpha(Production prod, int bIndex, int eIndex, double value) {
        this._alpha[bIndex][eIndex - bIndex][prod.getProdSeq()] = value;
    }

    private void addAlpha(Production prod, int bIndex, int eIndex, double value) {
        this._alpha[bIndex][eIndex - bIndex][prod.getProdSeq()] += value;
    }

    /**
     * Put Alpha Temporary Table Methods.
     */
    private void putAlpha_w(Production prod, int bIndex, int eIndex, double count) {
        this.alpha_w[prod.getProdSeq()][bIndex][eIndex - bIndex] += count;
    }

    private void putAlpha_Xw(Production prod, int cIndex, int bIndex, int eIndex, double count) {
        this.alpha_Xw[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex] += count;
    }

    private void putAlpha_wX(Production prod, int cIndex, int bIndex, int eIndex, double count) {
        this.alpha_wX[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex] += count;
    }

    private void putAlpha_w1Xw2(Production prod, int cIndex, int bIndex, int eIndex, double count) {
        this.alpha_w1Xw2[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex] += count;
    }

    /**
     * Put Beta Temporary Table Methods.
     */
    private void putBeta_w(Production prod, int bIndex, int eIndex, double count) {
        this.beta_w[prod.getProdSeq()][bIndex][eIndex - bIndex] += count;
    }

    private void putBeta_Xw(Production prod, int cIndex, int bIndex, int eIndex, double count) {
        this.beta_Xw[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex] += count;
    }

    private void putBeta_wX(Production prod, int cIndex, int bIndex, int eIndex, double count) {
        this.beta_wX[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex] += count;
    }

    private void putBeta_w1Xw2(Production prod, int cIndex, int bIndex, int eIndex, double count) {
        this.beta_w1Xw2[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex] += count;
    }

    private void putBeta_X(Production prod, int cIndex, int bIndex, int eIndex, double count) {
        this.beta_X[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex] += count;
    }

    /**
     * Update the Word Parameter
     * @param prod Production
     * @param word Word
     * @param count double
     */
    protected void updateWordParam(Production prod, Word word1, Word word, double count) {
        if (Double.isInfinite(count)) {
            throw new RuntimeException("x");
        }

        if (count == 0) {
            return;
        }

        count *= this._weight;
        ArrayList<InnerFeature> innerFeatures = prod.getGProduction().getInnerFeatures();

        double ratio = 1.0 / (double) innerFeatures.size();
        for (int i = 0; i < innerFeatures.size(); i++) {
            InnerFeature innerFeature = innerFeatures.get(i);

            double subProb;

            subProb = count / this.computeOneWordProb(prod, word1, word) * this._parameter.getInnerFeatureProb(innerFeature, word) * ratio;
            this._parameter.putCount(innerFeature, word1, word, subProb);

        }

    }

    protected double computeOneWordProb(Production prod, Word word1, Word word) {
        return this._parameter.getProb(prod.getGProduction(), word);
    }

}
