/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.core.learn;

/**
 * @author LU, Wei
 * @version 1.0
 */

import java.util.ArrayList;

import hybridtree.core.*;
import hybridtree.data.InnerFeature;
import hybridtree.data.Production;
import hybridtree.data.Rule;
import hybridtree.data.TrainExample;
import hybridtree.data.Word;


public class UGTrainTable extends TrainTable {

    private double beta_w[][][];
    private double beta_Xw[][][][];
    private double beta_wX[][][][];
    private double beta_w1Xw2[][][][];
    private double beta_X[][][][];

    private double alpha_w[][][];
    private double alpha_Xw[][][][];
    private double alpha_wX[][][][];
    private double alpha_w1Xw2[][][][];

    /**
     * Create a parse table for a training example
     * @param example the examples
     * @param parameter the parameter
     * @return a fresh ParseTable for this training example
     */
    public static UGTrainTable createTable(TrainExample example,
                                           Param parameter) {
        return new UGTrainTable(example, parameter);
    }

    /**
     * Constructor
     * @param example the Example
     * @param parameter the parameter
     */
    protected UGTrainTable(TrainExample example, Param parameter) {

        super(example, parameter);

        this._productions = example.getProductions();
        this._weight = example.getWeight();

    }

    /****************************************************************/
    /****************** End Of Public Methods ***********************/
    /****************************************************************/


    private double getBeta_w(Production prod, int bIndex, int eIndex) {
        return this.beta_w[prod.getProdSeq()][bIndex][eIndex - bIndex];
    }

    private double getBeta_X(Production prod, int cIndex, int bIndex,
                             int eIndex) {
        return this.beta_X[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private double getBeta_Xw(Production prod, int cIndex, int bIndex,
                              int eIndex) {
        return this.beta_Xw[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private double getBeta_wX(Production prod, int cIndex, int bIndex,
                              int eIndex) {
        return this.beta_wX[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private double getBeta_w1Xw2(Production prod, int cIndex, int bIndex,
                                 int eIndex) {
        return this.beta_w1Xw2[prod.getProdSeq()][cIndex][bIndex][eIndex -
                bIndex];
    }

    /**
     * Get Alpha Temporary Table Methods.
     */
    private double getAlpha_w(Production prod, int bIndex, int eIndex) {
        return this.alpha_w[prod.getProdSeq()][bIndex][eIndex - bIndex];
    }

    private double getAlpha_Xw(Production prod, int cIndex, int bIndex,
                               int eIndex) {
        return this.alpha_Xw[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private double getAlpha_wX(Production prod, int cIndex, int bIndex,
                               int eIndex) {
        return this.alpha_wX[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }

    private double getAlpha_w1Xw2(Production prod, int cIndex, int bIndex,
                                  int eIndex) {
        return this.alpha_w1Xw2[prod.getProdSeq()][cIndex][bIndex][eIndex -
                bIndex];
    }

    private void updateRuleProb(Production prod, int patternID, double count) {
        count *= this._weight;
        this._parameter.putPatternCount(prod.getGProduction(), patternID, count);
    }


    /****************************************************************/
    /****************** End Of Update Methods ***********************/
    /****************************************************************/

    /**
     * Reset the Outside Probability Tables
     */
    private void resetAlpha() {

        this._alpha = new double[this._N][][]; //bIndex, length, prod

        for (int i = 0; i < this._alpha.length; i++) {
            this._alpha[i] = new double[this._N - i + 1][];
            for (int j = 1; j < this._alpha[i].length; j++) {
                this._alpha[i][j] = new double[this._productions.size()];
            }
        }

        this.alpha_w = new double[this._productions.size()][][];
        this.alpha_Xw = new double[this._productions.size()][][][];
        this.alpha_wX = new double[this._productions.size()][][][];
        this.alpha_w1Xw2 = new double[this._productions.size()][][][];

        for (int prod_seq = 0; prod_seq < this._productions.size(); prod_seq++) {
            Production prod = (Production)this._productions.get(prod_seq);
            int childrensize = prod.countChildren();

            this.alpha_w[prod_seq] = new double[this._N][];
            this.alpha_Xw[prod_seq] = new double[childrensize][this._N][];
            this.alpha_wX[prod_seq] = new double[childrensize][this._N][];
            this.alpha_w1Xw2[prod_seq] = new double[childrensize][this._N][];

            for (int bIndex = 0; bIndex <= this._N - 1; bIndex++) {
                this.alpha_w[prod_seq][bIndex] = new double[this._N - bIndex +
                                                 1];
                for (int cIndex = 0; cIndex < childrensize; cIndex++) {
                    this.alpha_Xw[prod_seq][cIndex][bIndex] = new double[this.
                            _N - bIndex + 1];
                    this.alpha_wX[prod_seq][cIndex][bIndex] = new double[this.
                            _N - bIndex + 1];
                    this.alpha_w1Xw2[prod_seq][cIndex][bIndex] = new double[this.
                            _N - bIndex + 1];
                }
            }
        }

    }

    /**
     * Reset the Inside Probability Tables
     */
    private void resetBeta() {

        this._beta = new double[this._N][][]; //bIndex, length, prod
        for (int i = 0; i < this._beta.length; i++) {
            this._beta[i] = new double[this._N - i + 1][];
            for (int j = 1; j < this._beta[i].length; j++) {
                this._beta[i][j] = new double[this._productions.size()];
            }
        }

        this.beta_w = new double[this._productions.size()][][];
        this.beta_X = new double[this._productions.size()][][][];
        this.beta_Xw = new double[this._productions.size()][][][];
        this.beta_wX = new double[this._productions.size()][][][];
        this.beta_w1Xw2 = new double[this._productions.size()][][][];

        for (int prod_seq = this._productions.size() - 1; prod_seq >= 0;
                            prod_seq--) {
            Production prod = this._productions.get(prod_seq);
            int childrensize = prod.countChildren();

            this.beta_w[prod_seq] = new double[this._N][];
            this.beta_X[prod_seq] = new double[childrensize][this._N][];
            this.beta_Xw[prod_seq] = new double[childrensize][this._N][];
            this.beta_wX[prod_seq] = new double[childrensize][this._N][];
            this.beta_w1Xw2[prod_seq] = new double[childrensize][this._N][];

            for (int bIndex = 0; bIndex <= this._N - 1; bIndex++) {
                this.beta_w[prod_seq][bIndex] = new double[this._N - bIndex + 1];
                for (int cIndex = 0; cIndex < childrensize; cIndex++) {
                    this.beta_X[prod_seq][cIndex][bIndex] = new double[this._N -
                            bIndex + 1];
                    this.beta_Xw[prod_seq][cIndex][bIndex] = new double[this._N -
                            bIndex + 1];
                    this.beta_wX[prod_seq][cIndex][bIndex] = new double[this._N -
                            bIndex + 1];
                    this.beta_w1Xw2[prod_seq][cIndex][bIndex] = new double[this.
                            _N - bIndex + 1];
                }
            }
        }
    }

    /****************************************************************/
    /******************* End Of Reset Methods ***********************/
    /****************************************************************/

    /**
     * Computing the outside Probability
     */
    public void computeAlpha() {

        Production root = this._productions.get(0);
        double global_beta = getBeta(root, 0, this._N);

        if (global_beta == 0) {
            System.err.println(
                    "[Instance Ignored...] Beta Probability for this instance is zero! \n\t" +
                    this._sentence);
            _ignoreFlag = true;
            return;
        }

        this.resetAlpha();

        setAlpha(root, 0, this._N, 1.0);
        for (int k = 1; k < this._N; k++) {
            setAlpha(root, 0, k, 0.0);
        }

        for (int prod_seq = 0; prod_seq < this._productions.size(); prod_seq++) {

            Production prod = (Production)this._productions.get(prod_seq);

            ArrayList<Production> children = prod.getChildren();

            int childrensize = children.size();

            for (int cIndex = 0; cIndex < childrensize; cIndex++) {
                if (cIndex == 0) {
                    this.updateWordParam(prod, Word.CHILD_A, 1.0);
                } else {
                    this.updateWordParam(prod, Word.CHILD_B, 1.0);
                }
            }

            this.updateWordParam(prod, Word.END, 1.0);

            for (int len = this._N; len >= 1; len--) {
                for (int bIndex = 0; bIndex <= this._N - len; bIndex++) {
                    int eIndex = bIndex + len;

                    //for the contiguous case..
                    if (childrensize == 0) {

                        double subProb;
                        double count;

                        subProb = this.getAlpha(prod, bIndex, eIndex);
                        this.putAlpha_w(prod, bIndex, eIndex, subProb);

                        //update the rule probabilities..
                        {
                            double ruleCount = Math.log(subProb) +
                                               Math.log(this.
                                    computeEndProb(prod)) +
                                               Math.log(this.
                                    getBeta_w(prod, bIndex, eIndex)) +
                                               Math.log(this.
                                    computeRuleTypeProb(prod, Rule.RULE_w)) -
                                               Math.log(global_beta);
                            ruleCount = Math.exp(ruleCount);

                            this.updateRuleProb(prod, Rule.RULE_w, ruleCount);
                        }

                        //wIndex = this._words.get(bIndex).getWordIndex();
                        Word word = this._words.get(bIndex);
                        subProb = this.getAlpha_w(prod, bIndex, eIndex) *
                                  this.computeOneWordProb(prod, word);

                        if (len > 1) {
                            this.putAlpha_w(prod, bIndex + 1, eIndex, subProb);

                            //update...
                            count = Math.log(subProb) +
                                    Math.log(this.getBeta_w(prod, bIndex + 1,
                                    eIndex)) +
                                    Math.log(this.computeEndProb(prod)) +
                                    Math.log(this.computeRuleTypeProb(prod,
                                    Rule.RULE_w)) - Math.log(global_beta);
                            count = Math.exp(count);

                            this.updateWordParam(prod, word, count);
                        } else {

                            //update...
                            count = Math.log(subProb) +
                                    Math.log(this.computeEndProb(prod)) +
                                    Math.
                                    log(this.computeRuleTypeProb(prod,
                                    Rule.RULE_w)) -
                                    Math.log(global_beta);
                            count = Math.exp(count);

                            this.updateWordParam(prod, word, count);
                        }
                    } else if (childrensize == 1) {

                        int cIndex = 0;
                        Production child = children.get(cIndex);

                        double subProb;

                        subProb = this.getAlpha(prod, bIndex, eIndex);

                        this.putAlpha_w1Xw2(prod, cIndex, bIndex, eIndex,
                                            subProb *
                                            this.computeRuleTypeProb(prod,
                                Rule.RULE_w1Xw2));
                        this.putAlpha_wX(prod, cIndex, bIndex, eIndex,
                                         subProb *
                                         this.
                                         computeRuleTypeProb(prod, Rule.RULE_wX));
                        this.putAlpha_Xw(prod, cIndex, bIndex, eIndex,
                                         subProb *
                                         this.
                                         computeRuleTypeProb(prod, Rule.RULE_Xw));

                        //update rule pattern probabilities...
                        {
                            double ruleCount;

                            ruleCount = Math.log(subProb) +
                                        Math.log(this.
                                                 getBeta_w1Xw2(prod, cIndex,
                                    bIndex, eIndex)) +
                                        Math.log(this.computeEndProb(prod)) +
                                        Math.log(this.computeRuleTypeProb(prod,
                                    Rule.RULE_w1Xw2)) -
                                        Math.log(global_beta);
                            ruleCount = Math.exp(ruleCount);

                            this.updateRuleProb(prod, Rule.RULE_w1Xw2,
                                                ruleCount);

                            ruleCount = Math.log(subProb) +
                                        Math.log(this.
                                                 getBeta_wX(prod, cIndex,
                                    bIndex,
                                    eIndex)) +
                                        Math.log(this.computeEndProb(prod)) +
                                        Math.log(this.computeRuleTypeProb(prod,
                                    Rule.RULE_wX)) -
                                        Math.log(global_beta);
                            ruleCount = Math.exp(ruleCount);

                            this.updateRuleProb(prod, Rule.RULE_wX, ruleCount);

                            ruleCount = Math.log(subProb) +
                                        Math.log(this.
                                                 getBeta_Xw(prod, cIndex,
                                    bIndex,
                                    eIndex)) +
                                        Math.log(this.computeEndProb(prod)) +
                                        Math.log(this.computeRuleTypeProb(prod,
                                    Rule.RULE_Xw)) -
                                        Math.log(global_beta);
                            ruleCount = Math.exp(ruleCount);

                            this.updateRuleProb(prod, Rule.RULE_Xw, ruleCount);
                        }

                        /**...enclosed by its parent...**/
                        {
                            subProb = this.getAlpha(prod, bIndex, eIndex);
                            subProb *= this.computeTransProb(prod, cIndex) *
                                    computeEndProb(prod) *
                                    this.computeRuleTypeProb(prod, Rule.RULE_X);
                            this.addAlpha(child, bIndex, eIndex, subProb);

                            //update rule pattern probabilities...
                            double ruleCount = Math.log(subProb) +
                                               Math.log(this.
                                    getBeta(child, bIndex, eIndex)) -
                                               Math.log(global_beta);
                            ruleCount = Math.exp(ruleCount);

                            this.updateRuleProb(prod, Rule.RULE_X, ruleCount);
                        }

                        /**enclosed by w1Xw2**/
                        if (len >= 3) {
                            Word word = this._words.get(eIndex - 1);
                            subProb = this.getAlpha_w1Xw2(prod, cIndex, bIndex,
                                    eIndex) *
                                      this.computeOneWordProb(prod, word);
                            if (len >= 4) {
                                this.putAlpha_w1Xw2(prod, cIndex, bIndex,
                                        eIndex - 1, subProb);
                            }
                            this.putAlpha_wX(prod, cIndex, bIndex, eIndex - 1,
                                             subProb);

                            subProb *= this.computeEndProb(prod);

                            //update:
                            double count = Math.log(subProb) -
                                           Math.log(global_beta) +
                                           Math.log(this.getBeta_w1Xw2(prod,
                                    cIndex, bIndex, eIndex - 1) +
                                    this.getBeta_wX(prod, cIndex, bIndex,
                                    eIndex - 1));
                            count = Math.exp(count);

                            this.updateWordParam(prod, word, count);
                        }

                        /**enclosed by Xw**/
                        if (len >= 2) {
                            Word word = this._words.get(eIndex - 1);
                            subProb = this.getAlpha_Xw(prod, cIndex, bIndex,
                                    eIndex) *
                                      this.computeOneWordProb(prod, word);
                            if (len >= 3) {
                                this.putAlpha_Xw(prod, cIndex, bIndex,
                                                 eIndex - 1, subProb);
                            }

                            subProb *= this.computeEndProb(prod);

                            //propagate
                            double subAlpha = subProb *
                                              this.computeTransProb(prod,
                                    cIndex);
                            this.addAlpha(child, bIndex, eIndex - 1, subAlpha);

                            //update:
                            double count = Math.log(subProb) +
                                           Math.log(this.getBeta_Xw(prod,
                                    cIndex, bIndex, eIndex - 1) +
                                    this.getBeta_X(prod, cIndex, bIndex,
                                    eIndex - 1)) - Math.log(global_beta);
                            count = Math.exp(count);

                            this.updateWordParam(prod, word, count);
                        }

                        /**enclosed by wX**/
                        if (len >= 2) {
                            Word word = this._words.get(bIndex);
                            subProb = this.computeOneWordProb(prod, word) *
                                      this.
                                      getAlpha_wX(prod, cIndex, bIndex, eIndex);
                            if (len >= 3) {
                                this.putAlpha_wX(prod, cIndex, bIndex + 1,
                                                 eIndex, subProb);
                            }

                            subProb *= this.computeEndProb(prod);

                            //propagate
                            double subAlpha = subProb *
                                              this.computeTransProb(prod,
                                    cIndex);
                            this.addAlpha(child, bIndex + 1, eIndex, subAlpha);

                            //update:
                            /*
                             double count = (subProb) *
                                           (this.getBeta_X(prod, cIndex, bIndex + 1, eIndex) + this.getBeta_wX(prod, cIndex, bIndex + 1, eIndex)) /
                                           global_beta;
                             */
                            double count = Math.log(subProb) +
                                           Math.log(this.getBeta_X(prod, cIndex,
                                    bIndex + 1, eIndex) +
                                    this.getBeta_wX(prod, cIndex, bIndex + 1,
                                    eIndex)) - Math.log(global_beta);
                            count = Math.exp(count);

                            this.updateWordParam(prod, word, count);
                        }

                    } else if (childrensize == 2) {

                        double prob = this.getAlpha(prod, bIndex, eIndex);
                        double subProb;

                        //update rule probabilities...
                        {

                            //CASE OF FIVE PARTS...
                            if (len >= 5) {
                                // w1 X w2 | Y w3
                                // w1 Y w2 | X w3
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    double ruleCount = 0;
                                    for (int sIndex = bIndex + 3;
                                            sIndex <= eIndex - 2; sIndex++) {
                                        ruleCount +=
                                                this.getBeta_w1Xw2(prod, cIndex,
                                                bIndex, sIndex) *
                                                this.getBeta_Xw(prod,
                                                1 - cIndex, sIndex, eIndex);
                                    }
                                    if (cIndex == 0) {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_w1Xw2Yw3)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod,
                                                Rule.RULE_w1Xw2Yw3, ruleCount);
                                    } else {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_w1Yw2Xw3)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod,
                                                Rule.RULE_w1Yw2Xw3, ruleCount);
                                    }
                                }
                            }

                            //CASE OF FOUR PARTS..
                            if (len >= 4) {
                                // w1 X | w2 Y
                                // w1 Y | w2 X
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    double ruleCount = 0;
                                    for (int sIndex = bIndex + 2;
                                            sIndex <= eIndex - 2; sIndex++) {
                                        ruleCount +=
                                                this.getBeta_wX(prod, cIndex,
                                                bIndex, sIndex) *
                                                this.
                                                getBeta_wX(prod, 1 - cIndex,
                                                sIndex, eIndex);
                                    }
                                    if (cIndex == 0) {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_w1Xw2Y)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod,
                                                Rule.RULE_w1Xw2Y, ruleCount);
                                    } else {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_w1Yw2X)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod,
                                                Rule.RULE_w1Yw2X, ruleCount);
                                    }
                                }

                                // X w1 | Y w2
                                // Y w1 | X w2
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    double ruleCount = 0;
                                    for (int sIndex = bIndex + 2;
                                            sIndex <= eIndex - 2; sIndex++) {
                                        ruleCount +=
                                                this.getBeta_Xw(prod, cIndex,
                                                bIndex, sIndex) *
                                                this.
                                                getBeta_Xw(prod, 1 - cIndex,
                                                sIndex, eIndex);
                                    }
                                    if (cIndex == 0) {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_Xw1Yw2)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod,
                                                Rule.RULE_Xw1Yw2, ruleCount);
                                    } else {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_Yw1Xw2)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod,
                                                Rule.RULE_Yw1Xw2, ruleCount);
                                    }
                                }

                                // w1 X | Y w2
                                // w1 Y | X w2
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    double ruleCount = 0;
                                    for (int sIndex = bIndex + 2;
                                            sIndex <= eIndex - 2; sIndex++) {
                                        ruleCount +=
                                                this.getBeta_wX(prod, cIndex,
                                                bIndex, sIndex) *
                                                this.
                                                getBeta_Xw(prod, 1 - cIndex,
                                                sIndex, eIndex);
                                    }
                                    if (cIndex == 0) {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_w1XYw2)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod,
                                                Rule.RULE_w1XYw2, ruleCount);
                                    } else {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_w1YXw2)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod,
                                                Rule.RULE_w1YXw2, ruleCount);
                                    }
                                }
                            }

                            //CASE OF THREE PARTS.
                            if (len >= 3) {
                                // w X | Y
                                // w Y | X
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    double ruleCount = 0;
                                    for (int sIndex = bIndex + 2;
                                            sIndex <= eIndex - 1; sIndex++) {
                                        ruleCount +=
                                                this.getBeta_wX(prod, cIndex,
                                                bIndex, sIndex) *
                                                this.
                                                getBeta_X(prod, 1 - cIndex,
                                                sIndex,
                                                eIndex);
                                    }
                                    if (cIndex == 0) {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_wXY)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_wXY,
                                                ruleCount);
                                    } else {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_wYX)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_wYX,
                                                ruleCount);
                                    }
                                }

                                // X w | Y
                                // Y w | X
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    double ruleCount = 0;
                                    for (int sIndex = bIndex + 2;
                                            sIndex <= eIndex - 1; sIndex++) {
                                        ruleCount +=
                                                this.getBeta_Xw(prod, cIndex,
                                                bIndex, sIndex) *
                                                this.
                                                getBeta_X(prod, 1 - cIndex,
                                                sIndex,
                                                eIndex);
                                    }
                                    if (cIndex == 0) {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_XwY)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_XwY,
                                                ruleCount);
                                    } else {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_YwX)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_YwX,
                                                ruleCount);
                                    }
                                }

                                // X | Y w
                                // Y | X w
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    double ruleCount = 0;
                                    for (int sIndex = bIndex + 1;
                                            sIndex <= eIndex - 2; sIndex++) {
                                        ruleCount +=
                                                this.getBeta_X(prod, cIndex,
                                                bIndex, sIndex) *
                                                this.
                                                getBeta_Xw(prod, 1 - cIndex,
                                                sIndex, eIndex);
                                    }
                                    if (cIndex == 0) {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_XYw)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_XYw,
                                                ruleCount);
                                    } else {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_YXw)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_YXw,
                                                ruleCount);
                                    }
                                }
                            }

                            //CASE OF TWO PARTS.
                            if (len >= 2) {
                                // X Y
                                // Y X
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    double ruleCount = 0;
                                    for (int sIndex = bIndex + 1;
                                            sIndex <= eIndex - 1; sIndex++) {
                                        ruleCount +=
                                                this.getBeta_X(prod, cIndex,
                                                bIndex, sIndex) *
                                                this.
                                                getBeta_X(prod, 1 - cIndex,
                                                sIndex,
                                                eIndex);
                                    }
                                    if (cIndex == 0) {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_XY)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_XY,
                                                ruleCount);
                                    } else {
                                        ruleCount = Math.log(ruleCount) +
                                                Math.log(prob) +
                                                Math.log(this.
                                                computeRuleTypeProb(prod,
                                                Rule.RULE_YX)) +
                                                Math.log(this.computeEndProb(
                                                prod)) - Math.log(global_beta);
                                        ruleCount = Math.exp(ruleCount);

                                        this.updateRuleProb(prod, Rule.RULE_YX,
                                                ruleCount);
                                    }
                                }
                            }
                        }

                        //CASE OF FIVE PARTS.
                        if (len >= 5) {
                            //case : w1 X w2 | Y w3 / w1 Y w2 | X w3
                            for (int sIndex = bIndex + 3; sIndex <= eIndex - 2;
                                              sIndex++) {
                                int cIndex;

                                cIndex = 0;

                                // w1 Y w2 | * X w3 *
                                subProb = prob *
                                          this.getBeta_w1Xw2(prod, 1 - cIndex,
                                        bIndex, sIndex) *
                                          this.computeRuleTypeProb(prod,
                                        Rule.RULE_w1Yw2Xw3);
                                this.putAlpha_Xw(prod, cIndex, sIndex, eIndex,
                                                 subProb);

                                // * w1 X w2 * | Y w3
                                subProb = prob *
                                          this.getBeta_Xw(prod, 1 - cIndex,
                                        sIndex, eIndex) *
                                          this.computeRuleTypeProb(prod,
                                        Rule.RULE_w1Xw2Yw3);
                                this.putAlpha_w1Xw2(prod, cIndex, bIndex,
                                        sIndex, subProb);

                                cIndex = 1;

                                // w1 X w2 | * Y w3 *
                                subProb = prob *
                                          this.getBeta_w1Xw2(prod, 1 - cIndex,
                                        bIndex, sIndex) *
                                          this.computeRuleTypeProb(prod,
                                        Rule.RULE_w1Xw2Yw3);
                                this.putAlpha_Xw(prod, cIndex, sIndex, eIndex,
                                                 subProb);

                                // * w1 Y w2 * | X w3
                                subProb = prob *
                                          this.getBeta_Xw(prod, 1 - cIndex,
                                        sIndex, eIndex) *
                                          this.computeRuleTypeProb(prod,
                                        Rule.RULE_w1Yw2Xw3);
                                this.putAlpha_w1Xw2(prod, cIndex, bIndex,
                                        sIndex, subProb);
                            }
                        }

                        //CASE OF FOUR PARTS.
                        if (len >= 4) {
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2;
                                              sIndex++) {
                                //case : w1 X | w2 Y / w1 Y | w2 X
                                {
                                    int cIndex;

                                    cIndex = 0;
                                    // w1 Y | * w2 X *
                                    subProb = prob *
                                              this.getBeta_wX(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_w1Yw2X);
                                    this.putAlpha_wX(prod, cIndex, sIndex,
                                            eIndex, subProb);

                                    // * w1 X * | w2 Y
                                    subProb = prob *
                                              this.getBeta_wX(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_w1Xw2Y);
                                    this.putAlpha_wX(prod, cIndex, bIndex,
                                            sIndex, subProb);

                                    cIndex = 1;
                                    // w1 X | * w2 Y *
                                    subProb = prob *
                                              this.getBeta_wX(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_w1Xw2Y);
                                    this.putAlpha_wX(prod, cIndex, sIndex,
                                            eIndex, subProb);

                                    // * w1 Y * | w2 X
                                    subProb = prob *
                                              this.getBeta_wX(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_w1Yw2X);
                                    this.putAlpha_wX(prod, cIndex, bIndex,
                                            sIndex, subProb);
                                }

                                //case : X w1 | Y w2 / Y w1 | X w2
                                {
                                    int cIndex;

                                    cIndex = 0;

                                    // * X w1 * | Y w2
                                    subProb = prob *
                                              this.getBeta_Xw(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_Xw1Yw2);
                                    this.putAlpha_Xw(prod, cIndex, bIndex,
                                            sIndex, subProb);

                                    // Y w1 | * X w2 *
                                    subProb = prob *
                                              this.getBeta_Xw(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_Yw1Xw2);
                                    this.putAlpha_Xw(prod, cIndex, sIndex,
                                            eIndex, subProb);

                                    cIndex = 1;

                                    // * Y w1 * | X w2
                                    subProb = prob *
                                              this.getBeta_Xw(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_Yw1Xw2);
                                    this.putAlpha_Xw(prod, cIndex, bIndex,
                                            sIndex, subProb);

                                    // X w1 | * Y w2 *
                                    subProb = prob *
                                              this.getBeta_Xw(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_Xw1Yw2);
                                    this.putAlpha_Xw(prod, cIndex, sIndex,
                                            eIndex, subProb);

                                }

                                //case : w1 X | Y w2 / w1 Y | X w2
                                {
                                    int cIndex;

                                    cIndex = 0;
                                    // * w1 X * | Y w2
                                    subProb = prob *
                                              this.getBeta_Xw(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_w1XYw2);
                                    this.putAlpha_wX(prod, cIndex, bIndex,
                                            sIndex, subProb);

                                    // w1 Y | * X w2 *
                                    subProb = prob *
                                              this.getBeta_wX(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_w1YXw2);
                                    this.putAlpha_Xw(prod, cIndex, sIndex,
                                            eIndex, subProb);

                                    cIndex = 1;
                                    // * w1 Y * | X w2
                                    subProb = prob *
                                              this.getBeta_Xw(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_w1YXw2);
                                    this.putAlpha_wX(prod, cIndex, bIndex,
                                            sIndex, subProb);

                                    // w1 X | * Y w2 *
                                    subProb = prob *
                                              this.getBeta_wX(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.computeRuleTypeProb(prod,
                                            Rule.RULE_w1XYw2);
                                    this.putAlpha_Xw(prod, cIndex, sIndex,
                                            eIndex, subProb);

                                }
                            }
                        }

                        //CASE OF THREE PARTS.
                        if (len >= 3) {
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 1;
                                              sIndex++) {
                                //case : w X | Y / w Y | X
                                {
                                    int cIndex;
                                    Production child;

                                    cIndex = 0;
                                    child = children.get(cIndex);

                                    // * w X * | Y
                                    subProb = prob *
                                              this.getBeta_X(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_wXY);

                                    this.putAlpha_wX(prod, cIndex, bIndex,
                                            sIndex, subProb);

                                    // w Y | * X *
                                    subProb = prob *
                                              this.getBeta_wX(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_wYX);

                                    //propagate
                                    subProb *=
                                            this.computeTransProb(prod, cIndex) *
                                            this.computeEndProb(prod);
                                    this.addAlpha(child, sIndex, eIndex,
                                                  subProb);

                                    cIndex = 1;
                                    child = children.get(cIndex);

                                    // * w Y * | X
                                    subProb = prob *
                                              this.getBeta_X(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_wYX);

                                    this.putAlpha_wX(prod, cIndex, bIndex,
                                            sIndex, subProb);

                                    // w X | * Y *
                                    subProb = prob *
                                              this.getBeta_wX(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_wXY);

                                    //propagate
                                    subProb *=
                                            this.computeTransProb(prod, cIndex) *
                                            this.computeEndProb(prod);
                                    this.addAlpha(child, sIndex, eIndex,
                                                  subProb);
                                }

                                //case : X w | Y / Y w | X
                                {
                                    int cIndex;
                                    Production child;

                                    cIndex = 0;
                                    child = children.get(cIndex);

                                    // * X w * | Y
                                    subProb = prob *
                                              this.getBeta_X(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_XwY);

                                    this.putAlpha_Xw(prod, cIndex, bIndex,
                                            sIndex, subProb);

                                    // Y w | * X *
                                    subProb = prob *
                                              this.getBeta_Xw(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_YwX);

                                    //propagate
                                    subProb *=
                                            this.computeTransProb(prod, cIndex) *
                                            this.computeEndProb(prod);
                                    this.addAlpha(child, sIndex, eIndex,
                                                  subProb);

                                    cIndex = 1;
                                    child = children.get(cIndex);

                                    // * Y w * | X
                                    subProb = prob *
                                              this.getBeta_X(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_YwX);

                                    this.putAlpha_Xw(prod, cIndex, bIndex,
                                            sIndex, subProb);

                                    // X w | * Y *
                                    subProb = prob *
                                              this.getBeta_Xw(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_XwY);

                                    //propagate
                                    subProb *=
                                            this.computeTransProb(prod, cIndex) *
                                            this.computeEndProb(prod);
                                    this.addAlpha(child, sIndex, eIndex,
                                                  subProb);
                                }
                            }

                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 2;
                                              sIndex++) {
                                //case : X | Y w / Y | X w
                                {
                                    int cIndex;
                                    Production child;

                                    cIndex = 0;
                                    child = children.get(cIndex);

                                    // * X * | Y w
                                    subProb = prob *
                                              this.getBeta_Xw(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_XYw);

                                    //propagate
                                    subProb *=
                                            this.computeTransProb(prod, cIndex) *
                                            this.computeEndProb(prod);
                                    this.addAlpha(child, bIndex, sIndex,
                                                  subProb);

                                    // Y | * X w *
                                    subProb = prob *
                                              this.getBeta_X(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_YXw);

                                    this.putAlpha_Xw(prod, cIndex, sIndex,
                                            eIndex, subProb);

                                    cIndex = 1;
                                    child = children.get(cIndex);

                                    // * Y * | X w
                                    subProb = prob *
                                              this.getBeta_Xw(prod, 1 - cIndex,
                                            sIndex, eIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_YXw);

                                    //propagate
                                    subProb *=
                                            this.computeTransProb(prod, cIndex) *
                                            this.computeEndProb(prod);
                                    this.addAlpha(child, bIndex, sIndex,
                                                  subProb);

                                    // X | * Y w *
                                    subProb = prob *
                                              this.getBeta_X(prod, 1 - cIndex,
                                            bIndex, sIndex) *
                                              this.
                                              computeRuleTypeProb(prod,
                                            Rule.RULE_XYw);

                                    this.putAlpha_Xw(prod, cIndex, sIndex,
                                            eIndex, subProb);
                                }
                            }
                        }

                        //CASE OF TWO PARTS..
                        //case : X | Y / Y | X
                        if (len >= 2) {
                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 1;
                                              sIndex++) {
                                int cIndex;
                                Production child;

                                cIndex = 0;
                                child = children.get(cIndex);

                                // * X * | Y
                                subProb = prob *
                                          this.getBeta_X(prod, 1 - cIndex,
                                        sIndex, eIndex) *
                                          this.
                                          computeRuleTypeProb(prod,
                                        Rule.RULE_XY);

                                subProb *= this.computeTransProb(prod, cIndex) *
                                        this.computeEndProb(prod);

                                //propagate
                                this.addAlpha(child, bIndex, sIndex, subProb);

                                // Y | * X *
                                subProb = prob *
                                          this.getBeta_X(prod, 1 - cIndex,
                                        bIndex, sIndex) *
                                          this.
                                          computeRuleTypeProb(prod,
                                        Rule.RULE_YX);

                                subProb *= this.computeTransProb(prod, cIndex) *
                                        this.computeEndProb(prod);

                                //propagate
                                this.addAlpha(child, sIndex, eIndex, subProb);

                                cIndex = 1;
                                child = children.get(cIndex);

                                // * Y * | X
                                subProb = prob *
                                          this.getBeta_X(prod, 1 - cIndex,
                                        sIndex, eIndex) *
                                          this.
                                          computeRuleTypeProb(prod,
                                        Rule.RULE_YX);

                                subProb *= this.computeTransProb(prod, cIndex) *
                                        this.computeEndProb(prod);

                                //propagate
                                this.addAlpha(child, bIndex, sIndex, subProb);

                                // X | * Y *
                                subProb = prob *
                                          this.getBeta_X(prod, 1 - cIndex,
                                        bIndex, sIndex) *
                                          this.
                                          computeRuleTypeProb(prod,
                                        Rule.RULE_XY);

                                subProb *= this.computeTransProb(prod, cIndex) *
                                        this.computeEndProb(prod);

                                //propagate
                                this.addAlpha(child, sIndex, eIndex, subProb);
                            }
                        }

                        /**POST PROCESSING BEGINS**/
                        {
                            double count;

                            for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                Production child = children.get(cIndex);
                                // case : w1 X w2
                                if (len >= 3) {
                                    Word word = this._words.get(bIndex);
                                    subProb = this.computeOneWordProb(prod,
                                            word) *
                                              this.getAlpha_w1Xw2(prod, cIndex,
                                            bIndex, eIndex);
                                    this.putAlpha_Xw(prod, cIndex, bIndex + 1,
                                            eIndex, subProb);
                                    this.putAlpha_w1Xw2(prod, cIndex,
                                            bIndex + 1, eIndex, subProb);

                                    subProb *= this.computeEndProb(prod);

                                    //update...
                                    /*
                                     count = (subProb) *
                                     (this.getBeta_w1Xw2(prod, cIndex, bIndex + 1, eIndex) + this.getBeta_Xw(prod, cIndex, bIndex + 1, eIndex)) /
                                            global_beta;
                                     */
                                    count = Math.log(subProb) +
                                            Math.log(this.getBeta_w1Xw2(prod,
                                            cIndex, bIndex + 1, eIndex) +
                                            this.getBeta_Xw(prod, cIndex,
                                            bIndex + 1, eIndex)) -
                                            Math.log(global_beta);
                                    count = Math.exp(count);

                                    this.updateWordParam(prod, word, count);
                                }

                                // case : w X
                                if (len >= 2) {
                                    Word word = this._words.get(bIndex);
                                    subProb = this.computeOneWordProb(prod,
                                            word) *
                                              this.getAlpha_wX(prod, cIndex,
                                            bIndex, eIndex);
                                    this.putAlpha_wX(prod, cIndex, bIndex + 1,
                                            eIndex, subProb);
                                    subProb *= this.computeEndProb(prod);

                                    //propagete
                                    double subAlpha = subProb *
                                            this.computeTransProb(prod, cIndex);
                                    this.addAlpha(child, bIndex + 1, eIndex,
                                                  subAlpha);

                                    //update..
                                    /*
                                     count = (subProb) *
                                            (this.getBeta_wX(prod, cIndex, bIndex + 1, eIndex) + this.getBeta_X(prod, cIndex, bIndex + 1, eIndex)) /
                                            global_beta;
                                     */
                                    count = Math.log(subProb) +
                                            Math.log(this.getBeta_wX(prod,
                                            cIndex, bIndex + 1, eIndex) +
                                            this.getBeta_X(prod, cIndex,
                                            bIndex + 1, eIndex)) -
                                            Math.log(global_beta);
                                    count = Math.exp(count);

                                    this.updateWordParam(prod, word, count);
                                }

                                // case : X w
                                if (len >= 2) {
                                    Word word = this._words.get(eIndex - 1);
                                    subProb = this.getAlpha_Xw(prod, cIndex,
                                            bIndex, eIndex) *
                                              this.
                                              computeOneWordProb(prod, word);
                                    this.putAlpha_Xw(prod, cIndex, bIndex,
                                            eIndex - 1, subProb);

                                    subProb *= this.computeEndProb(prod);

                                    //propagate
                                    double subAlpha = subProb *
                                            this.computeTransProb(prod, cIndex);
                                    this.addAlpha(child, bIndex, eIndex - 1,
                                                  subAlpha);

                                    //update...
                                    /*
                                     count = (subProb) *
                                            (this.getBeta_Xw(prod, cIndex, bIndex, eIndex - 1) + this.getBeta_X(prod, cIndex, bIndex, eIndex - 1)) /
                                            global_beta;
                                     */
                                    count = Math.log(subProb) +
                                            Math.log(this.getBeta_Xw(prod,
                                            cIndex, bIndex, eIndex - 1) +
                                            this.getBeta_X(prod, cIndex, bIndex,
                                            eIndex - 1)) - Math.log(global_beta);
                                    count = Math.exp(count);

                                    this.updateWordParam(prod, word, count);
                                }
                            }
                        }

                        /**POST PROCESSING ENDS**/

                    } else {
                        System.out.println("Invalid number of children:" +
                                           prod.toLine());
                        System.err.println(
                                "Currently only support at most two children.");
                        System.exit(1);
                    }
                }
            }
        }

        double oldSum = Double.NEGATIVE_INFINITY;
        for (int k = 0; k < this._productions.size(); k++) {
            Production prod = this._productions.get(k);
            double sum = 0;
            for (int i = 0; i <= this._N; i++) {
                for (int j = i + 1; j <= this._N; j++) {
                    sum += this.getAlpha(prod, i, j) * this.getBeta(prod, i, j);
                    if(_debugMode){
                        this.viewAlpha(prod, i, j);
                        this.viewBeta(prod, i, j);
                    }
                }
            }

            double subSum = 0;

            sum += subSum;

            if (oldSum == Double.NEGATIVE_INFINITY) {
                oldSum = sum;
            } else {
                if (Math.abs(oldSum - sum) > Math.pow(10, -10) * sum) {
                    System.err.println("Sum=" + sum + ", subSum = " + subSum +
                                       ",Total=" + (sum + subSum));
                    System.err.println("diff=" + Math.abs(oldSum - sum));
                    System.err.println("Math.pow(10, -10) * sum=" +
                                       Math.pow(10, -10) * sum);
                    System.err.println(this._productions);
                    this.viewBeta(this._productions.get(0), 0, this._N);
                    System.exit(1);
                } else {
                }
            }
        }
    }

    /**
     * Computing the Inside Probability
     */
    public void computeBeta() {

        this.resetBeta();

        /**
         * consider each production from bottom up...
         */
        for (int prod_seq = this._productions.size() - 1; prod_seq >= 0;
                            prod_seq--) {
            Production prod = this._productions.get(prod_seq);
            int childrensize = prod.countChildren();

            /**
             * varying length with increasing order: 1 to N
             */
            for (int len = 1; len <= this._N; len++) {

                /**
                 * the begin index of the span
                 */
                for (int bIndex = 0; bIndex <= this._N - len; bIndex++) {
                    int eIndex = bIndex + len;

                    double prob = 0;
                    if (childrensize == 0) {

                        //case 1: w
                        Word word = this.getWord(eIndex - 1);
                        double wProb;
                        if (len == 1) {
                            wProb = this.computeOneWordProb(prod, word);
                        } else {
                            wProb = this.getBeta_w(prod, bIndex, eIndex - 1) *
                                    this.computeOneWordProb(prod, word);
                        }
                        this.putBeta_w(prod, bIndex, eIndex, wProb);

                        prob = this.computeRuleTypeProb(prod, Rule.RULE_w) *
                               this.getBeta_w(prod, bIndex, eIndex) *
                               this.computeEndProb(prod);
                        this.setBeta(prod, bIndex, eIndex, prob);

                        if (_debugMode) {
                            viewBeta(prod, bIndex, eIndex);
                        }

                    } else if (childrensize == 1) {
                        int cIndex = 0;
                        Production child = prod.getChildren().get(cIndex);
                        double subProb;

                        //case of ONE part
                        {
                            //case : X
                            subProb = this.computeTransProb(prod, cIndex) *
                                      this.getBeta(child, bIndex, eIndex);
                            this.putBeta_X(prod, cIndex, bIndex, eIndex,
                                           subProb);

                            prob += this.computeRuleTypeProb(prod, Rule.RULE_X) *
                                    subProb;
                        }

                        //case of TWO parts
                        if (len >= 2) {
                            Word word;

                            //case : w X
                            word = this.getWord(bIndex);
                            subProb = this.computeOneWordProb(prod, word) *
                                      (this.getBeta_wX(prod, cIndex, bIndex + 1,
                                    eIndex) +
                                       this.getBeta_X(prod, cIndex, bIndex + 1,
                                    eIndex));

                            this.putBeta_wX(prod, cIndex, bIndex, eIndex,
                                            subProb);

                            prob += subProb *
                                    this.computeRuleTypeProb(prod, Rule.RULE_wX);

                            //case : X w
                            word = this.getWord(eIndex - 1);
                            subProb = (this.getBeta_Xw(prod, cIndex, bIndex,
                                    eIndex - 1) +
                                       this.getBeta_X(prod, cIndex, bIndex,
                                    eIndex - 1)) *
                                      this.computeOneWordProb(prod, word);

                            this.putBeta_Xw(prod, cIndex, bIndex, eIndex,
                                            subProb);

                            prob += subProb *
                                    this.computeRuleTypeProb(prod, Rule.RULE_Xw);
                        }

                        //case of THREE parts
                        if (len >= 3) {
                            Word word;

                            //case : w1 X w2
                            word = this._words.get(bIndex);
                            subProb = this.computeOneWordProb(prod, word) *
                                      (this.getBeta_w1Xw2(prod, cIndex,
                                    bIndex + 1, eIndex) +
                                       this.getBeta_Xw(prod, cIndex, bIndex + 1,
                                    eIndex));

                            this.putBeta_w1Xw2(prod, cIndex, bIndex, eIndex,
                                               subProb);

                            prob += subProb *
                                    this.computeRuleTypeProb(prod,
                                    Rule.RULE_w1Xw2);
                        }

                        prob *= this.computeEndProb(prod);
                        this.setBeta(prod, bIndex, eIndex, prob);

                        if (_debugMode) {
                            viewBeta(prod, bIndex, eIndex);
                        }

                    } else if (childrensize == 2) {
                        Production child;
                        double subProb;

                        /**
                         * BEGIN OF PREPROCESS
                         */
                        for (int cIndex = 0; cIndex <= 1; cIndex++) {

                            child = prod.getChildren().get(cIndex);

                            //case of ONE part
                            {
                                //case 1-(a): X
                                subProb = this.computeTransProb(prod, cIndex) *
                                          this.getBeta(child, bIndex, eIndex);

                                this.putBeta_X(prod, cIndex, bIndex, eIndex,
                                               subProb);
                            }

                            //case of TWO parts
                            if (len >= 2) {
                                Word word;

                                //case 1-(b) w X
                                word = this.getWord(bIndex);
                                subProb = this.computeOneWordProb(prod, word) *
                                          (this.getBeta_wX(prod, cIndex,
                                        bIndex + 1, eIndex) +
                                           this.getBeta_X(prod, cIndex,
                                        bIndex + 1, eIndex));
                                this.putBeta_wX(prod, cIndex, bIndex, eIndex,
                                                subProb);

                                //case 1-(c) X w
                                word = this.getWord(eIndex - 1);
                                subProb = (this.getBeta_Xw(prod, cIndex, bIndex,
                                        eIndex - 1) +
                                           this.getBeta_X(prod, cIndex, bIndex,
                                        eIndex - 1)) *
                                          this.computeOneWordProb(prod, word);

                                this.putBeta_Xw(prod, cIndex, bIndex, eIndex,
                                                subProb);
                            }

                            //case of THREE parts
                            if (len >= 3) {
                                //case 1: w1 X w2
                                Word word = this.getWord(bIndex);
                                subProb = this.computeOneWordProb(prod, word) *
                                          (this.getBeta_w1Xw2(prod, cIndex,
                                        bIndex + 1, eIndex) +
                                           this.getBeta_Xw(prod, cIndex,
                                        bIndex + 1, eIndex));

                                this.putBeta_w1Xw2(prod, cIndex, bIndex, eIndex,
                                        subProb);
                            }
                        }
                        /**
                         * END OF PREPROCESS
                         */

                        //case of TWO parts.
                        if (len >= 2) {
                            //case 2-a, 2-b : X | Y / Y | X
                            subProb = 0;
                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 1;
                                              sIndex++) {

                                int cIndex;

                                cIndex = 0;
                                subProb +=
                                        this.getBeta_X(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_X(prod, 1 - cIndex, sIndex,
                                                  eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_XY);

                                cIndex = 1;
                                subProb +=
                                        this.getBeta_X(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_X(prod, 1 - cIndex, sIndex,
                                                  eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_YX);
                            }
                            prob += subProb;
                        }

                        //case of THREE parts..
                        if (len >= 3) {
                            //case : X w | Y / Y w | X
                            subProb = 0;
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 1;
                                              sIndex++) {

                                int cIndex;
                                cIndex = 0;
                                subProb +=
                                        this.getBeta_Xw(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_X(prod, 1 - cIndex, sIndex,
                                                  eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_XwY);

                                cIndex = 1;
                                subProb +=
                                        this.getBeta_Xw(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_X(prod, 1 - cIndex, sIndex,
                                                  eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_YwX);
                            }

                            //case : X | Y w / Y | X w
                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 2;
                                              sIndex++) {
                                int cIndex;
                                cIndex = 0;
                                subProb +=
                                        this.getBeta_X(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_Xw(prod, 1 - cIndex, sIndex,
                                        eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_XYw);

                                cIndex = 1;
                                subProb +=
                                        this.getBeta_X(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_Xw(prod, 1 - cIndex, sIndex,
                                        eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_YXw);
                            }

                            //case : w X | Y / w Y | X
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 1;
                                              sIndex++) {
                                int cIndex;
                                cIndex = 0;
                                subProb +=
                                        this.getBeta_wX(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_X(prod, 1 - cIndex, sIndex,
                                                  eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_wXY);

                                cIndex = 1;
                                subProb +=
                                        this.getBeta_wX(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_X(prod, 1 - cIndex, sIndex,
                                                  eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_wYX);
                            }
                            prob += subProb;
                        }

                        //case of FOUR parts
                        if (len >= 4) {
                            //case : X w1 | Y w2 / Y w1 | X w2
                            subProb = 0;
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2;
                                              sIndex++) {
                                int cIndex;
                                cIndex = 0;
                                subProb +=
                                        this.getBeta_Xw(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_Xw(prod, 1 - cIndex, sIndex,
                                        eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_Xw1Yw2);

                                cIndex = 1;
                                subProb +=
                                        this.getBeta_Xw(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_Xw(prod, 1 - cIndex, sIndex,
                                        eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_Yw1Xw2);
                            }

                            //case : w1 X | w2 Y / w1 Y | w2 X
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2;
                                              sIndex++) {
                                int cIndex;
                                cIndex = 0;
                                subProb +=
                                        this.getBeta_wX(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_wX(prod, 1 - cIndex, sIndex,
                                        eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_w1Xw2Y);

                                cIndex = 1;
                                subProb +=
                                        this.getBeta_wX(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_wX(prod, 1 - cIndex, sIndex,
                                        eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_w1Yw2X);
                            }

                            //case : w1 X | Y w2 / w1 Y | X w2
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2;
                                              sIndex++) {
                                int cIndex;
                                cIndex = 0;
                                subProb +=
                                        this.getBeta_wX(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_Xw(prod, 1 - cIndex, sIndex,
                                        eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_w1XYw2);

                                cIndex = 1;
                                subProb +=
                                        this.getBeta_wX(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.
                                        getBeta_Xw(prod, 1 - cIndex, sIndex,
                                        eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_w1YXw2);
                            }
                            prob += subProb;
                        }

                        //case of FIVE parts
                        if (len >= 5) {
                            subProb = 0;
                            //case : w1 X w2 | Y w3 / w1 Y w2 | X w3
                            for (int sIndex = bIndex + 3; sIndex <= eIndex - 2;
                                              sIndex++) {
                                int cIndex;
                                cIndex = 0;
                                subProb +=
                                        this.getBeta_w1Xw2(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.getBeta_Xw(prod, 1 - cIndex,
                                        sIndex, eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_w1Xw2Yw3);

                                cIndex = 1;
                                subProb +=
                                        this.getBeta_w1Xw2(prod, cIndex, bIndex,
                                        sIndex) *
                                        this.getBeta_Xw(prod, 1 - cIndex,
                                        sIndex, eIndex) *
                                        this.computeRuleTypeProb(prod,
                                        Rule.RULE_w1Yw2Xw3);
                            }
                            prob += subProb;
                        }

                        prob *= this.computeEndProb(prod);
                        this.setBeta(prod, bIndex, eIndex, prob);

                        if (_debugMode) {
                            viewBeta(prod, bIndex, eIndex);
                        }

                    } else {
                        System.out.println("Invalid number of children:" +
                                           prod.toLine());
                        System.err.println(
                                "Currently only support two children for a production.");
                        System.exit(1);
                        //error!
                    }
                }
            }
        }

        //if the inside probability is too small, we simply discard it...
        Production root = this._productions.get(0);

        if (this.getBeta(root, 0, this._N) < 1E-310) {
            System.err.println(
                    "[warning]\tDiscarded due to small inside probability!");
            this.setBeta(root, 0, this._N, 0);
        }

    }

    /****************************************************************/
    /*************** End Of Inside/Outside Methods ******************/
    /****************************************************************/

    private void viewAlpha(Production prod, int bIndex, int eIndex) {
        double alpha_value = this._alpha[bIndex][eIndex -
                             bIndex][prod.getProdSeq()];

        StringBuffer part = new StringBuffer();
        for (int i = bIndex; i < eIndex; i++) {
            part.append(this._words.get(i).getWordStr() + " ");
        }
        System.out.println("Alfa [" + prod.toLine() + "] : [" + part + "] " +
                           alpha_value);
    }

    private void viewBeta(Production prod, int bIndex, int eIndex) {
        double beta_value = this._beta[bIndex][eIndex - bIndex][prod.getProdSeq()];

        StringBuffer part = new StringBuffer();
        for (int i = bIndex; i < eIndex; i++) {
            part.append(this._words.get(i).getWordStr() + " ");
        }
        System.out.println("Beta [" + prod.toLine() + "] : [" + part + "] " +
                           beta_value);
    }

    private void setBeta(Production prod, int bIndex, int eIndex, double value) {
        this._beta[bIndex][eIndex - bIndex][prod.getProdSeq()] = value;
    }

    private void setAlpha(Production prod, int bIndex, int eIndex, double value) {
        this._alpha[bIndex][eIndex - bIndex][prod.getProdSeq()] = value;
    }

    private void addAlpha(Production prod, int bIndex, int eIndex, double value) {
        this._alpha[bIndex][eIndex - bIndex][prod.getProdSeq()] += value;
    }

    /**
     * Put Alpha Temporary Table Methods.
     */
    private void putAlpha_w(Production prod, int bIndex, int eIndex,
                            double count) {
        this.alpha_w[prod.getProdSeq()][bIndex][eIndex - bIndex] += count;
    }

    private void putAlpha_Xw(Production prod, int cIndex, int bIndex,
                             int eIndex, double count) {
        this.alpha_Xw[prod.getProdSeq()][cIndex][bIndex][eIndex -
                bIndex] += count;
    }

    private void putAlpha_wX(Production prod, int cIndex, int bIndex,
                             int eIndex, double count) {
        this.alpha_wX[prod.getProdSeq()][cIndex][bIndex][eIndex -
                bIndex] += count;
    }

    private void putAlpha_w1Xw2(Production prod, int cIndex, int bIndex,
                                int eIndex, double count) {
        this.alpha_w1Xw2[prod.getProdSeq()][cIndex][bIndex][eIndex -
                bIndex] += count;
    }

    /**
     * Put Beta Temporary Table Methods.
     */
    private void putBeta_w(Production prod, int bIndex, int eIndex,
                           double count) {
        this.beta_w[prod.getProdSeq()][bIndex][eIndex - bIndex] += count;
    }

    private void putBeta_Xw(Production prod, int cIndex, int bIndex, int eIndex,
                            double count) {
        this.beta_Xw[prod.getProdSeq()][cIndex][bIndex][eIndex -
                bIndex] += count;
    }

    private void putBeta_wX(Production prod, int cIndex, int bIndex, int eIndex,
                            double count) {
        this.beta_wX[prod.getProdSeq()][cIndex][bIndex][eIndex -
                bIndex] += count;
    }

    private void putBeta_w1Xw2(Production prod, int cIndex, int bIndex,
                               int eIndex, double count) {
        this.beta_w1Xw2[prod.getProdSeq()][cIndex][bIndex][eIndex -
                bIndex] += count;
    }

    private void putBeta_X(Production prod, int cIndex, int bIndex, int eIndex,
                           double count) {
        this.beta_X[prod.getProdSeq()][cIndex][bIndex][eIndex -
                bIndex] += count;
    }

    /**
     * Update the Word Parameter
     * @param prod Production
     * @param word Word
     * @param count double
     */
    private void updateWordParam(Production prod, Word word, double count) {

        if (count == 0) {
            return;
        }

        count *= this._weight;
        ArrayList<InnerFeature>
                innerFeatures = prod.getGProduction().getInnerFeatures();
        double ratio = 1.0 / (double) innerFeatures.size();

        for (int i = 0; i < innerFeatures.size(); i++) {
            InnerFeature innerFeature = innerFeatures.get(i);
            double subProb = count / this.computeOneWordProb(prod, word) *
                             this.
                             _parameter.getInnerFeatureProb(innerFeature, word) *
                             ratio;
            this._parameter.putCount(innerFeature, word, subProb);
        }

    }

    private double computeEndProb(Production prod) {
        return this._parameter.getProb(prod.getGProduction(), Word.END);
    }

    private double computeOneWordProb(Production prod, Word word) {
        return this._parameter.getProb(prod.getGProduction(), word);
    }

    private double computeTransProb(Production prod, int cIndex) {

        double prob;
        if (cIndex == 0) {
            prob = this._parameter.getProb(prod.getGProduction(), Word.CHILD_A);
        } else {
            prob = this._parameter.getProb(prod.getGProduction(), Word.CHILD_B);
        }
        return prob;

    }

}
