/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.core.parse;

/**
 * @author LU, Wei
 * @version 1.0
 */


import java.util.ArrayList;

import hybridtree.core.Candidate;
import hybridtree.core.CandidateSet;
import hybridtree.core.DeriveTable;
import hybridtree.core.HybridNode;
import hybridtree.core.Param;
import hybridtree.core.ProdHybridNode;
import hybridtree.core.ResElem;
import hybridtree.core.ResElemList;
import hybridtree.data.GProduction;
import hybridtree.data.Production;
import hybridtree.data.Rule;
import hybridtree.data.TestExample;
import hybridtree.data.Word;



public abstract class ParseTable extends DeriveTable {

    private TestExample _example;

    protected double[][][] _beta;

    /**
     * all GProductions
     */
    protected ArrayList<GProduction> _allGProds;

    /**
     * the sorted GProductions
     */
    protected ArrayList<GProduction> _sortedGProds;

    /**
     * Constructor
     * @param words all the words in string representations
     * @param wordIndices the words' indices
     * @param allprods all the GProductions
     * @param allnonts ArrayList
     * @param parameter Param
     * @param sorted ArrayList
     */
    protected ParseTable(TestExample example, Param parameter) {

        super(example, parameter);

        this._example = example;
        this._sortedGProds = parameter.getModel().sortProds();
        this._allGProds = parameter.getAllGProds();

        /*
        //remove unknown words..
        for(int i =0; i<this._words.size(); i++){
            Word word = this._words.get(i);
            if(word.equals(Word.UNKNOWN)){
                this._words.remove(i);
                i--;
            }
        }
        */

        this._N = this._words.size();

    }

    /**
     * Get the Loglikelihood
     * @return the log-likelhood value
     */
    public double getLogLikelihood() {
        return Math.log(getLikelihood());
    }

    /**
     * Get the likelihood
     * @return the likelihood value
     */
    public double getLikelihood() {
        return this._beta[0][this._N][0];
    }

    /**
     * Check whether this parse instance should be ignored.
     * @return ignore flag
     */
    public boolean ignore() {
        return this._ignoreFlag;
    }

    public ArrayList<GProduction> getAllProds() {
        return this._allGProds;
    }

    protected int _K = 1; //default..

    public abstract void parseMax(int K);

    public static UGParseTable createUGTable(TestExample example, Param parameter) {
        return new UGParseTable(example, parameter);
    }


    public static BGParseTable createBGTable(TestExample example, Param parameter) {
        return new BGParseTable(example, parameter);
    }

    public static MBGParseTable createMBGTable(TestExample example, Param parameter) {
        return new MBGParseTable(example, parameter);
    }

    public static MGParseTable createMGTable(TestExample example, Param parameter_unigram, Param parameter_bigram) {
        return new MGParseTable(example, parameter_unigram, parameter_bigram);
    }

    public ArrayList<Production> restoreMaxMRL_Kbest() {
        ArrayList<Production> roots = new ArrayList<Production>();
        for (int rank = 0; rank < this._final_list.size(); rank++) {
            ResElem result = this._final_list.get(rank);
            if (result.getProb() > 0) {
                Production root = this.recoverProd(result, null);
                roots.add(0, root);
            }
        }
        return roots;
    }

    private Production recoverProd(ResElem res, Production parent) {
        GProduction gp = res.getGP();
        Production prod = new Production(gp);

        if (parent != null) {
            ArrayList<Production> childrenOfPar = parent.getChildren();
            childrenOfPar.add(prod);
            prod.setParent(parent);
        }

        ResElem[] children = res.getChildren();

        for (int i = 0; i < children.length; i++) {
            ResElem child = children[i];
            this.recoverProd(child, prod);
        }

        return prod;
    }

    public CandidateSet getTopCandidates() {
        ArrayList<Production> roots = this.restoreMaxMRL_Kbest();
        CandidateSet candidates = new CandidateSet(this._exampleId);
        ArrayList<ProdHybridNode> rootNodes = this.getOptimalHybridTrees();
        if (!this._ignoreFlag) {
            for (int rank = 0; rank < rootNodes.size(); rank++) {
                Candidate candidate = Candidate.createCandidate(candidates, rootNodes.get(rank), rank);
                 {
                    boolean isCorrect = this._example.verifyCorrect(roots.get(rank).getAllChildren());
                    if (isCorrect) {
                        candidate.setCorrect();
                    }
                }
            }
        }
        return candidates;
    }

    protected ResElemList _final_list;

    public ArrayList<ProdHybridNode> getOptimalHybridTrees() {
        ArrayList<ProdHybridNode> trees = new ArrayList<ProdHybridNode>();
        if (this._ignoreFlag) {
            return trees;
        }
        for (int i = 0; i < this._final_list.size(); i++) {
            ResElem rootElem = this._final_list.get(i);
            if (rootElem.getProb() > 0) {
                ProdHybridNode rootNode = this.convertToHybridNode(rootElem, 0, this._words);
                trees.add(0, rootNode);
            }
        }
        return trees;
    }

    private ProdHybridNode convertToHybridNode(ResElem elem, int cIndex, ArrayList<Word> allwords) {
        ProdHybridNode node = HybridNode.createProdHybridNode(elem, cIndex, allwords);
        ResElem[] childrenElems = elem.getChildren();
        int bIndex = elem.getBIndex();
        int eIndex = elem.getEIndex();
        ArrayList<HybridNode> children = new ArrayList<HybridNode>();

        if (childrenElems.length == 0) {
            for (int wIndex = bIndex; wIndex < eIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                HybridNode cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
        } else if (childrenElems.length == 1) {
            ResElem childElem = childrenElems[0];
            HybridNode cNode;

            int cBIndex, cEIndex;
            cBIndex = bIndex;
            cEIndex = childElem.getBIndex();
            for (int wIndex = cBIndex; wIndex < cEIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
            cNode = convertToHybridNode(childElem, 0, allwords);
            children.add(cNode);
            cNode.setParent(node);

            cBIndex = childElem.getEIndex();
            cEIndex = eIndex;
            for (int wIndex = cBIndex; wIndex < cEIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
        } else if (childrenElems.length == 2) {
            ResElem childElem1, childElem2;
            int childCIndex1, childCIndex2;

            if (!Rule.isReversed(elem.getRulePattern())) {
                childElem1 = childrenElems[0];
                childElem2 = childrenElems[1];
                childCIndex1 = 0;
                childCIndex2 = 1;
            } else {
                childElem1 = childrenElems[1];
                childElem2 = childrenElems[0];
                childCIndex1 = 1;
                childCIndex2 = 0;
            }
            HybridNode cNode;

            int cBIndex, cEIndex;
            cBIndex = bIndex;
            cEIndex = childElem1.getBIndex();
            for (int wIndex = cBIndex; wIndex < cEIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
            cNode = convertToHybridNode(childElem1, childCIndex1, allwords);
            children.add(cNode);
            cNode.setParent(node);

            cBIndex = childElem1.getEIndex();
            cEIndex = childElem2.getBIndex();
            for (int wIndex = cBIndex; wIndex < cEIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
            cNode = convertToHybridNode(childElem2, childCIndex2, allwords);
            children.add(cNode);
            cNode.setParent(node);

            cBIndex = childElem2.getEIndex();
            cEIndex = eIndex;
            for (int wIndex = cBIndex; wIndex < cEIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
        }
        node.setChildren(children);

        return node;
    }

}
