/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package hybridtree.core.perceptron;

/**
 * @author Wei LU
 * @version 1.0
 */
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Collections;

import java.io.Serializable;

public class PFeature implements Serializable, Comparable<PFeature> {

	private static final long serialVersionUID = -7543302050489416032L;

	private int _fId = -1; // the feature id

    private int[] _feature; //perceptron feature representation

    private String[] _frep;

    private double _weight; // the weight associated to this feature

    private double _avgWeight; // the old weight

    private boolean _disabled = false;

    public void disable() {
        this._disabled = true;
    }

    public void enable(){
        this._disabled = false;
    }

    public boolean isDisabled() {
        return this._disabled;
    }

    private boolean _appearedOnce = true;

    //private int _appearedCount = 0;

    /**
     * Create a new PFeature, given a string representation and a list of all available features
     * If it appeared to be a already created feature, the old feature will get returned, otherwise
     * a new PFeature will be created.
     * @param feature String
     * @param allFeatures ArrayList
     * @return PFeature
     */
    public static PFeature createFeature(int[] feature, String[] frep, ArrayList<PFeature> allFeatures, boolean disallowMoreFeatures) {
        PFeature pf = new PFeature(feature, frep);
        @SuppressWarnings({ "unchecked", "rawtypes" })
		int index = Collections.binarySearch((List) allFeatures, pf);
        if (index < 0) {
            if (!disallowMoreFeatures) {
                pf.disable();
                System.out.println("New Feature:" + pf.toString());
            }
            index = -index - 1;
//            pf.setFeatureId(allFeatures.size());
            allFeatures.add(index, pf);
            return pf;
        } else {
            PFeature f = allFeatures.get(index);
//            f.increaseAppeaedCount();
            f.setAppearedMoreThanOnce();
            return f;
        }
    }

    /*
    public int getAppearedCount(){
        return _appearedCount;
    }
    */

    public boolean appearedOnceOnly(){
        return this._appearedOnce;
    }

    public void setAppearedMoreThanOnce(){
        this._appearedOnce = false;
    }

    /*
    public void increaseAppeaedCount(){
        this._appearedCount++;
    }
    */

    public int compareTo(PFeature pf) {
        if (this._feature[0] > pf._feature[0]) {
            return +1;
        } else if (this._feature[0] < pf._feature[0]) {
            return -1;
        } else {

            if (this._feature.length > pf._feature.length) {
                return +1;
            } else if (this._feature.length < pf._feature.length) {
                return -1;
            }

            for (int i = 1; i < this._feature.length; i++) {
                if (this._feature[i] > pf._feature[i]) {
                    return +1;
                } else if (this._feature[i] < pf._feature[i]) {
                    return -1;
                }
            }
            return 0;
        }
    }

    protected void normWeight(double norm) {
        this._weight /= norm;
    }

    /**
     * Two PFeatures are identical if they share the same string representation
     * @param o Object
     * @return boolean
     */
    public boolean equals(Object o) {
        if (o instanceof PFeature) {
            PFeature f = (PFeature) o;
            return Arrays.equals(f._feature, this._feature);
        }
        return false;
    }

    /**
     * Constructor
     * @param feature String
     */
    private PFeature(int[] feature, String[] frep) {
        this._feature = feature;
        this._frep = frep;
        this._weight = 0;
        this._avgWeight = 0;
    }

    /**
     * Set the feature id
     * @param fId int
     */
    public void setFeatureId(int fId) {
        this._fId = fId;
    }

    /**
     * Get the feature id
     * @return int
     */
    public int getFeatureId() {
        return this._fId;
    }

    /**
     * Get the representation of this PFeature
     * @return String
     */
    public int[] getFeatureRep() {
        return this._feature;
    }

    protected void weight_average(int i, int n, int T) {
        this._avgWeight = (this._avgWeight * (double) (i + n * T) + this._weight) / (double) ((i + 1) + n * T);
    }

    public void swap() {
        this._weight = this._avgWeight;
    }

    protected void weight_minus(double diff) {
        this._weight -= diff;
    }

    protected void weight_add(double diff) {
        this._weight += diff;
    }

    /**
     * Get the weight for this feature
     * @return double
     */
    public double getWeight() {
        return this._weight;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("PFeature:<");
        for (int i = 0; i < this._frep.length; i++) {
            if (i != 0) {
                sb.append(',');
            }
            sb.append(this._frep[i]);
        }
        sb.append('>');
        sb.append(':');
        sb.append(this._weight);

        return sb.toString();
    }
}
