/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package hybridtree.core.perceptron;

/**
 * @author Wei LU
 * @version 1.0
 */
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;

import hybridtree.core.*;
import hybridtree.data.*;
import hybridtree.util.Pair;

import java.io.*;

public class Perceptron implements Serializable {
	
	private static final long serialVersionUID = 6138598777764248172L;

	private ArrayList<PFeature> _allFeatures; // all the features

    private int _numInstances; // the number of instances.

    private double _bias; // the b

    private double _margin; // the margin

    public Perceptron(int numInstances) {
        this._allFeatures = new ArrayList<PFeature>();
        this._numInstances = numInstances;
    }

	public void sortAndIndexFeatures() {
        Collections.sort(this._allFeatures);
        for (int i = 0; i < this._allFeatures.size(); i++) {
            PFeature feature = this._allFeatures.get(i);
            feature.setFeatureId(i + 1);
        }
    }

    public FVPair createFVPair(int[] feature, String[] frep, double value) {
        PFeature f = PFeature.createFeature(feature, frep, this._allFeatures,
                                            this._allowMoreFeatures);
        FVPair fv = new FVPair(f, value);
        return fv;
    }

    private String diff(ArrayList<FVPair> fv1, ArrayList<FVPair> fv2) {
        for (int i = 0; i < fv1.size(); i++) {
            FVPair f = fv1.get(i);
            if (fv2.contains(f)) {
                fv2.remove(f);
                fv1.remove(f);
                i--;
            }
        }
        for (int i = 0; i < fv2.size(); i++) {
            FVPair f = fv2.get(i);
            if (fv1.contains(f)) {
                fv1.remove(f);
                fv2.remove(f);
                i--;
            }
        }

        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < fv1.size(); i++) {
            FVPair f = fv1.get(i);
            sb.append('<');
            sb.append(f.toString());
            sb.append('\n');
        }
        sb.append('\n');
        for (int i = 0; i < fv2.size(); i++) {
            FVPair f = fv2.get(i);
            sb.append('>');
            sb.append(f.toString());
            sb.append('\n');
        }
        return sb.toString();
    }

    /**
     * compute the cost associated with the whole feature ArrayList
     * @return double
     */
    public double computeFeatureCost(Candidate candidate, double bias,
                                     double margin) {
        return (candidate.computeFeatureCost() + bias);// / margin;
    }

    /**
     * Pick the best Candidate
     * @param canSet CandidateSet
     * @return Candidate
     */
    public ArrayList<Candidate> perceptPick(CandidateSet canSet) {
        ArrayList<Candidate> candidates = canSet.getCandidates();
        double maxCost = Double.NEGATIVE_INFINITY;

        ArrayList<Candidate> allMaxCandidates = new ArrayList<Candidate>();
        Candidate rank0 = candidates.get(0);
        double rank0Cost = rank0.computeFeatureCost();
        //this.computeFeatureCost(rank0, this._bias,
                //this._margin);
        for (int i = 0; i < candidates.size(); i++) {
            Candidate candidate = candidates.get(i);
//            double featureCost = this.computeFeatureCost(candidate, this._bias,
//                    this._margin);
            double featureCost = candidate.computeFeatureCost();
            if (featureCost >= maxCost) {
                maxCost = featureCost;
                allMaxCandidates = new ArrayList<Candidate>();
                allMaxCandidates.add(candidate);
            }
        }

        //should not comment out this line!!
        if (this.computeFeatureCost(allMaxCandidates.get(0), this._bias,
                                    this._margin) < 0) {
            return new ArrayList<Candidate>();
        }

        if (rank0.isCorrect() && !allMaxCandidates.get(0).isCorrect()) {
            System.out.println("BAD!!");
            System.out.println("Old:(correct)");
            System.out.println(rank0Cost);
            System.out.println(rank0.getRootNode().toString());
            ArrayList<FVPair> corrFeatures = rank0.getFeatures(this);
            for (int k = 0; k < allMaxCandidates.size(); k++) {
                Candidate maxCandidate = allMaxCandidates.get(k);
                System.out.println("New:" + maxCandidate.isCorrect());
                System.out.println(maxCost);
                System.out.println(maxCandidate.getRootNode().toString());
                System.out.println("Features:");
                ArrayList<FVPair> features = maxCandidate.getFeatures(this);
                System.out.println(this.diff(corrFeatures, features));
            }
        } else if (!rank0.isCorrect() && allMaxCandidates.get(0).isCorrect()) {
            System.out.println("GREAT!!");
            System.out.println("Old:(wrong)");
            System.out.println(rank0Cost + this._bias);
            System.out.println(rank0.getRootNode().toString());
            ArrayList<FVPair> corrFeatures = rank0.getFeatures(this);
            for (int k = 0; k < allMaxCandidates.size(); k++) {
                Candidate maxCandidate = allMaxCandidates.get(k);
                System.out.println("New:" + maxCandidate.isCorrect());
                System.out.println(maxCost);
                System.out.println(maxCandidate.getRootNode().toString());
                System.out.println("Features:");
                ArrayList<FVPair> features = maxCandidate.getFeatures(this);
                System.out.println(this.diff(corrFeatures, features));
            }
        } else {

            Candidate corrCandidate = null;
            for (int i = 0; i < canSet.size(); i++) {
                Candidate candidate = canSet.getCandidateAt(i);
                if (candidate.isCorrect()) {
                    corrCandidate = candidate;
                }
            }
            if (corrCandidate != null && !allMaxCandidates.get(0).isCorrect()) {
                System.out.println("CAN BE BETTER!!");
                System.out.println("Old:(wrong)");
                System.out.println(rank0.computeFeatureCost() + "\t" +
                                   rank0Cost);
                System.out.println(rank0.getRootNode().toString());
                System.out.println("At Best:");
                System.out.println(corrCandidate.computeFeatureCost() + "\t" +
                                   this.
                                   computeFeatureCost(corrCandidate, this._bias,
                        this._margin));
                System.out.println(corrCandidate.getRootNode().toString());
                ArrayList<FVPair> corrFeatures = corrCandidate.getFeatures(this);
                for (int k = 0; k < allMaxCandidates.size(); k++) {
                    Candidate maxCandidate = allMaxCandidates.get(k);
                    System.out.println("New:" + maxCandidate.isCorrect());
                    System.out.println(maxCandidate.computeFeatureCost() + "\t" +
                                       maxCost);
                    System.out.println(maxCandidate.getRootNode().toString());
                    System.out.println("Features:");
                    ArrayList<FVPair> features = maxCandidate.getFeatures(this);
                    System.out.println(this.diff(corrFeatures, features));
                }
            } else {

                ArrayList<FVPair> corrFeatures = rank0.getFeatures(this);
                for (int k = 0; k < allMaxCandidates.size(); k++) {
                    Candidate maxCandidate = allMaxCandidates.get(k);
                    if (!maxCandidate.isCorrect()) {
                        System.out.println("FAILED!!");
                        System.out.println("Old:(wrong)");
                        System.out.println(rank0Cost);
                        System.out.println(rank0.getRootNode().toString());
                    } else {
                        System.out.println("OK!!");
                        System.out.println("Old:(already correct)");
                        System.out.println(rank0Cost);
                        System.out.println(rank0.getRootNode().toString());
                    }
                    System.out.println("New:" + maxCandidate.isCorrect());
                    System.out.println(maxCost);
                    System.out.println(maxCandidate.getRootNode().toString());
                    System.out.println("Features:");
                    ArrayList<FVPair> features = maxCandidate.getFeatures(this);
                    System.out.println(this.diff(corrFeatures, features));
                }
            }
        }
        return allMaxCandidates;
    }

    public String toFeatureString(Candidate candidate) {
        StringBuffer sb = new StringBuffer();
        ArrayList<FVPair> fvs = candidate.getFeatures(this);
        for (int i = 0; i < fvs.size(); i++) {
            PFeature pf = fvs.get(i).getFeature();
            if (pf.getWeight() != 0) {
                sb.append(fvs.get(i).getFeature() + "\t" + fvs.get(i).getValue());
                sb.append('\n');
            }
        }
        return sb.toString();
    }

    public StringBuffer toSVMTestFile(CandidateSet canSet, int qid) {

        StringBuffer sb = new StringBuffer();

        ArrayList<Candidate> candidates = canSet.getCandidates();

        for (int i = 0; i < candidates.size(); i++) {
            Candidate candidate = candidates.get(i);
            ArrayList<FVPair> fvs = candidate.getFeatures(this);
            if (candidate.isCorrect()) {
                sb.append('2');
            } else {
                sb.append('1');
            }
            sb.append(' ');
            sb.append("qid:");
            sb.append(qid);
            sb.append(' ');
            for (int k = 0; k < fvs.size(); k++) {
                FVPair fv = fvs.get(k);
                int featureId = fv.getFeature().getFeatureId();
                if (featureId != -1) {
                    sb.append(fv.getFeature().getFeatureId());
                    sb.append(':');
                    sb.append(fv.getValue());
                    sb.append(' ');
                }
            }
            sb.append('\n');
        }

        return sb;

    }


    public StringBuffer toSVMTrainFile(CandidateSet canSet, int qid) {

        StringBuffer sb = new StringBuffer();

        ArrayList<Candidate> candidates = canSet.getCandidates();

        Candidate corrCandidate = null;

        for (int i = 0; i < candidates.size(); i++) {
            Candidate candidate = candidates.get(i);
            if (corrCandidate == null && candidate.isCorrect()) {
                corrCandidate = candidate;
                candidates.remove(i);
                i--;
            }
        }

        CandidateSet corrCandidates = canSet.getCorrectCandidates();

        if (corrCandidate == null) {
            corrCandidate = corrCandidates.getCandidateAt(0);
        }

        sb.append('2');
        sb.append(' ');
        sb.append("qid:");
        sb.append(qid);
        sb.append(' ');
        ArrayList<FVPair> fvs = corrCandidate.getFeatures(this);
        for (int i = 0; i < fvs.size(); i++) {
            FVPair fv = fvs.get(i);
            sb.append(fv.getFeature().getFeatureId());
            sb.append(':');
            sb.append(fv.getValue());
            sb.append(' ');
        }
        sb.append('\n');

        for (int i = 0; i < candidates.size(); i++) {
            Candidate candidate = candidates.get(i);
            fvs = candidate.getFeatures(this);
            sb.append('1');
            sb.append(' ');
            sb.append("qid:");
            sb.append(qid);
            sb.append(' ');
            for (int k = 0; k < fvs.size(); k++) {
                FVPair fv = fvs.get(k);
                sb.append(fv.getFeature().getFeatureId());
                sb.append(':');
                sb.append(fv.getValue());
                sb.append(' ');
            }
            sb.append('\n');
        }

        return sb;

    }


    /**
     * Learn perceptron
     * @param canSet CandidateSet
     * @return boolean
     */
    public boolean perceptLearn(CandidateSet canSet) {

        ArrayList<Candidate> candidates = canSet.getCandidates();

        double maxCost = Double.NEGATIVE_INFINITY;
        Candidate maxCandidate = null;

        for (int i = 0; i < candidates.size(); i++) {
            Candidate candidate = candidates.get(i);
            double featureCost = candidate.computeFeatureCost();
            if (featureCost > maxCost) {
                maxCost = featureCost;
                maxCandidate = candidate;
            }
        }

        //the corrCandidate is the correct Candidate
        CandidateSet corrCandidates = canSet.getCorrectCandidates();

        System.err.println("Update!");

        System.err.println("maxProb=" + maxCandidate.getProb());
        System.err.println("corProb=" +
                           corrCandidates.getCandidateAt(0).getProb());
        if (maxCandidate.getProb() == corrCandidates.getCandidateAt(0).getProb()) {
            System.err.println("Identical!");
        } else {
            System.out.println("Correct:" +
                               corrCandidates.getCandidateAt(0).isCorrect());
            System.out.println(corrCandidates.getCandidateAt(0).getRootNode().
                               toString());
            System.out.println("Max:" + maxCandidate.isCorrect());
            System.out.println(maxCandidate.getRootNode().toString());
            System.out.println();
            System.out.println();
        }

        this.positive(corrCandidates.getCandidateAt(0), 1.0);
        this.negative(maxCandidate, 1.0);

        return true;

    }

    public void normalizeWeights() {
        double norm = 0;
        for (int i = 0; i < this._allFeatures.size(); i++) {
            PFeature feature = this._allFeatures.get(i);
            norm += Math.pow(feature.getWeight(), 2);
        }
        norm = Math.sqrt(norm);
        for (int i = 0; i < this._allFeatures.size(); i++) {
            PFeature feature = this._allFeatures.get(i);
            feature.normWeight(norm);
        }
    }

    public void setBias(double bias) {
        this._bias = bias;
    }

    public void setMargin(double margin) {
        this._margin = margin;
    }

    public void computeBiasAndMargin(Candidate badCand, Candidate goodCand) {
        double badProjection = badCand.computeFeatureCost();
        double goodProjection = goodCand.computeFeatureCost();
        this._bias = -(goodProjection + badProjection) / 2.0;
        this._margin = Math.abs(goodProjection - badProjection) / 2.0;

        System.err.println("The bias=" + this._bias);
        System.err.println("The margin=" + this._margin);
    }

    /**
     * Do Average. for Average perceptron..
     */
    public void perceptAvg(int index, int T) {
        for (int i = 0; i < this._allFeatures.size(); i++) {
            PFeature feature = this._allFeatures.get(i);
            if (!feature.isDisabled()) {
                feature.weight_average(index, this._numInstances, T);
            }
        }
    }

    public void perceptSwap() {
        for (int i = 0; i < this._allFeatures.size(); i++) {
            PFeature feature = this._allFeatures.get(i);
            if (!feature.isDisabled()) {
                feature.swap();
            }
        }
    }

    private void positive(Candidate candidate, double ratio) {
        ArrayList<FVPair> fpairs = candidate.getFeatures(this);
        for (int i = 0; i < fpairs.size(); i++) {
            FVPair pair = fpairs.get(i);
            PFeature feature = pair.getFeature();
            if (!feature.isDisabled()) {
                feature.weight_add(pair.getValue() * ratio);
            }
        }
    }

    private void negative(Candidate candidate, double ratio) {
        ArrayList<FVPair> fpairs = candidate.getFeatures(this);
        for (int i = 0; i < fpairs.size(); i++) {
            FVPair pair = fpairs.get(i);
            PFeature feature = pair.getFeature();
            if (!feature.isDisabled()) {
                feature.weight_minus(pair.getValue() * ratio);
            }
        }
    }

    private void addFV(ArrayList<FVPair> fv, FVPair f) {
        @SuppressWarnings({ "unchecked", "rawtypes" })
		int index = Collections.binarySearch((List) fv, f);
        if (index < 0) {
            index = -index - 1;
            fv.add(index, f);
        }
    }

    private boolean _allowMoreFeatures = true;

    public void disallowMoreFeatures() {
        this._allowMoreFeatures = false;
    }

    public boolean allowMoreFeatures() {
        return this._allowMoreFeatures;
    }

    public ArrayList<FVPair> extractFeatures(Candidate candidate) {
    	ArrayList<FVPair> fv = new ArrayList<FVPair>();

        //log-probability
        int[] f_rep = {0};
        String[] frep = {"probability"};
        double f_val = Math.log(candidate.getProb());
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        f.getFeature().enable();
        fv.add(f);

        this.extractFeatures(candidate.getRootNode(), fv);

        return fv;
    }

    private void extractFeatures(ProdHybridNode node, ArrayList<FVPair> fv) {

        this.extractRuleFeatures(node, fv, 13);

        this.extractExpandedRuleFeatures(node, fv, 25);

        this.extractRangeInFeatures(node, fv, 11);

        this.extract2MRL1NLFeatures(node, fv, 44);

//        this.extractSubtreePhraseFeatures(node, fv, 100);

//        this.extractBigramFeatures(node, fv, 66);

//        this.extractBigramFeatures(node, fv, 55);

//        this.extractRelativeMRLFeatures(node, fv, 23);
    }

    public void extractSubtreeFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                       int featureType) {

        GProduction gp = node.getElem().getGP();
        if (gp.isConstant()) {
            return;
        }
        /*
         if (gp.isUnum() || gp.isNum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */

        ArrayList<HybridNode> childrenNodes = node.getChildren();
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractSubtreeFeatures(pChildNode, fv, featureType);
            }
        }

        Pair<ArrayList<Integer>,
                ArrayList<String>>
                subtreePairs = this.extractSubtreeFeaturesHelper(node);
        ArrayList<Integer> subtree_int = subtreePairs.getFirst();
        ArrayList<String> subtree_str = subtreePairs.getSecond();

        int[] f_rep = new int[subtree_int.size() + 1];
        String[] frep = new String[subtree_str.size() + 1];

        f_rep[0] = featureType;
        frep[0] = "Subtree";

        for (int i = 0; i < subtree_int.size(); i++) {
            f_rep[i + 1] = subtree_int.get(i);
            frep[i + 1] = subtree_str.get(i);
        }

        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);

    }

    private Pair<ArrayList<Integer>,
            ArrayList<String>> extractSubtreeFeaturesHelper(ProdHybridNode node) {

        GProduction gp = node.getElem().getGP();
        ArrayList<Integer> result_int = new ArrayList<Integer>();
        ArrayList<String> result_str = new ArrayList<String>();

        if (gp.isUnum()) {
            result_int.add( -100);
            result_str.add("UNUM");
            return new Pair<ArrayList<Integer>,
                    ArrayList<String>>(result_int, result_str);
        } else if (gp.isNum()) {
            result_int.add( -200);
            result_str.add("NUM");
            return new Pair<ArrayList<Integer>,
                    ArrayList<String>>(result_int, result_str);
        } else if (gp.isIdent()) {
            result_int.add( -300);
            result_str.add("IDENT");
            return new Pair<ArrayList<Integer>,
                    ArrayList<String>>(result_int, result_str);
        } else if (gp.isConstant()) {
            result_int.add( -400);
            result_str.add("CONST");
            return new Pair<ArrayList<Integer>,
                    ArrayList<String>>(result_int, result_str);
        }

        Pair<int[], String[]> hybridForm = this.getHybridForm(node);

        int[] curr_int = hybridForm.getFirst();
        String[] curr_str = hybridForm.getSecond();

        for (int i = 0; i < curr_int.length; i++) {
            result_int.add(curr_int[i]);
            result_str.add(curr_str[i]);
        }

        result_int.add( -100000);
        result_str.add("BEGIN");

        ArrayList<HybridNode> childrenNodes = node.getChildren();
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                Pair<ArrayList<Integer>,
                        ArrayList<String>>
                        childResult = this.extractSubtreeFeaturesHelper(
                                pChildNode);
                ArrayList<Integer> child_int = childResult.getFirst();
                ArrayList<String> child_str = childResult.getSecond();

                for (int j = 0; j < child_int.size(); j++) {
                    result_int.add(child_int.get(j));
                    result_str.add(child_str.get(j));
                }
            }
        }

        result_int.add( -200000);
        result_str.add("END");

        return new Pair<ArrayList<Integer>, ArrayList<String>>(result_int, result_str);
    }

    private Pair<int[], String[]> getHybridForm(ProdHybridNode node) {
        GProduction gp = node.getElem().getGP();

        ArrayList<HybridNode> childrenNodes = node.getChildren();
        int[] result = new int[childrenNodes.size() + 1];
        String[] resultStr = new String[childrenNodes.size() + 1];

        result[0] = gp.getProdId();
        resultStr[0] = gp.toLine();
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            Word word = childNode.toWord();
            if (this.isIdent(word)) {
                result[i + 1] = -1;
                resultStr[i + 1] = "IDENT";
            } else if (this.isNumber(word)) {
                result[i + 1] = -2;
                resultStr[i + 1] = "NUMBER";
            } else {
                result[i + 1] = word.getWordId();
                resultStr[i + 1] = word.getWordStr();
            }
        }
        return new Pair<int[], String[]>(result, resultStr);
    }

    @SuppressWarnings("unused")
	private void extractTwoLevelRulesFeatures(ProdHybridNode node,
                                              ArrayList<FVPair> fv,
                                              int featureType) {

        ProdHybridNode parentNode = ((ProdHybridNode) node.getParent());
        Pair<int[], String[]> parentHybridFormPair;
        int[] parentHybridForm;
        String[] parentHybridFormStr;
        if (parentNode != null) {
            parentHybridFormPair = this.getHybridForm(parentNode);
            parentHybridForm = parentHybridFormPair.getFirst();
            parentHybridFormStr = parentHybridFormPair.getSecond();
        } else {
            parentHybridForm = new int[] { -100};
            parentHybridFormStr = new String[] {"NULL"};
        }

        Pair<int[], String[]> hybridFormPair = this.getHybridForm(node);
        int[] hybridForm = hybridFormPair.getFirst();
        String[] hybridFormStr = hybridFormPair.getSecond();

        int[] f_rep = new int[3 + hybridForm.length + parentHybridForm.length];
        String[] frep = new String[3 + hybridFormStr.length +
                        parentHybridFormStr.length];

        f_rep[0] = featureType;
        frep[0] = "TwoLevelRule";

        for (int i = 0; i < parentHybridForm.length; i++) {
            f_rep[i + 1] = parentHybridForm[i];
            frep[i + 1] = parentHybridFormStr[i];
        }

        f_rep[parentHybridForm.length + 1] = -10000;
        f_rep[parentHybridForm.length + 2] = node.toWord().getWordId();

        frep[parentHybridForm.length + 1] = "END OF PARENT";
        frep[parentHybridForm.length + 2] = node.toWord().getWordStr();

        for (int i = 0; i < hybridForm.length; i++) {
            f_rep[i + parentHybridForm.length + 3] = hybridForm[i];
            frep[i + parentHybridFormStr.length + 3] = hybridFormStr[i];
        }

        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);

        ArrayList<HybridNode> childrenNodes = node.getChildren();
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractTwoLevelRulesFeatures(pChildNode, fv, featureType);
            }
        }
    }

    @SuppressWarnings("unused")
	private void extractExpandedBigramsFeatures(ProdHybridNode node,
                                                ArrayList<FVPair> fv,
                                                int featureType) {

        GProduction gp = node.getElem().getGP();

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        Word word1 = Word.BEGIN;
        Word word;

        int word1Type, wordType;
        int word1Id, wordId;
        String word1Str, wordStr;

        word1Type = 0;
        word1Id = word1.getWordId();
        word1Str = word1.getWordStr();

        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            word = childNode.toWord();

            wordType = childNode instanceof WordHybridNode ? 0 : 1;
            wordId = childNode instanceof WordHybridNode ?
                     ((WordHybridNode) childNode).getWord().getWordId() :
                     ((ProdHybridNode) childNode).getElem().getGP().getProdId();
            wordStr = childNode instanceof WordHybridNode ?
                      ((WordHybridNode) childNode).getWord().getWordStr() :
                      ((ProdHybridNode) childNode).getElem().getGP().toLine();

            int[] f_rep = new int[] {featureType, word1Type, wordType, word1Id,
                          wordId};
            String[] frep = new String[] {"ExpandedBigram", word1Str, wordStr};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractExpandedBigramsFeatures(pChildNode, fv, featureType);
            }
            word1 = word;
            word1Type = wordType;
            word1Str = wordStr;
            word1Type = wordType;
        }

        word = Word.END;
        wordType = 0;
        wordId = word.getWordId();
        int[] f_rep = new int[] {featureType, word1Type, wordType, word1Id,
                      wordId};
        String[] frep = new String[] {"ExpandedBigram", gp.toLine(),
                        word1.getWordStr(), word.getWordStr()};
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);

    }

    @SuppressWarnings("unused")
	private void extractGrandParentRuleFeatures(ProdHybridNode node,
                                                ArrayList<FVPair> fv,
                                                int featureType) {

        GProduction gp = node.getElem().getGP();
        if (gp.isConstant()) {
            return;
        }
        /*
         if (gp.isConst() || gp.isNum() || gp.isUnum() || gp.isIdent()) {
            return;
                 }
         */

        ProdHybridNode parentNode = (ProdHybridNode) node.getParent();
        GProduction parent;
        if (parentNode != null) {
            parent = parentNode.getElem().getGP();
        } else {
            parent = null;
        }

        int pIndex = gp.getProdId();

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        int[] f_rep = new int[3 + childrenNodes.size()];
        String[] frep = new String[3 + childrenNodes.size()];

        f_rep[0] = featureType;
        frep[0] = "GrandParentRule";
        if (parent != null) {
            f_rep[1] = parent.getProdId();
            frep[1] = parent.toLine();
        } else {
            f_rep[1] = -1;
            frep[1] = "NULL";
        }
        f_rep[2] = pIndex;
        frep[2] = gp.toLine();

        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            f_rep[i + 3] = childNode.toWord().getWordId();
            frep[i + 3] = childNode.toWord().getWordStr();
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractGrandParentRuleFeatures(pChildNode, fv, featureType);
            }
        }
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);
    }

    @SuppressWarnings("unused")
	private void extractGrandParentBigramFeatures(ProdHybridNode node,
                                                  ArrayList<FVPair> fv,
                                                  int featureType) {

        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isIdent() || gp.isNum() || gp.isUnum() || gp.isConst()) {
            return;
                 }
         */

        int pIndex = gp.getProdId();
        ArrayList<HybridNode> childrenNodes = node.getChildren();

        GProduction parent;
        int ppIndex;
        ProdHybridNode parentNode = (ProdHybridNode) node.getParent();
        if (parentNode != null) {
            parent = parentNode.getElem().getGP();
            ppIndex = parent.getProdId();
        } else {
            parent = null;
            ppIndex = -1;
        }

        Word word1 = Word.BEGIN;
        Word word;
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            word = childNode.toWord();
            int[] f_rep;
            String[] frep;

            f_rep = new int[] {featureType, ppIndex, pIndex, word1.getWordId(),
                    word.getWordId()};
            if (parent != null) {
                frep = new String[] {"GrandParentBigram", parent.toLine(),
                       gp.toLine(), word1.getWordStr(), word.getWordStr()};
            } else {
                frep = new String[] {"GrandParentBigram", "NULL", gp.toLine(),
                       word1.getWordStr(), word.getWordStr()};
            }
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractGrandParentBigramFeatures(pChildNode, fv,
                        featureType);
            }
            word1 = word;
        }

        word = Word.END;
        int[] f_rep;
        String[] frep;

        f_rep = new int[] {featureType, ppIndex, pIndex, word1.getWordId(),
                word.getWordId()};
        if (parent != null) {
            frep = new String[] {"GrandParentBigram", parent.toLine(),
                   gp.toLine(), word1.getWordStr(), word.getWordStr()};
        } else {
            frep = new String[] {"GrandParentBigram", "NULL", gp.toLine(),
                   word1.getWordStr(), word.getWordStr()};
        }
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);

    }


    @SuppressWarnings("unused")
	private void extractExpandedBigramFeatures(ProdHybridNode node,
                                               ArrayList<FVPair> fv,
                                               int featureType) {

        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isIdent() || gp.isNum() || gp.isUnum() || gp.isConst()) {
            return;
                 }
         */

        int pIndex = gp.getProdId();
        ArrayList<HybridNode> childrenNodes = node.getChildren();

        Word word1 = Word.BEGIN;
        Word word;
        int word1Index = word1.getWordId();
        int wordIndex;
        String word1Str = word1.getWordStr();
        String wordStr;
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof WordHybridNode) {
                word = childNode.toWord();
                wordIndex = word.getWordId();
                wordStr = word.getWordStr();
            } else {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                GProduction pChild = pChildNode.getElem().getGP();
                if (pChild.isIdent()) {
                    wordIndex = -2;
                    wordStr = "IDENT";
                } else if (pChild.isNum()) {
                    wordIndex = -3;
                    wordStr = "NUM";
                } else if (pChild.isUnum()) {
                    wordIndex = -4;
                    wordStr = "UNUM";
                } else if (pChild.isConstant()) {
                    wordIndex = -1;
                    wordStr = "CONSTANT";
                } else {
                    wordIndex = pChildNode.getElem().getGP().getProdId() +
                                10000;
                    wordStr = pChildNode.getElem().getGP().toLine();
                }
            }
            int[] f_rep = {featureType, pIndex, word1Index, wordIndex};
            String[] frep = {"ExpandedBigram", gp.toLine(), word1Str, wordStr};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractExpandedBigramFeatures(pChildNode, fv, featureType);
            }
            word1Index = wordIndex;
            word1Str = wordStr;
        }

        word = Word.END;
        int[] f_rep = new int[] {featureType, pIndex, word1Index,
                      word.getWordId()};
        String[] frep = new String[] {"ExpandedBigram", gp.toLine(), word1Str,
                        word.getWordStr()};
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);
    }


    @SuppressWarnings("unused")
	private void extractBigramFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                       int featureType) {

        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isIdent() || gp.isNum() || gp.isUnum() || gp.isConst()) {
            return;
                 }
         */

        int pIndex = gp.getProdId();
        ArrayList<HybridNode> childrenNodes = node.getChildren();

        Word word1 = Word.BEGIN;
        Word word;
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            word = childNode.toWord();
            int[] f_rep = new int[] {featureType, pIndex, word1.getWordId(),
                          word.getWordId()};
            String[] frep = new String[] {"Bigram", gp.toLine(),
                            word1.getWordStr(), word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractBigramFeatures(pChildNode, fv, featureType);
            }
            word1 = word;
        }

        word = Word.END;
        int[] f_rep = new int[] {featureType, pIndex, word1.getWordId(),
                      word.getWordId()};
        String[] frep = new String[] {"Bigram", gp.toLine(), word1.getWordStr(),
                        word.getWordStr()};
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);
    }

    @SuppressWarnings("unused")
	private void extractMRLRuleFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                        int featureType) {

        GProduction gp = node.getElem().getGP();
        int pIndex = gp.getProdId();

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        int ppIndex;
        String ppStr;
        ProdHybridNode parentNode = (ProdHybridNode) node.getParent();
        if (parentNode == null) {
            ppIndex = -1;
            ppStr = "NULL";
        } else {
            ppIndex = parentNode.getElem().getGP().getProdId();
            ppStr = parentNode.getElem().getGP().toLine();
        }

        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);

            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;

                int[] f_rep = {featureType, ppIndex, node.toWord().getWordId(),
                              node.getElem().getGP().getProdId(),
                              pChildNode.toWord().getWordId(),
                              pIndex};
                String[] frep = {"MRLRule", ppStr, node.toWord().getWordStr(),
                                node.getElem().getGP().toLine(),
                                pChildNode.toWord().getWordStr(),
                                pChildNode.getElem().getGP().toLine()};

                double f_val = 1.0;
                FVPair f = this.createFVPair(f_rep, frep, f_val);
                this.addFV(fv, f);

                this.extractMRLFeatures(pChildNode, fv, featureType);
            }
        }
    }

    private void extractExpandedRuleFeatures(ProdHybridNode node,
                                             ArrayList<FVPair> fv, int featureType) {

        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isIdent() || gp.isNum() || gp.isUnum() || gp.isConst()) {
            return;
                 }
         */

        int pIndex = gp.getProdId();

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        int[] f_rep = new int[2 + childrenNodes.size()];
        String[] frep = new String[2 + childrenNodes.size()];

        f_rep[0] = featureType;
        frep[0] = "ExpandedRule";

        f_rep[1] = pIndex;
        frep[1] = gp.toLine();

        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            Word word = childNode.toWord();
            if (this.isIdent(word)) {
                f_rep[i + 2] = -1;
                frep[i + 2] = "IDENT";
            } else if (this.isNumber(word)) {
                f_rep[i + 2] = -2;
                frep[i + 2] = "NUMBER";
            } else {
                if (childNode instanceof WordHybridNode) {
                    f_rep[i + 2] = childNode.toWord().getWordId();
                    frep[i + 2] = childNode.toWord().getWordStr();
                } else {
                    ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                    GProduction pChild = pChildNode.getElem().getGP();
                    if (pChild.isUnum()) {
                        f_rep[i + 2] = -100;
                        frep[i + 2] = "UNUM";
                    } else if (pChild.isNum()) {
                        f_rep[i + 2] = -200;
                        frep[i + 2] = "NUM";
                    } else if (pChild.isIdent()) {
                        f_rep[i + 2] = -300;
                        frep[i + 2] = "IDENT";
                    } else {
                        f_rep[i +
                                2] = pChildNode.getElem().getGP().getProdId() +
                                     10000;
                        frep[i + 2] = pChildNode.getElem().getGP().toLine();
                    }
                }
            }
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractExpandedRuleFeatures(pChildNode, fv, featureType);
            }
        }
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);
    }

    private void extractRuleFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                     int featureType) {

        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isIdent() || gp.isNum() || gp.isUnum() || gp.isConst()) {
            return;
                 }
         */

        int pIndex = gp.getProdId();

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        int[] f_rep = new int[2 + childrenNodes.size()];
        String[] frep = new String[2 + childrenNodes.size()];

        f_rep[0] = featureType;
        frep[0] = "Rule";

        f_rep[1] = pIndex;
        frep[1] = gp.toLine();

        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            Word word = childNode.toWord();
            if (this.isIdent(word)) {
                f_rep[i + 2] = -1;
                frep[i + 2] = "IDENT";
            } else if (this.isNumber(word)) {
                f_rep[i + 2] = -2;
                frep[i + 2] = "NUMBER";
            } else {
                f_rep[i + 2] = childNode.toWord().getWordId();
                frep[i + 2] = childNode.toWord().getWordStr();
            }
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractRuleFeatures(pChildNode, fv, featureType);
            }
        }
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);
    }

    private void extractMRLFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                    int featureType) {
        GProduction gp = node.getElem().getGP();
        int pIndex = gp.getProdId();

        ProdHybridNode parentNode = (ProdHybridNode) node.getParent();
        GProduction parent;
        int ppIndex;
        String ppStr;
        if (parentNode != null) {
            parent = parentNode.getElem().getGP();
            ppIndex = parent.getProdId();
            ppStr = parent.toLine();
        } else {
            ppIndex = -1;
            ppStr = "NULL";
        }

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            int cIndex = childNode.toWord().getWordId();
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                GProduction child = pChildNode.getElem().getGP();
                int cpIndex;
                if (child.isNum()) {
                    cpIndex = -2;
                } else if (child.isUnum()) {
                    cpIndex = -3;
                } else if (child.isIdent()) {
                    cpIndex = -4;
                } else if (child.isConstant()) {
                    cpIndex = -1;
                } else {
                    cpIndex = child.getProdId();
                }

                int[] f_rep = new int[] {featureType, ppIndex,
                              node.toWord().getWordId(), pIndex, cIndex,
                              cpIndex};
                String[] frep = new String[] {"MRL", ppStr, gp.toLine(),
                                String.valueOf(cIndex),
                                pChildNode.getElem().getGP().toLine()};
                double f_val = 1.0;
                FVPair f = this.createFVPair(f_rep, frep, f_val);
                this.addFV(fv, f);
                ProdHybridNode childPNode = (ProdHybridNode) childNode;
                this.extractMRLFeatures(childPNode, fv, featureType);
            }
        }
    }

    @SuppressWarnings("unused")
	private void extractRangeAllBigramFeatures(ProdHybridNode node,
                                               ArrayList<FVPair> fv,
                                               int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */

        int pIndex = gp.getProdId();

        ArrayList<Word> allwords = node.getAllWords();

        Word word1 = Word.BEGIN;
        Word word;
        for (int i = 0; i < allwords.size(); i++) {
            word = allwords.get(i);
            int[] f_rep = new int[] {featureType, pIndex, word1.getWordId(),
                          word.getWordId()};
            String[] frep = new String[] {"RangeAllBigram", gp.toLine(),
                            word1.getWordStr(), word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);

            word1 = word;
        }

        word = Word.END;
        int[] f_rep = new int[] {featureType, pIndex, word1.getWordId(),
                      word.getWordId()};
        String[] frep = new String[] {"RangeAllBigram", gp.toLine(),
                        word1.getWordStr(), word.getWordStr()};
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-NL feature
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode childPNode = (ProdHybridNode) childNode;
                this.extractRangeAllBigramFeatures(childPNode, fv, featureType);
            }
        }

    }

    @SuppressWarnings("unused")
	private void extractRangeRelationFeatures(ProdHybridNode node,
                                              ArrayList<FVPair> fv,
                                              int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        ArrayList<HybridNode> children = node.getChildren();
        for (int i = 0; i < children.size(); i++) {
            HybridNode child = children.get(i);
            if (child instanceof ProdHybridNode) {
                this.extractRangeRelationFeatures((ProdHybridNode) child, fv,
                                                  featureType);
            }
        }

        int pIndex = gp.getProdId();
        ArrayList<Word> allwords = node.getAllWords();

        int bIndex = node.getElem().getBIndex();
        int eIndex = node.getElem().getEIndex();

        ProdHybridNode parentNode = (ProdHybridNode) node.getParent();

        int pbIndex;
        int peIndex;

        int ppIndex;

        GProduction parent;
        if (parentNode != null) {
            parent = parentNode.getElem().getGP();
            pbIndex = parentNode.getElem().getBIndex();
            peIndex = parentNode.getElem().getEIndex();
            ppIndex = parent.getProdId();
        } else {
            return;
        }

        Word word;
        for (int i = pbIndex; i < bIndex; i++) {
            word = allwords.get(i);

            int[] f_rep = new int[] {featureType, -1, ppIndex, pIndex,
                          word.getWordId()};
            String[] frep = new String[] {"RangeLeft", parent.toLine(),
                            gp.toLine(), word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }
        if (pbIndex == bIndex) {
            int[] f_rep = new int[] {featureType, -1, ppIndex, pIndex, -1};
            String[] frep = new String[] {"RangeLeft", parent.toLine(),
                            gp.toLine(), "EMPTY"};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

        for (int i = bIndex; i < eIndex; i++) {
            word = allwords.get(i);

            int[] f_rep = new int[] {featureType, 0, ppIndex, pIndex,
                          word.getWordId()};
            String[] frep = new String[] {"RangeMiddle", parent.toLine(),
                            gp.toLine(), word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }
        if (bIndex == eIndex) {
            int[] f_rep = new int[] {featureType, 0, ppIndex, pIndex, -1};
            String[] frep = new String[] {"RangeMiddle", parent.toLine(),
                            gp.toLine(), "EMPTY"};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

        for (int i = eIndex; i < peIndex; i++) {
            word = allwords.get(i);

            int[] f_rep = new int[] {featureType, +1, ppIndex, pIndex,
                          word.getWordId()};
            String[] frep = new String[] {"RangeRight", parent.toLine(),
                            gp.toLine(), word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }
        if (eIndex == peIndex) {
            int[] f_rep = new int[] {featureType, +1, ppIndex, pIndex, -1};
            String[] frep = new String[] {"RangeRight", parent.toLine(),
                            gp.toLine(), "EMPTY"};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

    }

    private boolean isIdent(Word word) {
        return Character.isUpperCase(word.getWordStr().charAt(0));
    }

    private boolean isNumber(Word word) {
        try {
            Double.parseDouble(word.getWordStr());
            return true;
        } catch (NumberFormatException e) {
            return false;
        }
    }

    @SuppressWarnings("unused")
	private void extractRangeAllFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                         int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();

        ArrayList<Word> allwords = node.getAllWords();

        Word word;
        for (int i = 0; i < allwords.size(); i++) {
            word = allwords.get(i);
            if (this.isIdent(word) || this.isNumber(word)) {
                continue;
            }
            int[] f_rep = new int[] {featureType, pIndex, word.getWordId()};
            String[] frep = new String[] {"RangeAll", gp.toLine(),
                            word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-NL feature
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode childPNode = (ProdHybridNode) childNode;
                this.extractRangeAllFeatures(childPNode, fv, featureType);
            }
        }

    }

    @SuppressWarnings("unused")
	private void extractRangeOutBigramFeatures(ProdHybridNode node,
                                               ArrayList<FVPair> fv,
                                               int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();
        int bIndex = node.getElem().getBIndex();
        int eIndex = node.getElem().getEIndex();
        ArrayList<Word> allwords = node.getAllWords();

        Word word1 = Word.BEGIN;
        Word word;
        for (int i = 0; i < bIndex; i++) {
            word = allwords.get(i);
            int[] f_rep = new int[] {featureType, pIndex, word1.getWordId(),
                          word.getWordId()};
            String[] frep = new String[] {"RangeOutBigram", gp.toLine(),
                            word1.getWordStr(), word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
            word1 = word;
        }

        {
            word = Word.BEGIN;
            int[] f_rep = new int[] {featureType, pIndex, word1.getWordId(),
                          word.getWordId()};
            String[] frep = new String[] {"RangeOutBigram", gp.toLine(),
                            word1.getWordStr(), word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

        word1 = Word.END;
        for (int i = eIndex; i < allwords.size(); i++) {
            word = allwords.get(i);
            int[] f_rep = new int[] {featureType, pIndex, word1.getWordId(),
                          word.getWordId()};
            String[] frep = new String[] {"RangeOutBigram", gp.toLine(),
                            word1.getWordStr(), word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
            word1 = word;
        }

        {
            word = Word.END;
            int[] f_rep = new int[] {featureType, pIndex, word1.getWordId(),
                          word.getWordId()};
            String[] frep = new String[] {"RangeOutBigram", gp.toLine(),
                            word1.getWordStr(), word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-NL feature
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode childPNode = (ProdHybridNode) childNode;
                this.extractRangeOutBigramFeatures(childPNode, fv, featureType);
            }
        }

    }

    @SuppressWarnings("unused")
	private void extractRangeOutFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                         int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();
        int bIndex = node.getElem().getBIndex();
        int eIndex = node.getElem().getEIndex();
        ArrayList<Word> allwords = node.getAllWords();
        for (int i = 0; i < bIndex; i++) {
            Word word = allwords.get(i);
            int wordId = word.getWordId();
            if (this.isIdent(word)) {
                wordId = -1;
            } else if (this.isNumber(word)) {
                wordId = -2;
            }
            int[] f_rep = new int[] {featureType, pIndex, wordId};
            String[] frep = new String[] {"RangeOut", gp.toLine(),
                            word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

        for (int i = eIndex; i < allwords.size(); i++) {
            Word word = allwords.get(i);
            int wordId = word.getWordId();
            if (this.isIdent(word)) {
                wordId = -1;
            } else if (this.isNumber(word)) {
                wordId = -2;
            }
            int[] f_rep = new int[] {featureType, pIndex, wordId};
            String[] frep = new String[] {"RangeOut", gp.toLine(),
                            word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-NL feature
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode childPNode = (ProdHybridNode) childNode;
                this.extractRangeOutFeatures(childPNode, fv, featureType);
            }
        }

    }

    @SuppressWarnings("unused")
	private void getMRLAbove(ProdHybridNode node, ArrayList<Integer> v,
                             ArrayList<String> reps) {
        ProdHybridNode parentNode = (ProdHybridNode) node.getParent();
        if (parentNode == null) {
            return;
        }

        GProduction parent = parentNode.getElem().getGP();
        v.add(parent.getProdId());
        reps.add(parent.toLine());
        this.getMRLAbove(parentNode, v, reps);
    }

    private void getMRLBelow(ProdHybridNode node, ArrayList<Integer> v,
                             ArrayList<String> reps) {

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode childProdNode = (ProdHybridNode) childNode;
                GProduction childGP = childProdNode.getElem().getGP();
                int id;
                String rep;
                if (childGP.isIdent()) {
                    id = -2;
                    rep = "IDENT";
                } else if (childGP.isNum()) {
                    id = -3;
                    rep = "NUM";
                } else if (childGP.isUnum()) {
                    id = -4;
                    rep = "UNUM";
                } else
                if (childGP.isConstant()) {
                    id = -1;
                    rep = "CONST";
                } else {
                    id = childGP.getProdId();
                    rep = childGP.toLine();
                }
                if (!v.contains(id)) {
                    v.add(childGP.getProdId());
                    reps.add(rep);
                }

                this.getMRLBelow(childProdNode, v, reps);
            }
        }

    }

    @SuppressWarnings("unused")
	private void extractRelativeMRLFeatures(ProdHybridNode node,
                                            ArrayList<FVPair> fv, int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();

        ArrayList<Integer> belowMRLs = new ArrayList<Integer>();
        ArrayList<String> belowMRLreps = new ArrayList<String>();
        this.getMRLBelow(node, belowMRLs, belowMRLreps);

        for (int i = 0; i < belowMRLs.size(); i++) {
            int belowMRLId = belowMRLs.get(i);
            String belowMRLrep = belowMRLreps.get(i);

            int[] f_rep = new int[] {featureType, pIndex, belowMRLId};
            String[] frep = new String[] {"BelowMRL", gp.toLine(), belowMRLrep};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);

        }

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-NL feature
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode childPNode = (ProdHybridNode) childNode;
                this.extractRelativeMRLFeatures(childPNode, fv, featureType);
            }
        }
    }

    @SuppressWarnings("unused")
	private void extractRangeInBigramFeatures(ProdHybridNode node,
                                              ArrayList<FVPair> fv,
                                              int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();
        int bIndex = node.getElem().getBIndex();
        int eIndex = node.getElem().getEIndex();

        ArrayList<Word> allwords = node.getAllWords();

        Word word1 = Word.BEGIN;
        Word word;
        int word1Id = word1.getWordId();
        int wordId;
        for (int i = bIndex; i < eIndex; i++) {
            word = allwords.get(i);
            wordId = word.getWordId();
            if (this.isIdent(word)) {
                wordId = -1;
            } else if (this.isNumber(word)) {
                wordId = -2;
            }
            int[] f_rep = new int[] {featureType, pIndex, word1Id, wordId};
            String[] frep = new String[] {"RangeInBigram", gp.toLine(),
                            word1.toString(), word.toString()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
            word1 = word;
            word1Id = wordId;
        }

        {
            word = Word.END;
            wordId = word.getWordId();
            int[] f_rep = new int[] {featureType, pIndex, word1Id, wordId};
            String[] frep = new String[] {"RangeInBigram", gp.toLine(),
                            word1.toString(), word.toString()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-NL feature
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode childPNode = (ProdHybridNode) childNode;
                this.extractRangeInBigramFeatures(childPNode, fv, featureType);
            }
        }
    }

    @SuppressWarnings("unused")
	private void extractInsidePhraseFeatures(ProdHybridNode node,
                                             ArrayList<FVPair> fv, int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();
        int bIndex = node.getElem().getBIndex();
        int eIndex = node.getElem().getEIndex();

        ArrayList<Word> allwords = node.getAllWords();

        Word word;
        int wordId;
        int[] f_rep = new int[2 + eIndex - bIndex];
        String[] frep = new String[2 + eIndex - bIndex];
        f_rep[0] = featureType;
        f_rep[1] = pIndex;
        frep[0] = "InsidePhrase";
        frep[1] = gp.toLine();
        for (int i = bIndex; i < eIndex; i++) {
            word = allwords.get(i);
            wordId = word.getWordId();
            String wordStr = word.getWordStr();
            if (this.isIdent(word)) {
                wordId = -1;
                wordStr = "IDENT";
            } else if (this.isNumber(word)) {
                wordId = -2;
                wordStr = "NUMBER";
            }
            f_rep[i - bIndex + 2] = wordId;
            frep[i - bIndex + 2] = wordStr;
        }
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);

        ArrayList<HybridNode> childrenNodes = node.getChildren();
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                this.extractInsidePhraseFeatures((ProdHybridNode) childNode, fv,
                                                 featureType);
            }
        }
    }

    @SuppressWarnings("unused")
	private void extractTwoLevelRangeInFeatures(ProdHybridNode node,
                                                ArrayList<FVPair> fv,
                                                int featureType) {

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-NL feature
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode childPNode = (ProdHybridNode) childNode;
                this.extractTwoLevelRangeInFeatures(childPNode, fv, featureType);
            }
        }

        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */

        int pIndex = gp.getProdId();

        ArrayList<Word> allwords = node.getAllWords();

        ProdHybridNode parentNode = (ProdHybridNode) node.getParent();
        GProduction parent;
        int ppIndex;
        String ppStr;
        if (parentNode != null) {
            parent = parentNode.getElem().getGP();
            ppIndex = parent.getProdId();
            ppStr = parent.toLine();
        } else {
            return;
        }

        int bIndex = parentNode.getElem().getBIndex();
        int eIndex = parentNode.getElem().getEIndex();

        Word word;
        int wordId;
        for (int i = bIndex; i < eIndex; i++) {
            word = allwords.get(i);
            wordId = word.getWordId();
            if (this.isIdent(word)) {
                wordId = -1;
            } else if (this.isNumber(word)) {
                wordId = -2;
            }
            int[] f_rep = new int[] {featureType, ppIndex, pIndex, wordId};
            String[] frep = new String[] {"TwoLevelRangeIn", ppStr, gp.toLine(),
                            word.toString()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }
    }

    private Pair<ArrayList<Integer>,
            ArrayList<String>> getMRLSubtree(ProdHybridNode node) {
        ArrayList<Integer> results_int = new ArrayList<Integer>();
        ArrayList<String> results_str = new ArrayList<String>();
        GProduction gp = node.getElem().getGP();
        if (gp.isIdent()) {
            results_int.add( -20);
            results_str.add("IDENT");
            return new Pair<ArrayList<Integer>,
                    ArrayList<String>>(results_int, results_str);
        } else if (gp.isUnum()) {
            results_int.add( -30);
            results_str.add("UNUM");
            return new Pair<ArrayList<Integer>,
                    ArrayList<String>>(results_int, results_str);
        } else if (gp.isNum()) {
            results_int.add( -40);
            results_str.add("NUM");
            return new Pair<ArrayList<Integer>,
                    ArrayList<String>>(results_int, results_str);
        } else if (gp.isConstant()) {
            results_int.add( -10);
            results_str.add("CONST");
            return new Pair<ArrayList<Integer>,
                    ArrayList<String>>(results_int, results_str);
        }
        results_int.add(gp.getProdId());
        results_str.add(gp.toLine());

        results_int.add( -100);
        results_str.add("END");
        ArrayList<HybridNode> children = node.getChildren();
        for (int i = 0; i < children.size(); i++) {
            HybridNode child = children.get(i);
            if (child instanceof ProdHybridNode) {
                ProdHybridNode pChild = (ProdHybridNode) child;
                Pair<ArrayList<Integer>,
                        ArrayList<String>> subResults = this.getMRLSubtree(pChild);
                ArrayList<Integer> subResults_int = subResults.getFirst();
                ArrayList<String> subResults_str = subResults.getSecond();
                for (int k = 0; k < subResults_int.size(); k++) {
                    results_int.add(subResults_int.get(k));
                    results_str.add(subResults_str.get(k));
                }
            }
        }
        results_int.add( -200);
        results_str.add("END");

        return new Pair<ArrayList<Integer>,
                ArrayList<String>>(results_int, results_str);
    }

    @SuppressWarnings("unused")
	private void extractSubtreePhraseFeatures(ProdHybridNode node,
                                              ArrayList<FVPair> fv,
                                              int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant() || gp.isTerminal()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */

        Pair<ArrayList<Integer>, ArrayList<String>> subtree = this.getMRLSubtree(node);
        ArrayList<Integer> subtree_int = subtree.getFirst();
        ArrayList<String> subtree_str = subtree.getSecond();

        int bIndex = node.getElem().getBIndex();
        int eIndex = node.getElem().getEIndex();

        ArrayList<Word> allwords = node.getAllWords();

        Word word;
        int wordId;
        String wordStr;

        for (int i = bIndex; i < eIndex; i++) {
            word = allwords.get(i);
            wordId = word.getWordId();
            wordStr = word.getWordStr();
            if (this.isIdent(word)) {
                wordId = -1;
                wordStr = "IDENT";
            } else if (this.isNumber(word)) {
                wordId = -2;
                wordStr = "NUMBER";
            }

            int[] f_rep = new int[1 + subtree_int.size() + 2];
            String[] frep = new String[1 + subtree_int.size() + 2];

            f_rep[0] = featureType;
            frep[0] = "SubtreePhrase";

            for (int k = 0; k < subtree_int.size(); k++) {
                f_rep[1 + k] = subtree_int.get(k);
                frep[1 + k] = subtree_str.get(k);
            }

            f_rep[subtree_int.size() + 1] = -999;
            frep[subtree_str.size() + 1] = "ENDOFSUBTREE";

            f_rep[subtree_int.size() + 2] = wordId;
            frep[subtree_str.size() + 2] = wordStr;

            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-NL feature
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode childPNode = (ProdHybridNode) childNode;
                this.extractSubtreePhraseFeatures(childPNode, fv, featureType);
            }
        }
    }


    private void extractRangeInFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                        int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();
        int bIndex = node.getElem().getBIndex();
        int eIndex = node.getElem().getEIndex();

        ArrayList<Word> allwords = node.getAllWords();

        Word word;
        int wordId;

        for (int i = bIndex; i < eIndex; i++) {
            word = allwords.get(i);
            wordId = word.getWordId();
            if (this.isIdent(word)) {
                wordId = -1;
            } else if (this.isNumber(word)) {
                wordId = -2;
            }
            int[] f_rep = new int[] {featureType, pIndex, wordId};
            String[] frep = new String[] {"RangeIn", gp.toLine(), word.toString()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
        }

        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-NL feature
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode childPNode = (ProdHybridNode) childNode;
                this.extractRangeInFeatures(childPNode, fv, featureType);
            }
        }
    }


    @SuppressWarnings("unused")
	private void extractUnigramFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                        int featureType) {

        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();
        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-NL feature
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof WordHybridNode) {
                WordHybridNode wChildNode = (WordHybridNode) childNode;
                int[] f_rep = new int[] {featureType, pIndex,
                              wChildNode.getWord().getWordId()};
                String[] frep = new String[] {"Unigram", gp.toLine(),
                                wChildNode.getWord().getWordStr()};
                double f_val = 1.0;
                FVPair f = this.createFVPair(f_rep, frep, f_val);
                this.addFV(fv, f);
            } else {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractUnigramFeatures(pChildNode, fv, featureType);
            }
        }

    }

    @SuppressWarnings("unused")
	private void extractExpandedTrigramFeatures(ProdHybridNode node,
                                                ArrayList<FVPair> fv,
                                                int featureType) {

        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();
        ArrayList<HybridNode> childrenNodes = node.getChildren();

        Word word1 = Word.BEGIN;
        Word word2 = Word.BEGIN;
        Word word;
        int word1Index = word1.getWordId();
        int word2Index = word2.getWordId();
        int wordIndex;
        String word1Str = word1.getWordStr();
        String word2Str = word2.getWordStr();
        String wordStr;
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof WordHybridNode) {
                word = childNode.toWord();
                wordIndex = word.getWordId();
                wordStr = word.getWordStr();
            } else {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                GProduction pChild = pChildNode.getElem().getGP();
                if (pChild.isIdent()) {
                    wordIndex = -2;
                    wordStr = "IDENT";
                } else if (pChild.isNum()) {
                    wordIndex = -3;
                    wordStr = "NUM";
                } else if (pChild.isUnum()) {
                    wordIndex = -4;
                    wordStr = "UNUM";
                } else if (pChild.isConstant()) {
                    wordIndex = -1;
                    wordStr = "CONSTANT";
                } else {
                    wordIndex = pChildNode.getElem().getGP().getProdId() +
                                10000;
                    wordStr = pChildNode.getElem().getGP().toLine();
                }
            }
            int[] f_rep = {featureType, pIndex, word1Index, word2Index,
                          wordIndex};
            String[] frep = {"ExpandedTrigram", gp.toLine(), word1Str, word2Str,
                            wordStr};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractExpandedTrigramFeatures(pChildNode, fv, featureType);
            }

            word1Index = word2Index;
            word2Index = wordIndex;
            word1Str = word2Str;
            word2Str = wordStr;
        }

        word = Word.END;
        int[] f_rep = new int[] {featureType, pIndex, word1Index, word2Index,
                      word.getWordId()};
        String[] frep = new String[] {"ExpandedTrigram", gp.toLine(), word1Str,
                        word2Str, word.getWordStr()};
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);
    }

    @SuppressWarnings("unused")
	private void extractTrigramFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                        int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();
        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // MRL-(NL-NL) features
        Word word1 = Word.BEGIN;
        Word word2 = Word.BEGIN;
        for (int i = 0; i < childrenNodes.size(); i++) {
            Word word = childrenNodes.get(i).toWord();

            int[] f_rep = new int[] {featureType, pIndex, word1.getWordId(),
                          word2.getWordId(), word.getWordId()};
            String[] frep = new String[] {"Trigram", gp.toLine(),
                            word1.getWordStr(), word2.getWordStr(),
                            word.getWordStr()};
            double f_val = 1.0;
            FVPair f = this.createFVPair(f_rep, frep, f_val);
            this.addFV(fv, f);

            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof ProdHybridNode) {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractTrigramFeatures(pChildNode, fv, featureType);
            }
            word1 = word2;
            word2 = word;
        }

        int[] f_rep = new int[] {featureType, pIndex, word1.getWordId(),
                      word2.getWordId(), Word.END.getWordId()};
        String[] frep = new String[] {"Trigram", gp.toLine(), word1.getWordStr(),
                        word2.getWordStr(), Word.END.getWordStr()};
        double f_val = 1.0;
        FVPair f = this.createFVPair(f_rep, frep, f_val);
        this.addFV(fv, f);
    }

    private void extract2MRL1NLFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                        int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();
        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // (MRL-MRL)-NL features
        GProduction parent;
        ProdHybridNode parentNode = (ProdHybridNode) node.getParent();
        if (parentNode != null) {
            parent = parentNode.getElem().getGP();
        } else {
            parent = null;
        }

        int ppIndex = -1;
        if (parent != null) {
            ppIndex = parent.getProdId();
        }

        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof WordHybridNode) {
                WordHybridNode wChildNode = (WordHybridNode) childNode;
                int[] f_rep;
                String[] frep;
                double f_val;
                FVPair f;

                f_rep = new int[] {featureType, ppIndex, pIndex,
                        wChildNode.getWord().getWordId()};
                if (parent != null) {
                    frep = new String[] {"2MRL1NL", parent.toLine(), gp.toLine(),
                           wChildNode.getWord().getWordStr()};
                } else {
                    frep = new String[] {"2MRL1NL", "NULL", gp.toLine(),
                           wChildNode.getWord().getWordStr()};
                }
                f_val = 1.0;
                f = this.createFVPair(f_rep, frep, f_val);
                this.addFV(fv, f);

                f_rep = new int[] {featureType, ppIndex,
                        wChildNode.getWord().getWordId()};
                if (parent != null) {
                    frep = new String[] {"2MRL1NL-JUMP", parent.toLine(), "-",
                           wChildNode.getWord().getWordStr()};
                } else {
                    frep = new String[] {"2MRL1NL-JUMP", "NULL", "-",
                           wChildNode.getWord().getWordStr()};
                }
                f_val = 1.0;
                f = this.createFVPair(f_rep, frep, f_val);
                this.addFV(fv, f);

            } else {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extract2MRL1NLFeatures(pChildNode, fv, featureType);
            }
        }

    }

    @SuppressWarnings("unused")
	private void extractGrandParMRL1NLFeatures(ProdHybridNode node,
                                               ArrayList<FVPair> fv,
                                               int featureType) {
        GProduction gp = node.getElem().getGP();
        int pIndex = gp.getProdId();
        ArrayList<HybridNode> childrenNodes = node.getChildren();

        // (MRL-*)-NL features
        GProduction parent;
        ProdHybridNode parentNode = (ProdHybridNode) node.getParent();
        if (parentNode != null) {
            parent = parentNode.getElem().getGP();
        } else {
            parent = null;
        }

        int ppIndex = -1;
        if (parent != null) {
            ppIndex = parent.getProdId();
        }
        for (int i = 0; i < childrenNodes.size(); i++) {
            HybridNode childNode = childrenNodes.get(i);
            if (childNode instanceof WordHybridNode) {
                WordHybridNode wChildNode = (WordHybridNode) childNode;
                int[] f_rep = new int[] {featureType, ppIndex,
                              wChildNode.getWord().getWordId()};
                String[] frep = new String[] {"GrandParMRLNL", parent.toLine(),
                                wChildNode.getWord().getWordStr()};
                double f_val = 1.0;
                FVPair f = this.createFVPair(f_rep, frep, f_val);
                this.addFV(fv, f);
            } else {
                ProdHybridNode pChildNode = (ProdHybridNode) childNode;
                this.extractGrandParMRL1NLFeatures(pChildNode, fv, featureType);
            }
        }
    }

    @SuppressWarnings("unused")
	private void extractSiblingFeatures(ProdHybridNode node, ArrayList<FVPair> fv,
                                        int featureType) {
        GProduction gp = node.getElem().getGP();

        if (gp.isConstant()) {
            return;
        }

        /*
         if (gp.isNum() || gp.isUnum() || gp.isIdent() || gp.isConst()) {
            return;
                 }
         */


        int pIndex = gp.getProdId();
        ProdHybridNode parentNode = (ProdHybridNode) node.getParent();

        if (parentNode != null) {
            ArrayList<HybridNode> childrenOfParent = parentNode.getChildren();
            for (int i = 0; i < childrenOfParent.size(); i++) {
                HybridNode childOfParent = childrenOfParent.get(i);
                if (childOfParent != node) {
                    Word word = childOfParent.toWord();
                    int[] f_rep = new int[] {featureType, word.getWordId(),
                                  pIndex};
                    String[] frep = new String[] {"Sibling", word.getWordStr(),
                                    gp.toLine()};
                    double f_val = 1.0;
                    FVPair f = this.createFVPair(f_rep, frep, f_val);
                    this.addFV(fv, f);
                }
            }
        }

        ArrayList<HybridNode> children = node.getChildren();
        for (int i = 0; i < children.size(); i++) {
            HybridNode child = children.get(i);
            if (child instanceof ProdHybridNode) {
                this.extractSiblingFeatures((ProdHybridNode) child, fv,
                                            featureType);
            }
        }
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < this._allFeatures.size(); i++) {
            PFeature feature = this._allFeatures.get(i);
            sb.append(feature.toString());
            sb.append('\n');
        }

        return sb.toString();
    }

}
