/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package hybridtree.core.viterbi;

import hybridtree.core.Candidate;
import hybridtree.core.CandidateSet;
import hybridtree.core.DeriveTable;
import hybridtree.core.HybridNode;
import hybridtree.core.Param;
import hybridtree.core.ProdHybridNode;
import hybridtree.core.ResElem;
import hybridtree.core.ResElemList;
import hybridtree.data.Production;
import hybridtree.data.Rule;
import hybridtree.data.TrainExample;
import hybridtree.data.Word;

/**
 * @author Wei LU
 * @version 1.0
 */
import java.util.ArrayList;

public abstract class MaxTable extends DeriveTable {

    /**
     * the Productions
     */
    protected ArrayList<Production> _productions;

    protected int _K;


    protected MaxTable(TrainExample example, Param parameter) {
        super(example, parameter);
        this._productions = example.getProductions();
    }

    /**
     * Set the collection of Productions
     * @param prods a collection of Productions
     */
    public void setProductions(ArrayList<Production> prods) {
        this._productions = prods;
    }

    /**
     * Get the collection of Productions
     * @return ArrayList
     */
    public ArrayList<Production> getProductions() {
        return this._productions;
    }

    public abstract void alignMax(int K);

    public static UGMaxTable createUGTable(TrainExample example, Param parameter) {
        return new UGMaxTable(example, parameter);
    }

    public static BGMaxTable createBGTable(TrainExample example, Param parameter) {
        return new BGMaxTable(example, parameter);
    }

    public static MBGMaxTable createMBGTable(TrainExample example, Param parameter) {
        return new MBGMaxTable(example, parameter);
    }

    public static MGMaxTable createMGTable(TrainExample example, Param parameter_unigram, Param parameter_bigram) {
        return new MGMaxTable(example, parameter_unigram, parameter_bigram);
    }

    public CandidateSet getTopCandidates() {
        CandidateSet candidates = new CandidateSet(this._exampleId);
        ArrayList<ProdHybridNode> rootNodes = this.getOptimalHybridTrees();
        if (!this._ignoreFlag) {
            for (int rank = 0; rank < rootNodes.size(); rank++) {
                Candidate candidate = Candidate.createCandidate(candidates, rootNodes.get(rank), rank);
                candidate.setCorrect();
            }
        }
        return candidates;
    }


    protected ResElemList _final_list;

    public ResElemList getOptResults() {
        if (this._ignoreFlag) {
            return ResElemList._emptyList;
        }
        return this._final_list;
    }

    public ArrayList<ProdHybridNode> getOptimalHybridTrees() {
        ArrayList<ProdHybridNode> trees = new ArrayList<ProdHybridNode>();
        if (this._ignoreFlag) {
            return trees;
        }
        for (int i = 0; i < this._final_list.size(); i++) {
            ResElem rootElem = this._final_list.get(i);
            if (rootElem.getProb() > 0) {
                ProdHybridNode rootNode = this.convertToHybridNode(rootElem, 0, this._words);
                trees.add(0, rootNode);
            }
        }
        return trees;
    }

    private ProdHybridNode convertToHybridNode(ResElem elem, int cIndex, ArrayList<Word> allwords) {
        ProdHybridNode node = HybridNode.createProdHybridNode(elem, cIndex, allwords);
        ResElem[] childrenElems = elem.getChildren();
        int bIndex = elem.getBIndex();
        int eIndex = elem.getEIndex();
        ArrayList<HybridNode> children = new ArrayList<HybridNode>();

        if (childrenElems.length == 0) {
            for (int wIndex = bIndex; wIndex < eIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                HybridNode cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
        } else if (childrenElems.length == 1) {
            ResElem childElem = childrenElems[0];
            HybridNode cNode;

            int cBIndex, cEIndex;
            cBIndex = bIndex;
            cEIndex = childElem.getBIndex();
            for (int wIndex = cBIndex; wIndex < cEIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
            cNode = convertToHybridNode(childElem, 0, allwords);
            children.add(cNode);
            cNode.setParent(node);

            cBIndex = childElem.getEIndex();
            cEIndex = eIndex;
            for (int wIndex = cBIndex; wIndex < cEIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
        } else if (childrenElems.length == 2) {
            ResElem childElem1, childElem2;
            int childCIndex1, childCIndex2;

            if (!Rule.isReversed(elem.getRulePattern())) {
                childElem1 = childrenElems[0];
                childElem2 = childrenElems[1];
                childCIndex1 = 0;
                childCIndex2 = 1;
            } else {
                childElem1 = childrenElems[1];
                childElem2 = childrenElems[0];
                childCIndex1 = 1;
                childCIndex2 = 2;
            }
            HybridNode cNode;

            int cBIndex, cEIndex;
            cBIndex = bIndex;
            cEIndex = childElem1.getBIndex();
            for (int wIndex = cBIndex; wIndex < cEIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
            cNode = convertToHybridNode(childElem1, childCIndex1, allwords);
            children.add(cNode);
            cNode.setParent(node);

            cBIndex = childElem1.getEIndex();
            cEIndex = childElem2.getBIndex();
            for (int wIndex = cBIndex; wIndex < cEIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
            cNode = convertToHybridNode(childElem2, childCIndex2, allwords);
            children.add(cNode);
            cNode.setParent(node);

            cBIndex = childElem2.getEIndex();
            cEIndex = eIndex;
            for (int wIndex = cBIndex; wIndex < cEIndex; wIndex++) {
                Word word = this.getWord(wIndex);
                cNode = HybridNode.createWordHybridNode(word);
                children.add(cNode);
                cNode.setParent(node);
            }
        }
        node.setChildren(children);

        return node;
    }

}
