/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package hybridtree.core.viterbi;

import hybridtree.core.*;
import hybridtree.data.Production;
import hybridtree.data.Rule;
import hybridtree.data.TrainExample;
import hybridtree.data.Word;

import java.util.ArrayList;

/**
 * @author Wei LU
 * @version 1.0
 */
public class UGMaxTable extends MaxTable {

    private ResElemList[][][] _results;
    private ResElem[][][] _results_w;

    private ResElemList[][][][] _results_X;
    private ResElemList[][][][] _results_wX;
    private ResElemList[][][][] _results_Xw;
    private ResElemList[][][][] _results_w1Xw2;

    private void resetMax() {

        this._results = new ResElemList[this._productions.size()][][];
        this._results_w = new ResElem[this._productions.size()][][];
        for (int pIndex = 0; pIndex < this._results.length; pIndex++) {
            this._results[pIndex] = new ResElemList[this._N][];
            this._results_w[pIndex] = new ResElem[this._N][];
            for (int bIndex = 0; bIndex < this._results[pIndex].length; bIndex++) {
                this._results[pIndex][bIndex] = new ResElemList[this._N - bIndex + 1];
                this._results_w[pIndex][bIndex] = new ResElem[this._N - bIndex + 1];
                for (int len = 1; len < this._results[pIndex][bIndex].length; len++) {
                    this._results[pIndex][bIndex][len] = new ResElemList(this._K);
                }
            }
        }

        this._results_X = new ResElemList[this._productions.size()][][][];
        this._results_wX = new ResElemList[this._productions.size()][][][];
        this._results_Xw = new ResElemList[this._productions.size()][][][];
        this._results_w1Xw2 = new ResElemList[this._productions.size()][][][];

        for (int pIndex = 0; pIndex < this._productions.size(); pIndex++) {
            Production gp = this._productions.get(pIndex);
            int childrenSize = gp.countRHSNonterminals();

            this._results_X[pIndex] = new ResElemList[childrenSize][][];
            this._results_wX[pIndex] = new ResElemList[childrenSize][][];
            this._results_Xw[pIndex] = new ResElemList[childrenSize][][];
            this._results_w1Xw2[pIndex] = new ResElemList[childrenSize][][];

            for (int cIndex = 0; cIndex < this._results_X[pIndex].length; cIndex++) {
                this._results_X[pIndex][cIndex] = new ResElemList[this._N][];
                this._results_wX[pIndex][cIndex] = new ResElemList[this._N][];
                this._results_Xw[pIndex][cIndex] = new ResElemList[this._N][];
                this._results_w1Xw2[pIndex][cIndex] = new ResElemList[this._N][];

                for (int bIndex = 0; bIndex < this._results_X[pIndex][cIndex].length; bIndex++) {
                    this._results_X[pIndex][cIndex][bIndex] = new ResElemList[this._N - bIndex + 1];
                    this._results_wX[pIndex][cIndex][bIndex] = new ResElemList[this._N - bIndex + 1];
                    this._results_Xw[pIndex][cIndex][bIndex] = new ResElemList[this._N - bIndex + 1];
                    this._results_w1Xw2[pIndex][cIndex][bIndex] = new ResElemList[this._N - bIndex + 1];

                    for (int len = 1; len < this._results_X[pIndex][cIndex][bIndex].length; len++) {
                        this._results_X[pIndex][cIndex][bIndex][len] = new ResElemList(this._K);
                        this._results_wX[pIndex][cIndex][bIndex][len] = new ResElemList(this._K);
                        this._results_Xw[pIndex][cIndex][bIndex][len] = new ResElemList(this._K);
                        this._results_w1Xw2[pIndex][cIndex][bIndex][len] = new ResElemList(this._K);
                    }
                }
            }
        }
    }

    /**
     * Constructor
     * @param example the Example
     * @param parameter the parameter
     */
    protected UGMaxTable(TrainExample example, Param parameter) {
        super(example, parameter);
    }

    private double computeEndProb(Production prod) {
        return this._parameter.getProb(prod.getGProduction(), Word.END);
    }

    protected double computeOneWordProb(Production prod, Word word) {
        return this._parameter.getProb(prod.getGProduction(), word);
    }


    private boolean addRes(ResElemList[][][][] re, Production prod, int cIndex, int bIndex, int eIndex, double prob, int rulePattern,
                           ResElem[] children) {
        if (Double.isInfinite(prob) || Double.isNaN(prob)) {
            return false;
        }
        ResElem res = new ResElem(prod.getGProduction(), bIndex, eIndex, prob, rulePattern, children);
        return re[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex].add(res);
    }

    private ResElemList getRes(ResElemList[][][][] re, Production prod, int cIndex, int bIndex, int eIndex) {
        return re[prod.getProdSeq()][cIndex][bIndex][eIndex - bIndex];
    }


    public ResElem getRes_w(Production prod, int bIndex, int eIndex) {
        return this._results_w[prod.getProdSeq()][bIndex][eIndex - bIndex];
    }

    private void setRes_w(Production prod, int bIndex, int eIndex, double prob) {
        ResElem res = new ResElem(prod.getGProduction(), bIndex, eIndex, prob);
        this._results_w[prod.getProdSeq()][bIndex][eIndex - bIndex] = res;
    }

    private boolean addRESULT(Production prod, int bIndex, int eIndex, double prob, int rulePattern, ResElem[] children) {
        if (Double.isInfinite(prob) || Double.isNaN(prob)) {
            return false;
        }
        ResElem res = new ResElem(prod.getGProduction(), bIndex, eIndex, prob, rulePattern, children);
        return this._results[prod.getProdSeq()][bIndex][eIndex - bIndex].add(res);
    }

    private ResElemList getRESULT(Production gp, int bIndex, int eIndex) {
        return this._results[gp.getProdSeq()][bIndex][eIndex - bIndex];
    }

    private double computeTransProb(Production prod, int cIndex) {
        return cIndex == 0 ? this._parameter.getProb(prod.getGProduction(), Word.CHILD_A) :
                this._parameter.getProb(prod.getGProduction(), Word.CHILD_B);
    }


    public void alignMax(int K) {

        this._K = K;
        this.resetMax();

        //varying length: 1 to N ...
        for (int len = 1; len <= this._N; len++) {
            //System.err.println("len:" + len + "/" + this._N);

            //the begin index of the span ...
            for (int bIndex = 0; bIndex <= this._N - len; bIndex++) {
                int eIndex = bIndex + len;

                for (int prod_seq = this._productions.size() - 1; prod_seq >= 0; prod_seq--) {
                    Production prod = this._productions.get(prod_seq);

                    int childrensize = prod.countChildren();

                    if (childrensize == 0) {

                        Word word = this.getWord(eIndex - 1);
                        int rulePattern = Rule.RULE_w;
                        double subProb;
                        if (len == 1) {
                            subProb = this.computeOneWordProb(prod, word);
                            this.setRes_w(prod, bIndex, eIndex, subProb);
                        } else {

                            ResElem res = this.getRes_w(prod, bIndex, eIndex - 1);
                            subProb = res.getProb() * this.computeOneWordProb(prod, word);
                            this.setRes_w(prod, bIndex, eIndex, subProb);
                        }
                        double prob = this.computeRuleTypeProb(prod, rulePattern) * subProb * this.computeEndProb(prod);
                        this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, new ResElem[0]);

                    } else if (childrensize == 1) {

                        Production child = prod.getChildren().get(0);

                        int cIndex = 0;
                        //X
                        {
                            int rulePattern = Rule.RULE_X;

                            ResElemList cResults = this.getRESULT(child, bIndex, eIndex);
                            for (int index = cResults.size() - 1; index >= 0; index--) {
                                ResElem cResult = cResults.get(index);
                                if (cResult.getProb() <= 0) {
                                    break;
                                }
                                double subProb = this.computeMRLLMProb(prod, cIndex, child) * this.computeTransProb(prod, cIndex) * cResult.getProb();

                                ResElem[] children = new ResElem[] {cResult};
                                this.addRes(this._results_X, prod, cIndex, bIndex, eIndex, subProb, rulePattern, children);
                            }
                        }

                        if (len >= 2) {

                            //w X
                            {
                                int rulePattern = Rule.RULE_wX;

                                Word word = this.getWord(bIndex);

                                // w | w X
                                if (len >= 3) {

                                    ResElemList cResults = this.getRes(this._results_wX, prod, cIndex, bIndex + 1, eIndex);
                                    for (int index = cResults.size() - 1; index >= 0; index--) {
                                        ResElem cResult = cResults.get(index);
                                        if (cResult.getProb() <= 0) {
                                            break;
                                        }
                                        double subProb = this.computeOneWordProb(prod, word) * cResult.getProb();
                                        if (!this.addRes(this._results_wX, prod, cIndex, bIndex, eIndex, subProb, rulePattern, cResult.getChildren())) {
                                            break;
                                        }
                                    }
                                }

                                // w | X
                                ResElemList cResults = this.getRes(this._results_X, prod, cIndex, bIndex + 1, eIndex);
                                for (int index = cResults.size() - 1; index >= 0; index--) {
                                    ResElem cResult = cResults.get(index);
                                    if (cResult.getProb() <= 0) {
                                        break;
                                    }
                                    double subProb = this.computeOneWordProb(prod, word) * cResult.getProb();
                                    if (!this.addRes(this._results_wX, prod, cIndex, bIndex, eIndex, subProb, rulePattern, cResult.getChildren())) {
                                        break;
                                    }
                                }
                            }

                            //X w
                            {
                                int rulePattern = Rule.RULE_Xw;

                                Word word = this.getWord(eIndex - 1);

                                // X w | w
                                if (len >= 3) {

                                    ResElemList cResults = this.getRes(this._results_Xw, prod, cIndex, bIndex, eIndex - 1);
                                    for (int index = cResults.size() - 1; index >= 0; index--) {
                                        ResElem cResult = cResults.get(index);
                                        if (cResult.getProb() <= 0) {
                                            break;
                                        }
                                        double subProb = this.computeOneWordProb(prod, word) * cResult.getProb();
                                        if (!this.addRes(this._results_Xw, prod, cIndex, bIndex, eIndex, subProb, rulePattern, cResult.getChildren())) {
                                            break;
                                        }
                                    }
                                }

                                // X | w
                                ResElemList cResults = this.getRes(this._results_X, prod, cIndex, bIndex, eIndex - 1);
                                for (int index = cResults.size() - 1; index >= 0; index--) {
                                    ResElem cResult = cResults.get(index);
                                    if (cResult.getProb() <= 0) {
                                        break;
                                    }
                                    double subProb = this.computeOneWordProb(prod, word) * cResult.getProb();
                                    if (!this.addRes(this._results_Xw, prod, cIndex, bIndex, eIndex, subProb, rulePattern, cResult.getChildren())) {
                                        break;
                                    }
                                }
                            }
                        }

                        if (len >= 3) {
                            // w1 X w2
                            {
                                int rulePattern = Rule.RULE_w1Xw2;

                                Word word = this.getWord(eIndex - 1);

                                // w1 X w2 | w
                                if (len >= 4) {

                                    ResElemList cResults = this.getRes(this._results_w1Xw2, prod, cIndex, bIndex, eIndex - 1);
                                    for (int index = cResults.size() - 1; index >= 0; index--) {
                                        ResElem cResult = cResults.get(index);
                                        if (cResult.getProb() <= 0) {
                                            break;
                                        }
                                        double subProb = cResult.getProb() * this.computeOneWordProb(prod, word);
                                        if (!this.addRes(this._results_w1Xw2, prod, cIndex, bIndex, eIndex, subProb, rulePattern, cResult.getChildren())) {
                                            break;
                                        }
                                    }
                                }

                                // w X | w
                                ResElemList cResults = this.getRes(this._results_wX, prod, cIndex, bIndex, eIndex - 1);
                                for (int index = cResults.size() - 1; index >= 0; index--) {
                                    ResElem cResult = cResults.get(index);
                                    if (cResult.getProb() <= 0) {
                                        break;
                                    }
                                    double subProb = cResult.getProb() * this.computeOneWordProb(prod, word);
                                    if (!this.addRes(this._results_w1Xw2, prod, cIndex, bIndex, eIndex, subProb, rulePattern, cResult.getChildren())) {
                                        break;
                                    }
                                }
                            }
                        }

                        //add global values
                        int rulePattern;
                        ResElemList cResults;

                        // X
                        rulePattern = Rule.RULE_X;
                        cResults = this.getRes(this._results_X, prod, cIndex, bIndex, eIndex);
                        for (int index = cResults.size() - 1; index >= 0; index--) {
                            ResElem cResult = cResults.get(index);
                            if (cResult.getProb() <= 0) {
                                break;
                            }
                            double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult.getProb() * this.computeEndProb(prod);

                            if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, cResult.getChildren())) {
                                break;
                            }
                        }

                        if (len >= 2) {
                            // w X



                            rulePattern = Rule.RULE_wX;
                            cResults = this.getRes(this._results_wX, prod, cIndex, bIndex, eIndex);
                            for (int index = cResults.size() - 1; index >= 0; index--) {
                                ResElem cResult = cResults.get(index);
                                if (cResult.getProb() <= 0) {
                                    break;
                                }
                                double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult.getProb() * this.computeEndProb(prod);

                                if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, cResult.getChildren())) {
                                    break;
                                }
                            }

                            // X w
                            rulePattern = Rule.RULE_Xw;
                            cResults = this.getRes(this._results_Xw, prod, cIndex, bIndex, eIndex);
                            for (int index = cResults.size() - 1; index >= 0; index--) {
                                ResElem cResult = cResults.get(index);
                                if (cResult.getProb() <= 0) {
                                    break;
                                }
                                double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult.getProb() * this.computeEndProb(prod);

                                if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, cResult.getChildren())) {
                                    break;
                                }
                            }
                        }

                        if (len >= 3) {
                            // w1 X w2



                            rulePattern = Rule.RULE_w1Xw2;
                            cResults = this.getRes(this._results_w1Xw2, prod, cIndex, bIndex, eIndex);
                            for (int index = cResults.size() - 1; index >= 0; index--) {
                                ResElem cResult = cResults.get(index);
                                if (cResult.getProb() <= 0) {
                                    break;
                                }
                                double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult.getProb() * this.computeEndProb(prod);

                                if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, cResult.getChildren())) {
                                    break;
                                }
                            }
                        }

                    } else if (childrensize == 2) {

                        if (len >= 2) {
                            // X | Y
                            // Y | X


                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 1; sIndex++) {
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    ResElemList cResults1 = this.getRes(this._results_X, prod, cIndex, bIndex, sIndex);
                                    ResElemList cResults2 = this.getRes(this._results_X, prod, 1 - cIndex, sIndex, eIndex);

                                    int rulePattern = cIndex == 0 ? Rule.RULE_XY : Rule.RULE_YX;

                                    boolean outerAttempt = true;
                                    for (int i = cResults1.size() - 1; i >= 0; i--) {
                                        ResElem cResult1 = cResults1.get(i);
                                        if (cResult1.getProb() <= 0) {
                                            break;
                                        }
                                        for (int j = cResults2.size() - 1; j >= 0; j--) {
                                            ResElem cResult2 = cResults2.get(j);
                                            if (cResult2.getProb() <= 0) {
                                                break;
                                            }
                                            double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult1.getProb() * cResult2.getProb() *
                                                    this.computeEndProb(prod);

                                            ResElem[] children = cIndex == 0 ? new ResElem[] {cResult1.getFirstChild(), cResult2.getFirstChild()}
                                                     : new ResElem[] {cResult2.getFirstChild(), cResult1.getFirstChild()};
                                            if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, children)) {
                                                if (j == cResults2.size() - 1) {
                                                    outerAttempt = false;
                                                }
                                                break;
                                            }
                                        }
                                        if (!outerAttempt) {
                                            break;
                                        }
                                    }
                                }
                            }
                        }

                        if (len >= 3) {
                            // w X | Y
                            // w Y | X



                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    ResElemList cResults1 = this.getRes(this._results_wX, prod, cIndex, bIndex, sIndex);
                                    ResElemList cResults2 = this.getRes(this._results_X, prod, 1 - cIndex, sIndex, eIndex);

                                    int rulePattern = cIndex == 0 ? Rule.RULE_wXY : Rule.RULE_wYX;

                                    boolean outerAttempt = true;
                                    for (int i = cResults1.size() - 1; i >= 0; i--) {
                                        ResElem cResult1 = cResults1.get(i);
                                        if (cResult1.getProb() <= 0) {
                                            break;
                                        }
                                        for (int j = cResults2.size() - 1; j >= 0; j--) {
                                            ResElem cResult2 = cResults2.get(j);
                                            if (cResult2.getProb() <= 0) {
                                                break;
                                            }
                                            double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult1.getProb() * cResult2.getProb() *
                                                    this.computeEndProb(prod);

                                            ResElem[] children = cIndex == 0 ? new ResElem[] {cResult1.getFirstChild(), cResult2.getFirstChild()}
                                                     : new ResElem[] {cResult2.getFirstChild(), cResult1.getFirstChild()};
                                            if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, children)) {
                                                if (j == cResults2.size() - 1) {
                                                    outerAttempt = false;
                                                }
                                                break;
                                            }
                                        }
                                        if (!outerAttempt) {
                                            break;
                                        }
                                    }
                                }
                            }

                            // X w | Y
                            // Y w | X
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 1; sIndex++) {
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    ResElemList cResults1 = this.getRes(this._results_Xw, prod, cIndex, bIndex, sIndex);
                                    ResElemList cResults2 = this.getRes(this._results_X, prod, 1 - cIndex, sIndex, eIndex);

                                    int rulePattern = cIndex == 0 ? Rule.RULE_XwY : Rule.RULE_YwX;

                                    boolean outerAttempt = true;
                                    for (int i = cResults1.size() - 1; i >= 0; i--) {
                                        ResElem cResult1 = cResults1.get(i);
                                        if (cResult1.getProb() <= 0) {
                                            break;
                                        }
                                        for (int j = cResults2.size() - 1; j >= 0; j--) {
                                            ResElem cResult2 = cResults2.get(j);
                                            if (cResult2.getProb() <= 0) {
                                                break;
                                            }
                                            double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult1.getProb() * cResult2.getProb() *
                                                    this.computeEndProb(prod);

                                            ResElem[] children = cIndex == 0 ? new ResElem[] {cResult1.getFirstChild(), cResult2.getFirstChild()}
                                                     : new ResElem[] {cResult2.getFirstChild(), cResult1.getFirstChild()};

                                            if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, children)) {
                                                if (j == cResults2.size() - 1) {
                                                    outerAttempt = false;
                                                }
                                                break;
                                            }
                                        }
                                        if (!outerAttempt) {
                                            break;
                                        }
                                    }
                                }
                            }

                            // X | Y w
                            // Y | X w
                            for (int sIndex = bIndex + 1; sIndex <= eIndex - 2; sIndex++) {
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    ResElemList cResults1 = this.getRes(this._results_X, prod, cIndex, bIndex, sIndex);
                                    ResElemList cResults2 = this.getRes(this._results_Xw, prod, 1 - cIndex, sIndex, eIndex);

                                    int rulePattern = cIndex == 0 ? Rule.RULE_XYw : Rule.RULE_YXw;





                                    boolean outerAttempt = true;
                                    for (int i = cResults1.size() - 1; i >= 0; i--) {
                                        ResElem cResult1 = cResults1.get(i);
                                        if (cResult1.getProb() <= 0) {
                                            break;
                                        }
                                        for (int j = cResults2.size() - 1; j >= 0; j--) {
                                            ResElem cResult2 = cResults2.get(j);
                                            if (cResult2.getProb() <= 0) {
                                                break;
                                            }
                                            double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult1.getProb() * cResult2.getProb() *
                                                    this.computeEndProb(prod);

                                            ResElem[] children = cIndex == 0 ? new ResElem[] {cResult1.getFirstChild(), cResult2.getFirstChild()}
                                                     : new ResElem[] {cResult2.getFirstChild(), cResult1.getFirstChild()};
                                            if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, children)) {
                                                if (j == cResults2.size() - 1) {
                                                    outerAttempt = false;
                                                }
                                                break;
                                            }
                                        }
                                        if (!outerAttempt) {
                                            break;
                                        }
                                    }
                                }
                            }

                        }

                        if (len >= 4) {

                            // X w1 | Y w2
                            // Y w1 | X w2



                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    ResElemList cResults1 = this.getRes(this._results_Xw, prod, cIndex, bIndex, sIndex);
                                    ResElemList cResults2 = this.getRes(this._results_Xw, prod, 1 - cIndex, sIndex, eIndex);

                                    int rulePattern = cIndex == 0 ? Rule.RULE_Xw1Yw2 : Rule.RULE_Yw1Xw2;






                                    boolean outerAttempt = true;
                                    for (int i = cResults1.size() - 1; i >= 0; i--) {
                                        ResElem cResult1 = cResults1.get(i);
                                        if (cResult1.getProb() <= 0) {
                                            break;
                                        }
                                        for (int j = cResults2.size() - 1; j >= 0; j--) {
                                            ResElem cResult2 = cResults2.get(j);
                                            if (cResult2.getProb() <= 0) {
                                                break;
                                            }
                                            double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult1.getProb() * cResult2.getProb() *
                                                    this.computeEndProb(prod);

                                            ResElem[] children = cIndex == 0 ? new ResElem[] {cResult1.getFirstChild(), cResult2.getFirstChild()}
                                                     : new ResElem[] {cResult2.getFirstChild(), cResult1.getFirstChild()};
                                            if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, children)) {
                                                if (j == cResults2.size() - 1) {
                                                    outerAttempt = false;
                                                }
                                                break;
                                            }
                                        }
                                        if (!outerAttempt) {
                                            break;
                                        }
                                    }
                                }
                            }

                            // w1 X | Y w2
                            // w1 Y | X w2
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    ResElemList cResults1 = this.getRes(this._results_wX, prod, cIndex, bIndex, sIndex);
                                    ResElemList cResults2 = this.getRes(this._results_Xw, prod, 1 - cIndex, sIndex, eIndex);

                                    int rulePattern = cIndex == 0 ? Rule.RULE_w1XYw2 : Rule.RULE_w1YXw2;






                                    boolean outerAttempt = true;
                                    for (int i = cResults1.size() - 1; i >= 0; i--) {
                                        ResElem cResult1 = cResults1.get(i);
                                        if (cResult1.getProb() <= 0) {
                                            break;
                                        }
                                        for (int j = cResults2.size() - 1; j >= 0; j--) {
                                            ResElem cResult2 = cResults2.get(j);
                                            if (cResult2.getProb() <= 0) {
                                                break;
                                            }
                                            double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult1.getProb() * cResult2.getProb() *
                                                    this.computeEndProb(prod);

                                            ResElem[] children = cIndex == 0 ? new ResElem[] {cResult1.getFirstChild(), cResult2.getFirstChild()}
                                                     : new ResElem[] {cResult2.getFirstChild(), cResult1.getFirstChild()};
                                            if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, children)) {
                                                if (j == cResults2.size() - 1) {
                                                    outerAttempt = false;
                                                }
                                                break;
                                            }
                                        }
                                        if (!outerAttempt) {
                                            break;
                                        }
                                    }
                                }
                            }

                            // w1 X | w2 Y
                            // w1 Y | w2 X
                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 2; sIndex++) {
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    ResElemList cResults1 = this.getRes(this._results_wX, prod, cIndex, bIndex, sIndex);
                                    ResElemList cResults2 = this.getRes(this._results_wX, prod, 1 - cIndex, sIndex, eIndex);

                                    int rulePattern = cIndex == 0 ? Rule.RULE_w1Xw2Y : Rule.RULE_w1Yw2X;






                                    boolean outerAttempt = true;
                                    for (int i = cResults1.size() - 1; i >= 0; i--) {
                                        ResElem cResult1 = cResults1.get(i);
                                        if (cResult1.getProb() <= 0) {
                                            break;
                                        }
                                        for (int j = cResults2.size() - 1; j >= 0; j--) {
                                            ResElem cResult2 = cResults2.get(j);
                                            if (cResult2.getProb() <= 0) {
                                                break;
                                            }
                                            double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult1.getProb() * cResult2.getProb() *
                                                    this.computeEndProb(prod);

                                            ResElem[] children = cIndex == 0 ? new ResElem[] {cResult1.getFirstChild(), cResult2.getFirstChild()}
                                                     : new ResElem[] {cResult2.getFirstChild(), cResult1.getFirstChild()};
                                            if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, children)) {
                                                if (j == cResults2.size() - 1) {
                                                    outerAttempt = false;
                                                }
                                                break;
                                            }
                                        }
                                        if (!outerAttempt) {
                                            break;
                                        }
                                    }
                                }
                            }
                        }

                        if (len >= 5) {

                            // w1 X | w2 Y w3
                            // w1 Y | w2 X w3


                            for (int sIndex = bIndex + 2; sIndex <= eIndex - 3; sIndex++) {
                                for (int cIndex = 0; cIndex <= 1; cIndex++) {
                                    ResElemList cResults1 = this.getRes(this._results_wX, prod, cIndex, bIndex, sIndex);
                                    ResElemList cResults2 = this.getRes(this._results_w1Xw2, prod, 1 - cIndex, sIndex, eIndex);

                                    int rulePattern = cIndex == 0 ? Rule.RULE_w1Xw2Yw3 : Rule.RULE_w1Yw2Xw3;






                                    boolean outerAttempt = true;
                                    for (int i = cResults1.size() - 1; i >= 0; i--) {
                                        ResElem cResult1 = cResults1.get(i);
                                        if (cResult1.getProb() <= 0) {
                                            break;
                                        }
                                        for (int j = cResults2.size() - 1; j >= 0; j--) {
                                            ResElem cResult2 = cResults2.get(j);
                                            if (cResult2.getProb() <= 0) {
                                                break;
                                            }
                                            double prob = this.computeRuleTypeProb(prod, rulePattern) * cResult1.getProb() * cResult2.getProb() *
                                                    this.computeEndProb(prod);

                                            ResElem[] children = cIndex == 0 ? new ResElem[] {cResult1.getFirstChild(), cResult2.getFirstChild()}
                                                     : new ResElem[] {cResult2.getFirstChild(), cResult1.getFirstChild()};
                                            if (!this.addRESULT(prod, bIndex, eIndex, prob, rulePattern, children)) {
                                                if (j == cResults2.size() - 1) {
                                                    outerAttempt = false;
                                                }
                                                break;
                                            }
                                        }
                                        if (!outerAttempt) {
                                            break;
                                        }
                                    }
                                }
                            }
                        }

                    } else {
                        System.err.println("Not Supported!");
                        System.exit(1);
                    }

                }

                //BEGIN OF POST PROCESS
                for (int prod_seq = this._productions.size() - 1; prod_seq >= 0; prod_seq--) {
                    Production prod = this._productions.get(prod_seq);
                    int childrenSize = prod.countRHSNonterminals();
                    if (childrenSize == 2) {
                        ArrayList<Production> children = prod.getChildren();
                        for (int cIndex = 0; cIndex <= 1; cIndex++) {

                            Production child = children.get(cIndex);

                            //X
                            {
                                int rulePattern = Rule.RULE_X;

                                ResElemList cResults = this.getRESULT(child, bIndex, eIndex);
                                for (int index = cResults.size() - 1; index >= 0; index--) {
                                    ResElem cResult = cResults.get(index);
                                    if (cResult.getProb() <= 0) {
                                        break;
                                    }
                                    double subProb = this.computeMRLLMProb(prod, cIndex, child) * this.computeTransProb(prod, cIndex) *
                                            cResult.getProb();
                                    this.addRes(this._results_X, prod, cIndex, bIndex, eIndex, subProb, rulePattern, new ResElem[] {cResult});
                                }
                            }

                            if (len >= 2) {

                                //w X
                                {
                                    int rulePattern = Rule.RULE_wX;
                                    Word word = this.getWord(bIndex);

                                    // w | w X
                                    if (len >= 3) {
                                        ResElemList cResults = this.getRes(this._results_wX, prod, cIndex, bIndex + 1, eIndex);
                                        for (int index = cResults.size() - 1; index >= 0; index--) {
                                            ResElem cResult = cResults.get(index);
                                            if (cResult.getProb() <= 0) {
                                                break;
                                            }
                                            double subProb = this.computeOneWordProb(prod, word) * cResult.getProb();
                                            if (!this.addRes(this._results_wX, prod, cIndex, bIndex, eIndex, subProb, rulePattern,
                                                    cResult.getChildren())) {
                                                break;
                                            }
                                        }
                                    }

                                    // w | X
                                    ResElemList cResults = this.getRes(this._results_X, prod, cIndex, bIndex + 1, eIndex);
                                    for (int index = cResults.size() - 1; index >= 0; index--) {
                                        ResElem cResult = cResults.get(index);
                                        if (cResult.getProb() <= 0) {
                                            break;
                                        }
                                        double subProb = this.computeOneWordProb(prod, word) * cResult.getProb();
                                        if (!this.addRes(this._results_wX, prod, cIndex, bIndex, eIndex, subProb, rulePattern, cResult.getChildren())) {
                                            break;
                                        }
                                    }
                                }

                                //X w
                                {
                                    int rulePattern = Rule.RULE_Xw;
                                    Word word = this.getWord(eIndex - 1);

                                    // X w | w
                                    if (len >= 3) {
                                        ResElemList cResults = this.getRes(this._results_Xw, prod, cIndex, bIndex, eIndex - 1);
                                        for (int index = cResults.size() - 1; index >= 0; index--) {
                                            ResElem cResult = cResults.get(index);
                                            if (cResult.getProb() <= 0) {
                                                break;
                                            }
                                            double subProb = this.computeOneWordProb(prod, word) * cResult.getProb();
                                            if (!this.addRes(this._results_Xw, prod, cIndex, bIndex, eIndex, subProb, rulePattern,
                                                    cResult.getChildren())) {
                                                break;
                                            }
                                        }
                                    }

                                    // X | w
                                    ResElemList cResults = this.getRes(this._results_X, prod, cIndex, bIndex, eIndex - 1);
                                    for (int index = cResults.size() - 1; index >= 0; index--) {
                                        ResElem cResult = cResults.get(index);
                                        if (cResult.getProb() <= 0) {
                                            break;
                                        }
                                        double subProb = this.computeOneWordProb(prod, word) * cResult.getProb();
                                        if (!this.addRes(this._results_Xw, prod, cIndex, bIndex, eIndex, subProb, rulePattern, cResult.getChildren())) {
                                            break;
                                        }
                                    }
                                }
                            }

                            if (len >= 3) {
                                // w1 X w2
                                {
                                    int rulePattern = Rule.RULE_w1Xw2;
                                    Word word = this.getWord(eIndex - 1);

                                    // w1 X w2 | w
                                    if (len >= 4) {
                                        ResElemList cResults = this.getRes(this._results_w1Xw2, prod, cIndex, bIndex, eIndex - 1);
                                        for (int index = cResults.size() - 1; index >= 0; index--) {
                                            ResElem cResult = cResults.get(index);
                                            if (cResult.getProb() <= 0) {
                                                break;
                                            }
                                            double subProb = this.computeOneWordProb(prod, word) * cResult.getProb();
                                            if (!this.addRes(this._results_w1Xw2, prod, cIndex, bIndex, eIndex, subProb, rulePattern,
                                                    cResult.getChildren())) {
                                                break;
                                            }
                                        }
                                    }

                                    // w X | w
                                    ResElemList cResults = this.getRes(this._results_wX, prod, cIndex, bIndex, eIndex - 1);
                                    for (int index = cResults.size() - 1; index >= 0; index--) {
                                        ResElem cResult = cResults.get(index);
                                        if (cResult.getProb() <= 0) {
                                            break;
                                        }
                                        double subProb = this.computeOneWordProb(prod, word) * cResult.getProb();
                                        if (!this.addRes(this._results_w1Xw2, prod, cIndex, bIndex, eIndex, subProb, rulePattern, cResult.getChildren())) {
                                            break;
                                        }
                                    }
                                }
                            }

                        }
                    }
                }
                //END OF POST PROCESS
            }
        }

        this._final_list = this.getRESULT(this._productions.get(0), 0, this._N);

        if (this._final_list.getBest().getProb() <= 0) {
            this._ignoreFlag = true;
        }
    }
}
