/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.data;

/**
 * @author Wei LU
 * @version 1.0
 */

import java.util.ArrayList;
import java.io.Serializable;

public abstract class Example implements Serializable {

	private static final long serialVersionUID = -1725168094035941772L;

	/**
     * the example id
     */
    protected int _exampleId;

    /**
     * the natural language sentence, in string form
     */
    protected String _nl;

    /**
     * the meaning representation language, in string form
     */
    protected String _mrl;

    /**
     * all the productions
     */
    protected ArrayList<Production> _productions;

    /**
     * all the words
     */
    protected ArrayList<Word> _words;


    /**
     * all the inner features associated with this Example
     */
    private ArrayList<InnerFeature> _innerFeatures;

    /**
     * all the first inner features
     */
    private ArrayList<InnerFeature> _firstInnerFeatures;

    /**
     * Get the Example's id
     * @return the id of this Example
     */
    public int getExampleId() {
        return this._exampleId;
    }

    protected Example(int exampleId, String nl, ArrayList<Word> words, String mrl,
    		ArrayList<Production> productions) {
        this._exampleId = exampleId;
        this._nl = nl;
        this._words = words;
        this._mrl = mrl;
        this._productions = productions;
    }

    /**
     * Verify the correctness of Productions
     * @param prods Vector
     * @return boolean
     */
    public boolean verifyCorrect(ArrayList<Production> prods) {
        Production prod_correct = this._productions.get(0);
        Production prod_predict = prods.get(0);
        return verifyCorrectHelper(prod_correct, prod_predict);
    }

    /**
     * Verify the correctness of two Productions
     * @param prod_correct Production
     * @param prod_predict Production
     * @return boolean
     */
    private boolean verifyCorrectHelper(Production prod_correct,
                                        Production prod_predict) {
        if (!prodEquals(prod_correct.getGProduction(),
                        prod_predict.getGProduction())) {
            return false;
        }
        GProduction gp = prod_correct.getGProduction();

        ArrayList<Production> children_correct = prod_correct.getChildren();
        ArrayList<Production> children_predict = prod_predict.getChildren();

        boolean result = true;

        if (gp.isCommutative()) {
            Production child_correct_0 = children_correct.get(0);
            Production child_correct_1 = children_correct.get(1);

            Production child_predict_0 = children_predict.get(0);
            Production child_predict_1 = children_predict.get(1);

            result &= this.verifyCorrectHelper(child_correct_0, child_predict_0) &&
                    this.verifyCorrectHelper(child_correct_1, child_predict_1) ||
                    this.verifyCorrectHelper(child_correct_0, child_predict_1) &&
                    this.verifyCorrectHelper(child_correct_1, child_predict_0);
        } else {
            for (int i = 0; i < children_correct.size(); i++) {
                Production child_correct = children_correct.get(i);
                Production child_predict = children_predict.get(i);
                result &= verifyCorrectHelper(child_correct, child_predict);
            }
        }
        return result;
    }

    /**
     * Check the equality of two GProductions
     * @param gp1 GProduction
     * @param gp2 GProduction
     * @return boolean
     */
    private boolean prodEquals(GProduction gp1, GProduction gp2) {

        if (gp1.toLine().equals(gp2.toLine())) {
            return true;
        }

        if (gp1.getLHS().getName().startsWith("*n:Num") &&
            gp2.getLHS().getName().startsWith("*n:Num")) {
            try {
                double value1 = Double.parseDouble(gp1.getRHS().get(0).getName());
                double value2 = Double.parseDouble(gp2.getRHS().get(0).getName());
                if (value1 == value2) {
                    return true;
                }
                return false;
            } catch (NumberFormatException e) {
                return false;
            }
        }

        return false;

    }


    public ArrayList<InnerFeature> getFirstInnerFeatures() {
        if (this._firstInnerFeatures != null) {
            return this._firstInnerFeatures;
        }

        this._firstInnerFeatures = new ArrayList<InnerFeature>();
        for (int i = 0; i < this._productions.size(); i++) {
            Production prod = this._productions.get(i);
            ArrayList<InnerFeature>
                    features = prod.getGProduction().getInnerFeatures();
            this._firstInnerFeatures.add(features.get(0));
        }

        return this._firstInnerFeatures;
    }

    public ArrayList<InnerFeature> getInnerFeatures() {
        if (this._innerFeatures != null) {
            return this._innerFeatures;
        }

        this._innerFeatures = new ArrayList<InnerFeature>();
        for (int i = 0; i < this._productions.size(); i++) {
            Production prod = this._productions.get(i);
            ArrayList<InnerFeature>
                    features = prod.getGProduction().getInnerFeatures();
            for (int j = 0; j < features.size(); j++) {
                this._innerFeatures.add(features.get(j));
            }
        }

        return this._innerFeatures;
    }


    /**
     * Get the natural language sentence in a String representation
     * @return the natural language sentence
     */
    public String getNL() {
        return this._nl;
    }

    /**
     * Set the string representation of the natural language sentence
     * @param nl the natural language sentence
     */
    public void setNL(String nl) {
        this._nl = nl;
    }

    /**
     * Get all the words
     * @return all the words
     */
    public ArrayList<Word> getWords() {
        return this._words;
    }

    /**
     * Set all the words
     * @param words Vector
     */
    public void setWords(ArrayList<Word> words) {
        this._words = words;
    }


    /**
     * Get the meaning representation in a String representation
     * @return the natural language sentence
     */
    public String getMRL() {
        return this._mrl;
    }
}
