/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.data;

/**
 * @author LU Wei
 * @version 1.0
 */
import java.io.Serializable;
import java.util.ArrayList;

public class GProduction implements Serializable {

	private static final long serialVersionUID = 3077267120231736200L;

	/**
     * The LHS nonterminal node
     */
    private NonterminalNode _lhs;

    /**
     * The RHS nodes
     */
    private ArrayList<Node> _rhs;

    /**
     * The nonterminal symbols that appeared in the RHS
     */
    private ArrayList<NonterminalNode> _RHSNonterminals = null;

    /**
     * The predicate nodes that appeared in the RHS
     */
    private ArrayList<String> _predicates = null;

    /**
     * The global Id for this GProduction
     */
    private int _prodId = -1;

    /**
     * this GProduction's internal features
     */
    private ArrayList<InnerFeature> _innerFeatures;

    /**
     * the parent associated with this GProduction
     */
    private GProduction _parent;

    /**
     * whether the two children are interchangable
     */
    private boolean _commutativeFlag = false;

    private boolean _isConstant = false;

    /**
     * Constructor
     * @param lhs the LHS nonterminal node
     * @param rhs a collection of RHS nodes
     */
    protected GProduction(NonterminalNode lhs, ArrayList<Node> rhs) {
        this._lhs = lhs;
        this._rhs = rhs;
    }

    /**
     * Check if this GProduction is a constant GProduction
     * @return boolean
     */
    public boolean isConstant() {
        return this._isConstant;
    }

    /**
     * Set this GProduction as constant GProduction
     */
    public void setConstant() {
        this._isConstant = true;
    }

    /**
     * Set this GProduction as commutative
     */
    public void setCommutative() {
        this._commutativeFlag = true;
    }

    /**
     * Check whether this GProduction is commutative
     * @return boolean
     */
    public boolean isCommutative() {
        return this._commutativeFlag;
    }

    /**
     * Set the parent for this GProduction
     * @param parent GProduction
     */
    public void setParent(GProduction parent) {
        this._parent = parent;
    }

    /**
     * Get the parent of this GProduction
     * @return GProduction
     */
    public GProduction getParent() {
        return this._parent;
    }

    /**
     * Set the features
     * @param innerFeatures a collection of inner features
     */
    public void setInnerFeatures(ArrayList<InnerFeature> innerFeatures) {
        this._innerFeatures = innerFeatures;
    }

    /**
     * Get all the inner features
     * @return ArrayList
     */
    public ArrayList<InnerFeature> getInnerFeatures() {
        return this._innerFeatures;
    }

    /**
     * Get the LHS index
     * @return the index of the LHS nonterminal
     */
    public int getLHSNontIndex() {
        return this._lhs.getNodeIndex();
    }

    /**
     * Set the GProduction's id
     * @param prodId int
     */
    public void setProdId(int prodId) {
        this._prodId = prodId;
    }

    /**
     * Get the GProduction's id
     * @return int
     */
    public int getProdId() {
        return this._prodId;
    }

    /**
     * Check whether this is a TerminalNode
     * @return boolean
     */
    public boolean isTerminal() {
        return this.countRHSNonterminals() == 0;
    }

    /*
    public boolean isConst() {
        if (this._lhs.getName().equals("*n:CountryName")) {
            return true;
        } else if (this._lhs.getName().equals("*n:StateName")) {
            return true;
        } else if (this._lhs.getName().equals("*n:StateAbbrev")) {
            return true;
        } else if (this._lhs.getName().equals("*n:RiverName")) {
            return true;
        } else if (this._lhs.getName().equals("*n:PlaceName")) {
            return true;
        } else if (this._lhs.getName().equals("*n:CityName")) {
            return true;
        } else if (this._lhs.getName().equals("*n:AIRCRAFT_CODE0")) {
            return true;
        } else {
            if (this._lhs.getName().equals("*n:E") && this._rhs.size() == 1) {
                Node rhs = this._rhs.get(0);
                if (rhs instanceof LambdaTerminalNode) {
                    LambdaTerminalNode lrhs = (LambdaTerminalNode) rhs;
                    Expr rhsExpr = lrhs.getLambdaExpr();
                    if (rhsExpr instanceof Identifier) {
                        return true;
                    }
                }
            }
            return false;
        }
    }
    */

    public boolean isNum() {
        return this._lhs.getName().equals("*n:Num") &&
                this.getRHSNonterminals().size() == 0;
    }

    public boolean isUnum() {
        return this._lhs.getName().equals("*n:Unum");
    }

    public boolean isIdent() {
        return this._lhs.getName().equals("*n:Ident");
    }

    public void setLHS(NonterminalNode lhs) {
        this._lhs = lhs;
    }

    public void setRHS(ArrayList<Node> rhs) {
        this._rhs = rhs;
    }

    /**
     * Get the predicate symbol(s)
     * @return ArrayList
     */
    public ArrayList<String> getPredicates() {
        if (_predicates != null) {
            return this._predicates;
        } else {
            this._predicates = new ArrayList<String>();
            for (int i = 0; i < this._rhs.size(); i++) {
                Node rhsNode = this._rhs.get(i);
                if (rhsNode instanceof TerminalNode &&
                           isMeaningful(rhsNode.getName())) {
                    this._predicates.add(rhsNode.getName());
                }
            }
            return this._predicates;
        }
    }

    private boolean isMeaningful(String s) {
        return Character.isLetter(s.charAt(0));
    }

    /**
     * Count the number of RHS NonterminalNodes
     * @return int
     */
    public int countRHSNonterminals() {
        if (_RHSNonterminals != null) {
            return this._RHSNonterminals.size();
        } else {
            this._RHSNonterminals = getRHSNonterminals();
            return _RHSNonterminals.size();
        }
    }

    /**
     * Get all the RHS NonterminalNodes
     * @return ArrayList
     */
    public ArrayList<NonterminalNode> getRHSNonterminals() {
        if (_RHSNonterminals != null) {
            return this._RHSNonterminals;
        } else {
            this._RHSNonterminals = new ArrayList<NonterminalNode>();
            for (int i = 0; i < this._rhs.size(); i++) {
                Node rhsNode = (Node)this._rhs.get(i);
                if (rhsNode instanceof NonterminalNode) {
                    this._RHSNonterminals.add((NonterminalNode) rhsNode);
                }
            }
            return this._RHSNonterminals;
        }
    }

    /**
     * Get the LHS nonterminal node
     * @return NonterminalNode
     */
    public NonterminalNode getLHS() {
        return this._lhs;
    }

    /**
     * Get the RHS nodes
     * @return ArrayList
     */
    public ArrayList<Node> getRHS() {
        return this._rhs;
    }

    /**
     * Print the GProduction as a line.
     * @return String
     */
    public String toLine() {
        StringBuffer sb = new StringBuffer();

        sb.append(this.getLHS());
        sb.append(" -> ({ ");
        for (int i = 0; i < this.getRHS().size(); i++) {
            sb.append(this.getRHS().get(i));
            sb.append(' ');
        }
        sb.append("})");

//        sb.append(this.getLHS().getName() + " -> " + this.getRHS());
        return sb.toString();
    }

    public boolean equals(Object o) {
        if (!(o instanceof GProduction)) {
            return false;
        }
        GProduction gp = (GProduction) o;
        if (!this._lhs.equals(gp._lhs)) {
            return false;
        }
        if (!this._rhs.equals(gp._rhs)) {
            return false;
        }
        return true;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append(this.toLine());
        sb.append('|');
        sb.append(this._prodId);
        return sb.toString();
    }

}
