/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.data;

import java.util.ArrayList;
import java.io.Serializable;


import java.util.Hashtable;

public class Model implements Serializable {
	
	private static final long serialVersionUID = -7372556900958156439L;

	/**
     * all the Words
     */
    private ArrayList<Word> _allWords;

    /**
     * all the GProductions
     */
    private ArrayList<GProduction> _allGProds;

    /**
     * all the Nodes
     */
    private ArrayList<Node> _allNodes;

    /**
     * all the inner Features
     */
    private ArrayList<InnerFeature> _allInnerFeatures;

    /**
     * all the outer Features
     */
    private ArrayList<OuterFeature> _allOuterFeatures;

    /**
     * all the train examples
     */
    private ArrayList<Example> _trainExamples;

    /**
     * all the test examples
     */
    private ArrayList<Example> _testExamples;

    /**
     * sorted GProductions
     */
    private ArrayList<GProduction> _sortedGProds;

    /**
     * Constructor
     */
    public Model() {
        this._allInnerFeatures = new ArrayList<InnerFeature>();
        this._allOuterFeatures = new ArrayList<OuterFeature>();
        this._allGProds = new ArrayList<GProduction>();
        this._sortedGProds = new ArrayList<GProduction>();
        this._allNodes = new ArrayList<Node>();
        this._allWords = new ArrayList<Word>();
        this._trainExamples = new ArrayList<Example>();
        this._testExamples = new ArrayList<Example>();

        //insert initial words
        this._allWords.add(Word.BEGIN);
        this._allWords.add(Word.END);
        this._allWords.add(Word.CHILD_A);
        this._allWords.add(Word.CHILD_B);
        this._allWords.add(Word.UNKNOWN);
    }


    private Hashtable<Integer, ArrayList<GProduction>> _possibleChildrenTable = new Hashtable<Integer,
                                               ArrayList<GProduction>>();

    /**
     * You need to make sure that prods does not contain repeated GProductions
     */
    public ArrayList<GProduction> getPossibleChildren(GProduction prod,
            ArrayList<GProduction> prods, int cIndex) {
        Integer key = new Integer(prod.getProdId() * 1000 + cIndex);
        if (this._possibleChildrenTable.containsKey(key)) {
            return (ArrayList<GProduction>)this._possibleChildrenTable.get(key);
        }

        ArrayList<GProduction> children = new ArrayList<GProduction>();
        ArrayList<NonterminalNode> rhs_nodes = prod.getRHSNonterminals();
        Node rhs_node = rhs_nodes.get(cIndex);
        for (int i = 0; i < prods.size(); i++) {
            GProduction gp = prods.get(i);
            if (gp.getLHS().getName().equals(rhs_node.getName())) {
                children.add(gp);
            }
        }

        this._possibleChildrenTable.put(key, children);

        return children;
    }


    /**
     * Add the training example
     * @param example a training example
     */
    public void addTrainExample(TrainExample example) {
        this._trainExamples.add(example);
    }

    /**
     * Add the testing example
     * @param example a testing example
     */
    public void addTestExample(TestExample example) {
        this._testExamples.add(example);
    }

    /**
     * Get all the training examples
     * @return ArrayList
     */
    public ArrayList<Example> getTrainExamples() {
        return this._trainExamples;
    }

    /**
     * Get all the testing examples
     * @return ArrayList
     */
    public ArrayList<Example> getTestExamples() {
        return this._testExamples;
    }

    /**
     * Get all the words
     * @return ArrayList
     */
    public ArrayList<Word> getAllWords() {
        return this._allWords;
    }

    /**
     * Get all the GProductions
     * @return ArrayList
     */
    public ArrayList<GProduction> getAllGProds() {
        return this._allGProds;
    }

    /**
     * Get all the Nodes
     * @return ArrayList
     */
    public ArrayList<Node> getAllNodes() {
        return this._allNodes;
    }

    /**
     * Get all the inner features
     * @return ArrayList
     */
    public ArrayList<InnerFeature> getAllInnerFeatures() {
        return this._allInnerFeatures;
    }

    /**
     * Get all the outer features
     * @return ArrayList
     */
    public ArrayList<OuterFeature> getAllOuterFeatures() {
        return this._allOuterFeatures;
    }

    /**
     * Get the indexed word
     * @param w String
     * @return Word
     */
    public Word createWord(String w) {
        Word word = new Word(w);
        int wIndex = this._allWords.indexOf(word);
        if (wIndex == -1) {
            word.setWordId(this._allWords.size());
            this._allWords.add(word);
            return word;
        } else {
            return this._allWords.get(wIndex);
        }
    }

    /**
     * Attempt to create a word, if it appeared before, return that word,
     * if not, return UNKNOWN word
     * @param w String
     * @return Word
     */
    public Word attemptCreateWord(String w) {
        Word word = new Word(w);
        int wIndex = this._allWords.indexOf(word);
        if (wIndex == -1) {
            return Word.UNKNOWN;
        } else {
            return this._allWords.get(wIndex);
        }
    }


    /**
     * Remove the newly added word
     */
    public void removeNewWord() {
        this._allWords.remove(this._allWords.size() - 1);
    }

    /**
     * Get the indexed DoubleTerminalNode
     * @param num double
     * @return DoubleTerminalNode
     */
    public DoubleTerminalNode createDoubleTerminalNode(double num) {
        DoubleTerminalNode node = new DoubleTerminalNode(num);
        int nIndex = this._allNodes.indexOf(node);
        if (nIndex == -1) {
            node.setNodeIndex(this._allNodes.size());
            this._allNodes.add(node);
            return node;
        } else {
            return (DoubleTerminalNode)this._allNodes.get(nIndex);
        }
    }

    /**
     * Get the indexed IntTerminalNode
     * @param num int
     * @return IntTerminalNode
     */
    public IntTerminalNode createIntTerminalNode(int num) {
        IntTerminalNode node = new IntTerminalNode(num);
        int nIndex = this._allNodes.indexOf(node);
        if (nIndex == -1) {
            node.setNodeIndex(this._allNodes.size());
            this._allNodes.add(node);
            return node;
        } else {
            return (IntTerminalNode)this._allNodes.get(nIndex);
        }
    }

    /**
     * Get the indexed TerminalNode
     * @param n string representation of the node
     * @return TerminalNode
     */
    public TerminalNode createTerminalNode(String n) {
        TerminalNode node = new TerminalNode(n);
        int nIndex = this._allNodes.indexOf(node);
        if (nIndex == -1) {
            node.setNodeIndex(this._allNodes.size());
            this._allNodes.add(node);
            return node;
        } else {
            return (TerminalNode)this._allNodes.get(nIndex);
        }
    }

    /**
     * Get the indexed NonterminalNode
     * @param n string representation of the node
     * @return NonterminalNode
     */
    public NonterminalNode createNonterminalNode(String n) {
        NonterminalNode node = new NonterminalNode(n);
        int nIndex = this._allNodes.indexOf(node);
        if (nIndex == -1) {
            node.setNodeIndex(this._allNodes.size());
            this._allNodes.add(node);
            return node;
        } else {
            return (NonterminalNode)this._allNodes.get(nIndex);
        }
    }

    /**
     * Get the indexed GProduction
     * @param lhs the LHS NonterminalNode
     * @param rhs the RHS collecion of Node
     * @return the GProduction
     */
    public GProduction createGProduction(NonterminalNode lhs, ArrayList<Node> rhs) {
        GProduction gp = new GProduction(lhs, rhs);
        int gpIndex = this._allGProds.indexOf(gp);
        if (gpIndex == -1) {
            gp.setProdId(this._allGProds.size());
            this._allGProds.add(gp);
            extractInnerFeatures(gp);
        } else {
            gp = this._allGProds.get(gpIndex);
        }

        return gp;
    }

    /**
     * Get the indexed GProduction
     * @param lhs the LHS NonterminalNode
     * @param rhs the RHS collecion of Node
     * @return the GProduction
     */
    public GProduction attemptCreateGProduction(NonterminalNode lhs,
                                                ArrayList<Node> rhs) {
        GProduction gp = new GProduction(lhs, rhs);
        int gpIndex = this._allGProds.indexOf(gp);
        if (gpIndex == -1) {
            return null;
        } else {
            gp = this._allGProds.get(gpIndex);
        }

        return gp;
    }

    public Hashtable<GProduction,
            Integer> getPossibleParentGProductions(GProduction gp) {
        Hashtable<GProduction,
                Integer> results = new Hashtable<GProduction, Integer>();

        NonterminalNode lhs = gp.getLHS();
        for (int i = 0; i < this._allGProds.size(); i++) {
            GProduction curr = this._allGProds.get(i);
            ArrayList<NonterminalNode> rhsNonts = curr.getRHSNonterminals();
            for (int cIndex = 0; cIndex < rhsNonts.size(); cIndex++) {
                NonterminalNode rhsNont = rhsNonts.get(cIndex);
                if (rhsNont.equals(lhs)) {
                    results.put(curr, cIndex);
                }
            }
        }
        return results;
    }


    //this method can be further extended...
    /**
     * @todo
     * @param gp GProduction
     */
    private void extractInnerFeatures(GProduction gp) {
        ArrayList<InnerFeature> innerFeatures = new ArrayList<InnerFeature>();

        //the whole gp as feature
        InnerFeature f1 = indexInnerFeature("f1:" + gp.toLine());
        innerFeatures.add(f1);

        /**/
        //the predicate as feature
        ArrayList<String> predicates = gp.getPredicates();

        if (predicates.size() > 0) {
            InnerFeature f2 = indexInnerFeature("f2:" + predicates.toString());
            innerFeatures.add(f2);
        }

        /*
            ArrayList<Node> lambdaNodes = gp.getLambdaNontNodes();
            if (lambdaNodes.size() > 0) {
                InnerFeature f3 = indexInnerFeature("f3:" +
                        lambdaNodes.toString());
                innerFeatures.add(f3);
            }

                 //the lambda nodes as feature

                 //the rhs nodes as feature
                 ArrayList<Node> rhsNodes = gp.getRHS();
                 if(rhsNodes.size()>0){
            InnerFeature f4 = indexInnerFeature("f4:"+rhsNodes.toString());
            innerFeatures.add(f4);
                 }
         */

        gp.setInnerFeatures(innerFeatures);
    }

    /**
     * Index the InnerFeature
     * @param f the Feature to be indexed
     * @return the resulting Feature, after indexing
     */
    public InnerFeature indexInnerFeature(String f) {
        InnerFeature feature = new InnerFeature(f);
        int innerFeatureId = this._allInnerFeatures.indexOf(feature);
        if (innerFeatureId == -1) {
            feature.setFeatureId(this._allInnerFeatures.size());
            this._allInnerFeatures.add(feature);
            return feature;
        } else {
            return this._allInnerFeatures.get(innerFeatureId);
        }
    }

    /**
     * Index the OuterFeature
     * @param f the Feature to be indexed
     * @return the resulting Feature, after indexing
     */
    public OuterFeature indexOuterFeature(String f) {
        OuterFeature feature = new OuterFeature(f);
        int outerFeatureId = this._allOuterFeatures.indexOf(feature);
        if (outerFeatureId == -1) {
            feature.setFeatureId(this._allOuterFeatures.size());
            this._allOuterFeatures.add(feature);
            return feature;
        } else {
            return this._allOuterFeatures.get(outerFeatureId);
        }
    }


    /**
     * Sorting the productions for parsing.
     * just to make sure the parsing is reasonable. i.e. in correct order.
     * @todo
     * @todo This method needs improvement.
     * @return ArrayList
     */
    public ArrayList<GProduction> sortProds() {

        if (this._sortedGProds.size() != 0) {
            return this._sortedGProds;
        }

        //WARNING... FOR THE MOMENT...dirty fix...
        for (int k = 0; k < 1; k++) {
            for (int i = 0; i < this._allGProds.size(); i++) {
                GProduction gp = this._allGProds.get(this._allGProds.size() - 1 -
                        i);
                this._sortedGProds.add(gp);
            }
        }

        return this._sortedGProds;
        
        /*
        if (true) {
            return this._sortedGProds;
        }

        
        System.err.println("Started sorting " + this._allGProds.size() + " productions.");

        ArrayList<GProduction> additional = new ArrayList<GProduction>();

        boolean stablized = false;
        while (!stablized) {
            stablized = true;
            for (int i = 0; i < _sortedGProds.size(); i++) {
                GProduction prod = (GProduction) _sortedGProds.get(i);
                String lhs = prod.getLHS().getName();
                if (prod.countRHSNonterminals() == 1) {
                    String child = ((NonterminalNode) prod.getRHSNonterminals().
                                    get(0)).getName();
                    //if(!child.equals(lhs))
                    {
                        for (int j = i + 1; j < _sortedGProds.size(); j++) {
                            GProduction f_prod = _sortedGProds.get(j);
                            String f_lhs = f_prod.getLHS().getName();
                            //move it to front...
                            if (f_lhs.equals(child)) {
                                if (f_prod.countRHSNonterminals() != 1 ||
                                    !(f_prod.getRHSNonterminals().get(0).
                                      getName().equals(lhs))) {
                                    _sortedGProds.remove(j);
                                    _sortedGProds.add(i, f_prod);
                                    stablized = false;
                                    //i++;
                                    System.err.println(prod.toLine() + "\t" +
                                            f_prod.toLine());
                                }
                            } else if (f_prod.countRHSNonterminals() == 1 &&
                                       f_prod.getRHSNonterminals().get(0).
                                       getName().equals(lhs)) {
                                if (!additional.contains(prod)) {
                                    additional.add(prod);
                                }
                            }
                        }
                    }
                }
            }
            //System.err.println("status: "+stablized);
        }

        for (int i = 0; i < additional.size(); i++) {
            _sortedGProds.add(additional.get(i));
        }

        System.err.println("Sorting completed!");

        return _sortedGProds;
        
        */
    }


}
