/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.data;

/**
 * @author LU Wei
 * @version 1.0
 */
import java.io.Serializable;
import java.util.ArrayList;

public class Production implements Serializable {
	
	private static final long serialVersionUID = 5723655095327425539L;

	/**
     * the unique GProduction associated with this Production
     */
    private GProduction _gp; //the gproduction

    /**
     * its parent Production
     */
    private Production _parent;

    /**
     * its children Productions
     */
    private ArrayList<Production> _children;

    /**
     * its sequence number when appearing in a Production tree
     */
    private int _prodSeq = -1;

    /**
     * Constructor
     * @param gp the unique global GProduction
     */
    public Production(GProduction gp) {
        this._gp = gp;
        this._children = new ArrayList<Production>();
    }

    /**
     * Count the number of children Productions
     * @return the number of children Productions
     */
    public int countChildren() {
        if (this._children == null) {
            return 0;
        }
        return this._children.size();
    }

    /**
     * Get the LHS NonterminalNode
     * @return the Production's LHS NonterminalNode
     */
    public NonterminalNode getLHS() {
        return this._gp.getLHS();
    }

    /**
     * Set its LHS NonterminalNode
     * @param LHS the LHS NonterminalNode to be set
     */
    public void setLHS(NonterminalNode LHS) {
        this._gp.setLHS(LHS);
    }

    /**
     * Get its RHS Nodes
     * @return a collection of RHS Nodes
     */
    public ArrayList<Node> getRHS() {
        return this._gp.getRHS();
    }

    /**
     * Get its RHS NonterminalNodes
     * @return a collection of RHS NonterminalNodes
     */
    public ArrayList<NonterminalNode> getRHSNonterminals() {
        return this._gp.getRHSNonterminals();
    }

    /**
     * Set its GProduction
     * @param gp the GProduction to be set
     */
    public void setGproduction(GProduction gp) {
        this._gp = gp;
    }

    /**
     * Count the number of RHS NonterminalNodes
     * @return the count
     */
    public int countRHSNonterminals() {
        return this._gp.countRHSNonterminals();
    }

    /**
     * Get the GProduction
     * @return the GProduction associated with this Production
     */
    public GProduction getGProduction() {
        return this._gp;
    }

    /**
     * Set the parent Production
     * @param parent the parent Production
     */
    public void setParent(Production parent) {
        this._parent = parent;
    }

    /**
     * Get the parent Production
     * @return the parent Production
     */
    public Production getParent() {
        return this._parent;
    }

    /**
     * Set the children Productions
     * @param children a collection of children Productions
     */
    public void setChildren(ArrayList<Production> children) {
        this._children = children;
    }

    /**
     * Get the children Productions
     * @return the children Productions
     */
    public ArrayList<Production> getChildren() {
        return this._children;
    }

    /**
     * Set the Production sequence number when appearing in a Production tree
     * @param prodSeq the Production's sequence number
     */
    public void setProdSeq(int prodSeq) {
        this._prodSeq = prodSeq;
    }

    /**
     * Get the Production's sequence number
     * @return the Production's sequence number
     */
    public int getProdSeq() {
        return this._prodSeq;
    }

    /**
     * Get the Production's Id, which is also the Id of its associated GProduction
     * @return this Production's Id
     */
    public int getProdId() {
        return this._gp.getProdId();
    }

    /**
     * Get all the children Productions that appear under this Production
     * @return a collection of all children Productions
     */
    public ArrayList<Production> getAllChildren() {
    	ArrayList<Production> allChildren = new ArrayList<Production>();
        allChildren.add(this);

        ArrayList<Production> children = this.getChildren();
        if (children != null) {
            for (int i = 0; i < children.size(); i++) {
                Production child = children.get(i);
                ArrayList<Production> pv = child.getAllChildren();
                for (int j = 0; j < pv.size(); j++) {
                    allChildren.add(pv.get(j));
                }
            }
        }
        return allChildren;
    }

    /**
     * Get the index for this Production as child of its parent
     * @return int
     */
    public int getAsChildIndex() {
        if (this._parent == null) {
            return -1;
        }
        return this._parent.getChildren().indexOf(this);
    }

    /**
     * Get the Context
     * @return the Context
     */
    public Context getContext() {
        return Context.extractContext(this);
    }

    /**
     * Print this Production as a single line
     * @return a single line representation of the Production
     */
    public String toLine() {
        return this._gp.toLine();
    }

    /**
     * Print this Production as a tree structure
     * @return a tree structure representation of this Production as well as
     * all the Productions below it.
     */
    public String toTree() {
        StringBuffer sb = new StringBuffer();
        toTreeHelper(0, sb);
        return sb.toString();
    }

    private void toTreeHelper(int indent, StringBuffer sb) {
        for (int i = 0; i < indent; i++) {
            sb.append('\t');
        }

        sb.append(this._gp.toLine());
        sb.append('\n');

        for (int i = 0; i < this.countChildren(); i++) {
            Production child = this._children.get(i);
            child.toTreeHelper(indent + 1, sb);
        }
    }

    public String toString() {
        return this._prodSeq + "|" + this._gp.toString();
    }

}

