/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.eval;


/**
 * @author LU, Wei
 * @version 1.0
 */

import hybridtree.core.*;
import hybridtree.core.perceptron.*;
import hybridtree.data.Example;
import hybridtree.data.GProduction;
import hybridtree.data.Model;
import hybridtree.data.Production;
import hybridtree.data.reader.ModelReader;
import hybridtree.util.Utility;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.PrintWriter;

import java.util.ArrayList;



import java.io.File;

public class PerceptTest {

    private static String testMode = "parse";

    private static String gram = "bigram";

    private static String corpus = "x";

    private static int iteration = 99;

    private static int FOLD_START = 0;

    private static int FOLD_END = 9;

    private static int RUN = 0;

    private static int K = 50;

    private static String lang = "";

    private static int numTotal = -1;

    private static int numTrain = -1;

    private static void help() {
        System.err.println("Useage: java PerceptTest [options]");
        System.err.println("\toptions:");
        System.err.println(
                "\t\t corpus=robocup/geofunql (must be specified)");
        System.err.println("\t\t testMode=1,2,.. (default:parse)");
        System.err.println("\t\t run=0-9 (default:0)");
        System.err.println("\t\t folds=e.g.:3-8 (default:0-9)");
        System.err.println("\t\t iterations=10,20,.. (default:100)");
        System.err.println("\t\t gram=unigram/bigram (default:bigram)");
    }

    private static void dispArgs() {
        System.out.println("run:\t" + RUN);
        System.out.println("folds:" + FOLD_START + "~" + FOLD_END);
        System.out.println("corpus:\t" + corpus);
        System.out.println("testMode:\t" + testMode);
        System.out.println("gram:\t" + gram);
        System.out.println("parameter at iteration " + Utility.norm(iteration) +
                           " used.");
    }

    private static void readArgs(String args[]) {
        if (args.length % 2 != 0) {
            System.err.println("Invalid Argument: " + args[args.length - 1]);
            help();
            System.exit(1);
        }

        for (int i = 0; i < args.length; i += 2) {
            String option = args[i];
            String value = args[i + 1];
            if (option.equals("-testMode")) {
                if (value.equals("align")) {
                    testMode = "align";
                } else if (value.equals("parse")) {
                    testMode = "parse";
                } else {
                    System.err.println("Value " + value +
                                       " for testMode not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-run")) {
                try {
                    RUN = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for run not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-folds")) {
                try {
                    int index = value.indexOf("-");
                    FOLD_START = Integer.parseInt(value.substring(0, index));
                    FOLD_END = Integer.parseInt(value.substring(index + 1));
                } catch (Exception e) {
                    System.err.println("Value " + value +
                                       " for folds not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-numTotal")) {
                try {
                    numTotal = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for numTotal not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-numTrain")) {
                try {
                    numTrain = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for numTrain not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-lang")) {
                if (value.equals("en")) {
                    lang = "en";
                } else if (value.equals("es")) {
                    lang = "es";
                } else if (value.equals("ja")) {
                    lang = "ja";
                } else if (value.equals("tr")) {
                    lang = "tr";
                } else {
                    System.err.println("Value " + value +
                                       " for lang not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-gram")) {
                if (value.equals("unigram")) {
                    gram = "unigram";
                } else if (value.equals("bigram")) {
                    gram = "bigram";
                } else if (value.equals("mixbigram")) {
                    gram = "mixbigram";
                } else if (value.equals("newmixbigram")) {
                    gram = "newmixbigram";
                } else if (value.equals("mixgram")) {
                    gram = "mixgram";
                } else {
                    System.err.println("Value " + value +
                                       " for gram not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-corpus")) {
                if (value.equals("robocup")) {
                    corpus = "robocup-clang";
                } else if (value.equals("geofunql")) {
                    corpus = "geo-funql";
                } else if (value.equals("atis")) {
                    corpus = "atis-funql";
                } else {
                    System.err.println("Value " + value +
                                       " for corpus not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-iterations")) {
                try {
                    iteration = Integer.parseInt(value) - 1;
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for iteration not recognizable!");
                    help();
                    System.exit(1);
                }
            } else {
                System.err.println("Option " + option + " not recoginizable!");
                help();
                System.exit(1);
            }
        }
    }

    public static void main(String args[]) throws IOException,
            ClassNotFoundException {

        readArgs(args);
        dispArgs();

        for (int RUN = 0; RUN < 1; RUN++) {

            double totalTotal = 0;
            double totalMaxCorr = 0;
            double totalBestCorr = 0;
            double totalPred = 0;

            for (int FOLD = FOLD_START; FOLD <= FOLD_END; FOLD++) {
            	
                String prefix = lang + "/";
                prefix += Utility.norm(numTotal);
                prefix += '-';
                prefix += Utility.norm(numTrain);
                prefix += '/';

                File file_read;
                FileInputStream infile;
                ObjectInputStream ois;

                file_read = new File("result/" + corpus + "/top/hybridtree-" + gram +
                                     "/" + prefix + Utility.norm(iteration) +
                                     "." + RUN + "." + FOLD +
                                     ".percept." + Utility.norm(K));

                infile = new FileInputStream(file_read);

                ois = new ObjectInputStream(infile);

                Perceptron p = (Perceptron) ois.readObject();
                ois.close();
                infile.close();

                /*
                 File file = new File("result/" + corpus + "/parse/hybridtree-" + gram + "/" + Utility.norm(iteration) + "." + RUN + "." + FOLD +
                                     ".percept.result");
                 FileOutputStream outfile = new FileOutputStream(file);

                             PrintWriter pw = new PrintWriter(outfile);
                 */

                File file_test = new File("result/" + corpus + "/top/hybridtree-" +
                                          gram + "/" + prefix +
                                          Utility.norm(iteration) + "." + RUN +
                                          "." + FOLD +
                                          ".candidates." + Utility.norm(K) +
                                          ".[test]");
                FileInputStream infile_test = new FileInputStream(file_test);
                ObjectInputStream ois_test = new ObjectInputStream(infile_test);

                File file_result = new File("result/" + corpus + "/top/hybridtree-" +
                                            gram + "/" + prefix +
                                            Utility.norm(iteration) + "." + RUN +
                                            "." + FOLD +
                                            "." + Utility.norm(K) +
                                            ".predictions");
                FileOutputStream outfile_result = new FileOutputStream(
                        file_result);
                PrintWriter pw_result = new PrintWriter(outfile_result);

                Model m = null;
                if (corpus.equals("geo-funql")) {
                    m = ModelReader.readGeoqueryFunql(RUN, FOLD, lang, numTotal,
                            numTrain);
                }

                ArrayList<CandidateSet> candidateSets = new ArrayList<CandidateSet>();
                try {
                    while (true) {
                        candidateSets.add((CandidateSet) ois_test.readObject());
                    }
                } catch (Exception e) {
                }

                double total = 0;
                double maxCorr = 0;
                double bestCorr = 0;
                double pred = 0;

                for (int i = 0; i < candidateSets.size(); i++) {
                    total++;
                    boolean flag = false;
                    CandidateSet candidateSet = candidateSets.get(i);
                    if (candidateSet.size() == 0) {
                        continue;
                    }
                    for (int rank = 0; rank < candidateSet.size(); rank++) {
                        Candidate cand = candidateSet.getCandidateAt(rank);
                        if (cand.isCorrect()) {
                            flag = true;
                        }
                        cand.getFeatures(p);
                    }

                    if (flag) {
                        bestCorr++;
                    }

                    ArrayList<Candidate>
                            maxCandidates = p.perceptPick(candidateSet);
                    if (maxCandidates.size() == 0) {
                        continue;
                    }

                    pred++;

//                    boolean great = false;
                    if (maxCandidates.get(0).isCorrect()) {
//                        great = true;
                        maxCorr++;
                    }

                    Candidate maxCandidate = maxCandidates.get(0);
                    Production rootNode = recoverProd(maxCandidate.getRootNode().
                            getElem(), null);

//                System.err.println(rootNode.toTree());
//                System.err.println(mrl);
                    if (corpus.equals("geo-funql")) {
                        String mrl = Utility.toFunql(rootNode.getAllChildren());
                        ArrayList<Example> exs = m.getTestExamples();
                        Example ex = exs.get(i);
                        pw_result.println("Example " + i);
                        pw_result.println("pred:" + mrl);
                        pw_result.println("corr:" + ex.getMRL());
                        pw_result.println();
                        pw_result.flush();
                    }
                }

                pw_result.close();
                outfile_result.close();
                //file_result.setReadOnly();

                /*
                             System.err.println("Total:\t" + (int) total);
                             System.err.println("Pred:\t" + (int) pred);
                             System.err.println("Corr:\t" + (int) maxCorr);
                 System.err.println("Best Corr:\t" + (int) bestCorr);
                             System.err.println();
                             System.err.println();
                 */

                double precision = maxCorr / pred;
                double recall = maxCorr / total;
                double f = 2 / (1 / precision + 1 / recall);

                System.err.println("Precision:\t" + precision);
                System.err.println("Recall:\t" + recall);
                System.err.println("F:\t" + f);

                totalTotal += total;
                totalPred += pred;
                totalMaxCorr += maxCorr;
                totalBestCorr += bestCorr;

                System.err.println("\t" + precision + "\t" + recall + "\t" + f);
            }

            double precision = totalMaxCorr / totalPred;
            double recall = totalMaxCorr / totalTotal;
            double F = 2 / (1 / precision + 1 / recall);

            /**/
            System.err.println("-Summary-");
            System.err.println("Total:\t" + (int) totalTotal);
            System.err.println("Pred:\t" + (int) totalPred);
            System.err.println("Corr:\t" + (int) totalMaxCorr);
            System.err.println("Best Corr:\t" + (int) totalBestCorr);
            System.err.println("Precision:" + precision);
            System.err.println("Recall:" + recall);
            System.err.println("F:" + F);
            System.err.println();
            System.err.println();
            /**/

            System.err.println(precision + "\t" + recall + "\t" + F);

        }

    }


    private static Production recoverProd(ResElem res, Production parent) {
        GProduction gp = res.getGP();
        Production prod = new Production(gp);

        if (parent != null) {
            ArrayList<Production> childrenOfPar = parent.getChildren();
            childrenOfPar.add(prod);
            prod.setParent(parent);
        }

        ResElem[] children = res.getChildren();

        for (int i = 0; i < children.length; i++) {
            ResElem child = children[i];
            recoverProd(child, prod);
        }

        return prod;
    }
}
