
/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.eval;


/**
 * @author LU, Wei
 * @version 1.0
 */

import hybridtree.core.Param;
import hybridtree.core.parse.*;
import hybridtree.core.viterbi.*;
import hybridtree.data.Example;
import hybridtree.data.Model;
import hybridtree.data.Production;
import hybridtree.data.TestExample;
import hybridtree.data.TrainExample;
import hybridtree.util.Utility;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.PrintWriter;

import java.util.ArrayList;



import java.io.File;

public class SelfTest {

    private static String testMode = "parse";

    private static String gram = "bigram";

    private static String corpus = "x";

    private static int iteration = 99;

    private static int FOLD_START = 0;

    private static int FOLD_END = 9;

    private static int RUN = 0;

    private static void help() {
        System.err.println("Useage: java Test [options]");
        System.err.println("\toptions:");
        System.err.println("\t\t corpus=robocup/geofunql/geolambda (must be specified)");
        System.err.println("\t\t testMode=1,2,.. (default:parse)");
        System.err.println("\t\t run=0-9 (default:0)");
        System.err.println("\t\t folds=e.g.:3-8 (default:0-9)");
        System.err.println("\t\t iteration=9,19,.. (default:99)");
        System.err.println("\t\t gram=unigram/bigram (default:bigram)");
    }

    private static void dispArgs() {
        System.out.println("run:\t" + RUN);
        System.out.println("folds:" + FOLD_START + "~" + FOLD_END);
        System.out.println("corpus:\t" + corpus);
        System.out.println("testMode:\t" + testMode);
        System.out.println("gram:\t" + gram);
        System.out.println("parameter at iteration " + Utility.norm(iteration) + " used.");
    }

    private static void readArgs(String args[]) {
        if (args.length % 2 != 0) {
            System.err.println("Invalid Argument: " + args[args.length - 1]);
            help();
            System.exit(1);
        }

        for (int i = 0; i < args.length; i += 2) {
            String option = args[i];
            String value = args[i + 1];
            if (option.equals("-testMode")) {
                if (value.equals("align")) {
                    testMode = "align";
                } else if (value.equals("parse")) {
                    testMode = "parse";
                } else {
                    System.err.println("Value " + value + " for testMode not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-run")) {
                try {
                    RUN = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value + " for run not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-folds")) {
                try {
                    int index = value.indexOf("-");
                    FOLD_START = Integer.parseInt(value.substring(0, index));
                    FOLD_END = Integer.parseInt(value.substring(index + 1));
                } catch (Exception e) {
                    System.err.println("Value " + value + " for folds not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-gram")) {
                if (value.equals("unigram")) {
                    gram = "unigram";
                } else if (value.equals("bigram")) {
                    gram = "bigram";
                } else if (value.equals("mixbigram")) {
                    gram = "mixbigram";
                } else if (value.equals("newmixbigram")) {
                    gram = "newmixbigram";
                } else {
                    System.err.println("Value " + value + " for gram not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-corpus")) {
                if (value.equals("robocup")) {
                    corpus = "robocup-clang";
                } else if (value.equals("geofunql")) {
                    corpus = "geo-funql";
                }else {
                    System.err.println("Value " + value + " for corpus not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-iteration")) {
                try {
                    iteration = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value + " for iteration not recognizable!");
                    help();
                    System.exit(1);
                }
            } else {
                System.err.println("Option " + option + " not recoginizable!");
                help();
                System.exit(1);
            }
        }
    }

    public static void main(String args[]) throws IOException, ClassNotFoundException {

        readArgs(args);


        dispArgs();

        for (int FOLD = FOLD_START; FOLD <= FOLD_END; FOLD++) {

            long bTime = System.currentTimeMillis();

            System.err.println("Test..RUN " + RUN + ", FOLD " + FOLD);

            FileInputStream infile;
            String filename = "param/" + corpus + "/ham-" + gram + "/" + RUN + "." + FOLD + ".param." + Utility.norm(iteration);

            System.err.println("Reading the Param from: " + filename);

            infile = new FileInputStream(filename);

            ObjectInputStream ois = new ObjectInputStream(infile);
            Param parameter = (Param) ois.readObject();

            infile.close();

            Model model = parameter.getModel();

            ArrayList<Example> examples = model.getTrainExamples();

            System.err.println("There are totally " + examples.size() + " examples.");

            if (testMode.equals("parse")) {
                File file = new File("result/" + corpus + "/self/ham-" + gram + "/" + Utility.norm(iteration) + "." + RUN + "." + FOLD + ".result");
                FileOutputStream outfile = new FileOutputStream(file);

                PrintWriter p = new PrintWriter(outfile);

                int total = examples.size();
                int maxCorr = 0;
                int bestCorr = 0;

                int pred = 0;

                for (int i = 0; i < examples.size(); i++) {

                    TestExample example = ((TrainExample) examples.get(i)).toTestExample();
                    if (example.getExampleId() < 0) {
                        continue;
                    }

                    if (example.getExampleId() != 19) {
                        continue;
                    }

                    System.err.println("Example: " + example.getExampleId());

                    ParseTable table;

                    if (gram.equals("unigram")) {
                        table = ParseTable.createUGTable(example, parameter);
                    } else if (gram.equals("bigram")) {
                        table = ParseTable.createBGTable(example, parameter);
                    } else if (gram.equals("mixbigram")) {
                        if (!gram.equals("mixbigram")) {
                            System.err.println("Invalid gram:" + gram);
                            System.exit(1);
                        }
                        table = ParseTable.createMBGTable(example, parameter);
                    } else /*if(gram.equals("newmixbigram"))*/ {
                        if (!gram.equals("newmixbigram")) {
                            System.err.println("Invalid gram:" + gram);
                            System.exit(1);
                        }
                        table = ParseTable.createMBGTable(example, parameter);
                    }

                    table.setCorpus(corpus);

                    int K = 10;

                    table.parseMax(K);

                    ArrayList<Production> roots = table.restoreMaxMRL_Kbest();

                    if (!table.ignore()) {
                        pred++;
                        boolean corrFlag = false;

                        for (int rank = 0; rank < roots.size(); rank++) {
                            ArrayList<Production> results = roots.get(rank).getAllChildren();

                            {
                                boolean answer = example.verifyCorrect(results);

                                System.out.println("NL:\t" + example.getNL());
                                System.out.println("Predicted MRL:\t" + Utility.toFunql(results));
                                System.out.println("Correct MRL:\t" + example.getMRL());
                                System.out.println("Answer:\t" + answer);
                                System.out.println();

                                p.println("NL:\t" + example.getNL());
                                p.println("Predicted MRL rank " + rank + ":\t" + Utility.toFunql(results));
                                p.println("Correct MRL:\t" + example.getMRL());
                                p.println("Answer:\t" + answer);
                                p.println();
                                p.flush();

                                if (answer) {
                                    if (rank == 0) {
                                        maxCorr++;
                                    }
                                    corrFlag = true;
                                }
                            }
                        }

                        if (corrFlag) {
                            bestCorr++;
                        }
                    } else {

                        MaxTable pt = MaxTable.createBGTable((TrainExample) examples.get(i), parameter);
                        pt.alignMax(1);

                        if (!pt.ignored()) {
                            System.err.println(pt.getOptimalHybridTrees().get(0).toString());
                            System.exit(1);
                        }

                        System.out.println("NL:\t" + example.getNL());
                        System.out.println("Predicted MRL:\tFAILED...");
                        System.out.println("Correct MRL:\t" + example.getMRL());
                        System.out.println("Answer:\tfalse");
                        System.out.println();

                        p.println("NL:\t" + example.getNL());
                        p.println("Predicted MRL:\tFAILED...");
                        p.println("Correct MRL:\t" + example.getMRL());
                        p.println("Answer:\tfalse");
                        p.println();
                        p.flush();

                    }
                }

                System.err.println("-Summary-");
                System.err.println("Total        :\t" + total);
                System.err.println("Parsed       :\t" + pred);
                System.err.println("Max   Correct:\t" + maxCorr);
                System.err.println("Best Can Do:\t" + bestCorr);

                p.println("-Summary-");
                p.println("Total        :\t" + total);
                p.println("Parsed       :\t" + pred);
                p.println("Max   Correct:\t" + maxCorr);
                p.println("Best Can Do:\t" + bestCorr);
                p.flush();

                outfile.close();

                file.setReadOnly();
            }

            long eTime = System.currentTimeMillis();
            System.err.println("Total Testing Time:\t" + (eTime - bTime) / 1000.0);
        }
    }

}
