/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


package hybridtree.eval;


/**
 *
 * @author LU, Wei
 * @version 1.0
 */

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.ArrayList;

import hybridtree.core.Candidate;
import hybridtree.core.CandidateSet;
import hybridtree.core.Param;
import hybridtree.core.parse.ParseTable;
import hybridtree.data.Example;
import hybridtree.data.Model;
import hybridtree.data.TestExample;
import hybridtree.util.Utility;

public class Test {

    private static String testMode = "parse";

    private static String gram = "bigram";

    private static String corpus = "x";

    private static int iteration = 99;

    private static int FOLD_START = 0;

    private static int FOLD_END = 9;

    private static int RUN = 0;

    private static int K = 1;

    private static String lang = "en";

    private static int numTotal = -1;

    private static int numTrain = -1;

    private static void help() {
        System.err.println("Useage: java Test [options]");
        System.err.println("\toptions:");
        System.err.println(
                "\t\t corpus=robocup/geofunql/geolambda (must be specified)");
        System.err.println("\t\t testMode=1,2,.. (default:parse)");
        System.err.println("\t\t run=0-9 (default:0)");
        System.err.println("\t\t folds=e.g.:3-8 (default:0-9)");
        System.err.println("\t\t iterations=10,20,.. (default:99)");
        System.err.println("\t\t gram=unigram/bigram/mixgram (default:bigram)");
    }

    private static void dispArgs() {
        System.out.println("run:\t" + RUN);
        System.out.println("folds:" + FOLD_START + "~" + FOLD_END);
        System.out.println("corpus:\t" + corpus);
        System.out.println("testMode:\t" + testMode);
        System.out.println("gram:\t" + gram);
        System.out.println("parameter at iteration " + Utility.norm(iteration) +
                           " used.");
    }

    private static void readArgs(String args[]) {
        if (args.length % 2 != 0) {
            System.err.println("Invalid Argument: " + args[args.length - 1]);
            help();
            System.exit(1);
        }

        for (int i = 0; i < args.length; i += 2) {
            String option = args[i];
            String value = args[i + 1];
            if (option.equals("-testMode")) {
                if (value.equals("align")) {
                    testMode = "align";
                } else if (value.equals("parse")) {
                    testMode = "parse";
                } else {
                    System.err.println("Value " + value +
                                       " for testMode not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-run")) {
                try {
                    RUN = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for run not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-folds")) {
                try {
                    int index = value.indexOf("-");
                    FOLD_START = Integer.parseInt(value.substring(0, index));
                    FOLD_END = Integer.parseInt(value.substring(index + 1));
                } catch (Exception e) {
                    System.err.println("Value " + value +
                                       " for folds not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-gram")) {
                if (value.equals("unigram")) {
                    gram = "unigram";
                } else if (value.equals("bigram")) {
                    gram = "bigram";
                } else if (value.equals("mixgram")) {
                    gram = "mixgram";
                } else if (value.equals("mixbigram")) {
                    gram = "mixbigram";
                } else if (value.equals("newmixbigram")) {
                    gram = "newmixbigram";
                } else {
                    System.err.println("Value " + value +
                                       " for gram not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-numTotal")) {
                try {
                    numTotal = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for numTotal not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-numTrain")) {
                try {
                    numTrain = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for numTrain not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-lang")) {
                if (value.equals("en")) {
                    lang = "en";
                } else if (value.equals("es")) {
                    lang = "es";
                } else if (value.equals("ja")) {
                    lang = "ja";
                } else if (value.equals("tr")) {
                    lang = "tr";
                } else {
                    System.err.println("Value " + value +
                                       " for lang not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-corpus")) {
                if (value.equals("robocup")) {
                    corpus = "robocup-clang";
                } else if (value.equals("geofunql")) {
                    corpus = "geo-funql";
                }else if (value.equals("atis")) {
                    corpus = "atis-funql";
                } else {
                    System.err.println("Value " + value +
                                       " for corpus not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-iterations")) {
                try {
                    iteration = Integer.parseInt(value)-1;
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for iteration not recognizable!");
                    help();
                    System.exit(1);
                }
            } else {
                System.err.println("Option " + option + " not recoginizable!");
                help();
                System.exit(1);
            }
        }
    }

    public static void main(String args[]) throws IOException,
            ClassNotFoundException {

        readArgs(args);


        dispArgs();

        for (int FOLD = FOLD_START; FOLD <= FOLD_END; FOLD++) {

            String prefix = lang + "/";
            prefix += Utility.norm(numTotal);
            prefix += '-';
            prefix += Utility.norm(numTrain);
            prefix += '/';

            long bTime = System.currentTimeMillis();

            System.err.println("Test..RUN " + RUN + ", FOLD " + FOLD);

            FileInputStream infile;
            String filename = "param/" + corpus + "/hybridtree-" + gram + "/" + prefix +
                              RUN +
                              "." + FOLD + ".param." + Utility.norm(iteration);
            
            if(gram.equals("mixgram")){
                filename = "param/" + corpus + "/hybridtree-bigram/" + prefix +
                RUN +
                "." + FOLD + ".param." + Utility.norm(iteration);
            }

            System.err.println("Reading the Param from: " + filename);

            infile = new FileInputStream(filename);

            ObjectInputStream ois = new ObjectInputStream(infile);
            Param parameter = (Param) ois.readObject();
            
            Param parameter_unigram = null;
            
            if(gram.equals("mixgram")){
                FileInputStream infile_unigram;
                String filename_unigram = "param/" + corpus + "/hybridtree-unigram/" + prefix +
                                  RUN +
                                  "." + FOLD + ".param." + Utility.norm(iteration);

                System.err.println("Reading the Unigram Param from: " + filename_unigram);

                infile_unigram = new FileInputStream(filename_unigram);

                ObjectInputStream ois_unigram = new ObjectInputStream(infile_unigram);
                parameter_unigram = (Param) ois_unigram.readObject();
            }

            infile.close();

            Model model = parameter.getModel();

            ArrayList<Example> examples = model.getTestExamples();

            System.err.println("There are totally " + examples.size() +
                               " examples.");

            File tmpFile = new File("result/" + corpus + "/parse/hybridtree-" + gram +
                                    "/" + prefix);
            tmpFile.mkdirs();
            
            int total = examples.size();
            int maxCorr = 0;
            int bestCorr = 0;

            int pred = 0;

            for (int i = 0; i < examples.size(); i++) {

                TestExample example = (TestExample) examples.get(i);

                System.err.println("Example: " + example.getExampleId());

                ParseTable table;

                if (gram.equals("unigram")) {
                    table = ParseTable.createUGTable(example, parameter);
                } else if (gram.equals("bigram")) {
                    table = ParseTable.createBGTable(example, parameter);
                } else if (gram.equals("mixgram")) {
                    table = ParseTable.createMGTable(example, parameter_unigram,parameter);
                } else if (gram.equals("mixbigram")) {
                    table = ParseTable.createMBGTable(example, parameter);
                } else /*if(gram.equals("newmixbigram"))*/ {
                    if (!gram.equals("newmixbigram")) {
                        System.err.println("Invalid gram:" + gram);
                        System.exit(1);
                    }
                    table = ParseTable.createMBGTable(example, parameter);
                } 

                table.setCorpus(corpus);

                table.parseMax(K);

                if (!table.ignore()) {
                    pred++;

                    CandidateSet candidates = table.getTopCandidates();

                    boolean sumCorr = false;

                    for (int rank = 0; rank < candidates.size(); rank++) {
                        Candidate cand = candidates.getCandidateAt(rank);

                        System.out.println(rank + "=" + cand.getProb());
                        System.out.println(cand.isCorrect());
                        System.out.println(cand.getRootNode().toString());

                        if (cand.isCorrect()) {
                            if (rank == 0) {
                                maxCorr++;
                            }
                            sumCorr = true;
                        }
                    }

                    if (sumCorr) {
                        bestCorr++;
                    }

                } else {
                }
            }
            
            System.err.println("-Summary-");
            System.err.println("Total        :\t" + total);
            System.err.println("Parsed       :\t" + pred);
            System.err.println("Max   Correct:\t" + maxCorr);
            System.err.println("Best Can Do:\t" + bestCorr);

            long eTime = System.currentTimeMillis();
            System.err.println("Total Testing Time:\t" +
                               (eTime - bTime) / 1000.0);
        }
    }

}
