/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package hybridtree.eval;

import java.io.IOException;
import java.util.ArrayList;

import java.io.*;

import hybridtree.core.*;
import hybridtree.core.learn.InitLearner;
import hybridtree.core.learn.TrainTable;
import hybridtree.data.*;
import hybridtree.data.reader.*;
import hybridtree.util.Utility;

public class Train {

    private static boolean verboseMode = false;

    private static int _MAX_ITERATIONS = 100;

    private static int _ITERATION_START = 0;

    private static String gram = "bigram";

    private static int RUN = 0;

    private static int FOLD_START = 0;

    private static int FOLD_END = 9;

    private static String corpus = "x";

    private static String lang = "en";

    private static int numTotal = -1;

    private static int numTrain = -1;

    private static void help() {
        System.err.println("Useage: java Train [options]");
        System.err.println("\toptions:");
        System.err.println(
                "\t\t corpus=robocup/geofunql/geolambda (must be specified)");
        System.err.println("\t\t run=0-9 (default:0)");
        System.err.println("\t\t folds=e.g.:3-8 (default:0-9)");
        System.err.println("\t\t iterations=10,20,.. (default:100)");
        System.err.println("\t\t gram=unigram/bigram (default:bigram)");
        System.err.println("\t\t lang=en/es/ja/tr (default:en)");
        System.err.println("\t\t numTotal=880/300/250");
        System.err.println("\t\t numTrain=792/270/...");
    }

    private static void dispArgs() {
        System.out.println("run:\t" + RUN);
        System.out.println("folds:" + FOLD_START + "~" + FOLD_END);
        System.out.println("corpus:\t" + corpus);
        System.out.println("gram:\t" + gram);
        System.out.println("iterations:\t" + _MAX_ITERATIONS); ;
        System.out.println("lang:\t" + lang);
        System.out.println("numTotal:\t" + numTotal);
        System.out.println("numTrain:\t" + numTrain);
    }

    private static void readArgs(String args[]) {
        if (args.length % 2 != 0) {
            System.err.println("Invalid Argument: " + args[args.length - 1]);
            help();
            System.exit(1);
        }

        for (int i = 0; i < args.length; i += 2) {
            String option = args[i];
            String value = args[i + 1];

            if (option.equals("-iterations")) {
                try {
                    _MAX_ITERATIONS = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for iterations not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-corpus")) {
                if (value.equals("robocup")) {
                    corpus = "robocup-clang";
                } else if (value.equals("geofunql")) {
                    corpus = "geo-funql";
                } else if (value.equals("atis")) {
                    corpus = "atis-funql";
                } else {
                    System.err.println("Value " + value +
                                       " for corpus not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-verboseMode")) {
                if (value.equals("true")) {
                    verboseMode = true;
                } else if (value.equals("false")) {
                    verboseMode = false;
                } else {
                    System.err.println("Value " + value +
                                       " for verboseMode not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-run")) {
                try {
                    RUN = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for run not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-numTotal")) {
                try {
                    numTotal = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for numTotal not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-numTrain")) {
                try {
                    numTrain = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for numTrain not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-folds")) {
                try {
                    int index = value.indexOf("-");
                    FOLD_START = Integer.parseInt(value.substring(0, index));
                    FOLD_END = Integer.parseInt(value.substring(index + 1));
                } catch (Exception e) {
                    System.err.println("Value " + value +
                                       " for folds not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-gram")) {
                if (value.equals("unigram")) {
                    gram = "unigram";
                } else if (value.equals("bigram")) {
                    gram = "bigram";
                } else if (value.equals("mixbigram")) {
                    gram = "mixbigram";
                } else if (value.equals("newmixbigram")) {
                    gram = "newmixbigram";
                } else {
                    System.err.println("Value " + value +
                                       " for gram not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-lang")) {
                if (value.equals("en")) {
                    lang = "en";
                } else if (value.equals("es")) {
                    lang = "es";
                } else if (value.equals("ja")) {
                    lang = "ja";
                } else if (value.equals("tr")) {
                    lang = "tr";
                } else {
                    System.err.println("Value " + value +
                                       " for lang not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-beginIteration")) {
                try {
                    _ITERATION_START = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for beginIteration not recognizable!");
                    help();
                    System.exit(1);
                }
            } else {
                System.err.println("Option " + option + " not recoginizable!");
                help();
                System.exit(1);
            }
        }
    }

    public static void main(String args[]) throws Exception {

        readArgs(args);
        dispArgs();

        System.err.println("Started.");

        for (int FOLD = FOLD_START; FOLD <= FOLD_END; FOLD++) {

            System.err.println("RUN " + RUN + ", FOLD " + FOLD);

            long bTime = System.currentTimeMillis();

            Model model;

            ArrayList<Example> examples;

            if (corpus.equals("robocup-clang")) {
                model = ModelReader.readRobocupClang(RUN, FOLD, numTotal,
                        numTrain);
                examples = model.getTrainExamples();
            } else if (corpus.equals("geo-funql")) {
                model = ModelReader.readGeoqueryFunql(RUN, FOLD, lang, numTotal,
                        numTrain);
                examples = model.getTrainExamples();
            } else if (corpus.equals("atis-funql")) {
                model = ModelReader.readAtis(RUN, FOLD);
                examples = model.getTrainExamples();

                /*
                                 for (int k = 0; k < examples.size(); k++) {
                    Example ex = examples.get(k);

                    ArrayList<Production>
                            prods = ((TrainExample) ex).getProductions();
                    for (int t = 0; t < prods.size(); t++) {
                        System.out.println(prods.get(t));
                    }
                    System.exit(1);
                                 }
                 */

            } else {
                model = null;
                examples = null;
                System.err.println("Error Recognizing Corpus:" + corpus);
                System.exit(1);
            }

            //here goes the parameter initialization
            Param parameter = new Param(model);

            String prefix = lang + "/";
            prefix += Utility.norm(numTotal);
            prefix += '-';
            prefix += Utility.norm(numTrain);
            prefix += '/';

            if (_ITERATION_START == 0) {
                if (gram.equals("unigram")) {
                    System.err.println("Training IBM on " + examples.size() +
                                       " Examples only!");

                    //perform IBM model 1
                    InitLearner initer = new InitLearner(parameter, model);
                    initer.learn();
                    parameter = initer.getParam();

                    File tmp = new File("param/" + corpus + "/ibm1/" + prefix);
                    tmp.mkdirs();

                    File ibm1file = new File("param/" + corpus + "/ibm1/" +
                                             prefix + RUN + "." + FOLD +
                                             ".param");
                    ibm1file.createNewFile();
                    FileOutputStream outfile1 = new FileOutputStream(ibm1file);

                    ObjectOutputStream oos = new ObjectOutputStream(outfile1);

                    oos.writeObject(parameter);
                    oos.flush();
                    outfile1.close();
                } else if (gram.equals("bigram") || gram.equals("mixbigram") ||
                           gram.equals("newmixbigram")) {
                    System.err.println(
                            "Reading unigram output as initial parameter");
                    FileInputStream infile = new FileInputStream("param/" +
                            corpus + "/hybridtree-unigram/" + prefix + RUN + "." +
                            FOLD + ".param." +
                            Utility.norm(_MAX_ITERATIONS - 1));
                    ObjectInputStream ois = new ObjectInputStream(infile);

                    parameter = (Param) ois.readObject();
                    ois.close();
                } else {
                    System.err.println("SORRY, NOT SUPPORTED!");
                    System.exit(1);
                }
            } else {
                FileInputStream infile = new FileInputStream("param/" + corpus +
                        "/hybridtree-" + gram + "/" + prefix + RUN + "." + FOLD +
                        ".param." +
                        Utility.norm(_ITERATION_START - 1));
                ObjectInputStream ois = new ObjectInputStream(infile);

                parameter = (Param) ois.readObject();
                ois.close();
            }

            System.err.println("IBM Model 1 finished.");

            System.err.println("Hybrid Tree Learning started...");

            double oldLikelihood = Double.NEGATIVE_INFINITY;

            for (int iteration = _ITERATION_START; iteration < _MAX_ITERATIONS;
                                 iteration++) {

                System.err.println("[" + RUN + "," + FOLD + "]Iteration " +
                                   iteration + "/" + _MAX_ITERATIONS);

                long beTime = System.currentTimeMillis();

                double likelihood = 0;

                for (int i = 0; i < examples.size(); i++) {

                    TrainExample example = (TrainExample) examples.get(i);

                    TrainTable table;
                    if (gram.equals("unigram")) {
                        table = TrainTable.createUGTable(example, parameter);
                    } else if (gram.equals("bigram")) {
                        table = TrainTable.createBGTable(example, parameter);
                    } else if (gram.equals("mixbigram")) {
                        table = TrainTable.createMBGTable(example, parameter);
                    } else /*if(gram.equals("newmixbigram"))*/ {
                        if (!gram.equals("newmixbigram")) {
                            System.err.println("Invalid gram:" + gram);
                            System.exit(1);
                        }
                        table = TrainTable.createNMBGTable(example, parameter);
                    }

                    table.computeBeta();
                    table.computeAlpha();

                    if (!table.ignore()) {
                        double subLoglikelihood = table.getLogLikelihood() *
                                                  example.getWeight();
                        if (verboseMode) {
                            System.err.print("\rInstance " + i + " with id " +
                                             example.getExampleId() +
                                             " Loglikelihood:\t" +
                                             subLoglikelihood);
                        }
                        likelihood += subLoglikelihood;
                    }
                }

                if (verboseMode) {
                    System.err.println("\rOK!                                                                          ");
                }

                long enTime = System.currentTimeMillis();

                System.out.println("Time For This Iteration:" +
                                   (enTime - beTime) / 1000.0);
                System.out.println("Total LogLikelihood = " + likelihood+"\tThe Improvement  = " +
                                   (likelihood - oldLikelihood));
                if (likelihood - oldLikelihood < 0) {
                    System.out.println(
                            "likelihood decreased!!!!!!!!!!!!!!!!!!!!!!!!");
                    parameter.reset_new();
                    FileOutputStream fout = new FileOutputStream("fail.tmp");
                    ObjectOutputStream fos = new ObjectOutputStream(fout);

                    fos.writeObject(parameter);
                    fos.flush();
                    fos.close();
                    fout.close();
//                    System.exit(1);
                }
                oldLikelihood = likelihood;

                System.err.println("==Normalize and Swap==");
                parameter.normalizeAndSwap();

                try {
                    FileOutputStream outfile_param;
                    if (iteration % 10 == 9) {
                        File tmp = new File("param/" + corpus + "/hybridtree-" + gram +
                                            "/" + prefix);
                        tmp.mkdirs();

                        File of = new File("param/" + corpus + "/hybridtree-" + gram +
                                           "/" + prefix + RUN + "." + FOLD +
                                           ".param." + Utility.norm(iteration));
                        of.createNewFile();
                        outfile_param = new FileOutputStream(of);

                        ObjectOutputStream oos_param = new ObjectOutputStream(
                                outfile_param);
                        oos_param.writeObject(parameter);
                        oos_param.flush();
                        outfile_param.close();

//                        System.out.println(parameter.getModel().getTestExamples().size());
                    }
                } catch (IOException e) {
                    System.err.println("IO error! " + e);
                    System.exit(1);
                }
            }

            long eTime = System.currentTimeMillis();
            System.err.println("Total running time: " +
                               (eTime - bTime) / 1000.0 + " sec.");
        }
    }
}
