/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */



package hybridtree.eval;


import hybridtree.core.Param;
import hybridtree.core.learn.*;
import hybridtree.core.parse.*;
import hybridtree.core.viterbi.MaxTable;
import hybridtree.core.viterbi.UGMaxTable;
import hybridtree.data.Example;
import hybridtree.data.GProduction;
import hybridtree.data.Model;
import hybridtree.data.Production;
import hybridtree.data.TestExample;
import hybridtree.data.TrainExample;
import hybridtree.util.Utility;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;

import java.util.ArrayList;



/**
 *
 * @author Wei LU
 * @version 1.0
 */
public class Verify {

    private static String testMode = "parse";

    private static String gram = "bigram";

    private static String corpus = "x";

    private static int iteration = 99;

    private static int FOLD_START = 0;

    private static int FOLD_END = 9;

    private static int RUN = 0;

    private static void help() {
        System.err.println("Useage: java Test [options]");
        System.err.println("\toptions:");
        System.err.println("\t\t corpus=robocup/geofunql/geolambda (must be specified)");
        System.err.println("\t\t testMode=1,2,.. (default:parse)");
        System.err.println("\t\t run=0-9 (default:0)");
        System.err.println("\t\t folds=e.g.:3-8 (default:0-9)");
        System.err.println("\t\t iteration=9,19,.. (default:99)");
        System.err.println("\t\t gram=unigram/bigram (default:bigram)");
    }

    private static void dispArgs() {
        System.out.println("run:\t" + RUN);
        System.out.println("folds:" + FOLD_START + "~" + FOLD_END);
        System.out.println("corpus:\t" + corpus);
        System.out.println("testMode:\t" + testMode);
        System.out.println("gram:\t" + gram);
        System.out.println("parameter at iteration " + Utility.norm(iteration) + " used.");
    }

    private static void readArgs(String args[]) {
        if (args.length % 2 != 0) {
            System.err.println("Invalid Argument: " + args[args.length - 1]);
            help();
            System.exit(1);
        }

        for (int i = 0; i < args.length; i += 2) {
            String option = args[i];
            String value = args[i + 1];
            if (option.equals("-testMode")) {
                if (value.equals("align")) {
                    testMode = "align";
                } else if (value.equals("parse")) {
                    testMode = "parse";
                } else {
                    System.err.println("Value " + value + " for testMode not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-run")) {
                try {
                    RUN = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value + " for run not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-folds")) {
                try {
                    int index = value.indexOf("-");
                    FOLD_START = Integer.parseInt(value.substring(0, index));
                    FOLD_END = Integer.parseInt(value.substring(index + 1));
                } catch (Exception e) {
                    System.err.println("Value " + value + " for folds not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-gram")) {
                if (value.equals("unigram")) {
                    gram = "unigram";
                } else if (value.equals("bigram")) {
                    gram = "bigram";
                } else if (value.equals("mixbigram")) {
                    gram = "mixbigram";
                } else if (value.equals("newmixbigram")) {
                    gram = "newmixbigram";
                } else {
                    System.err.println("Value " + value + " for gram not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-corpus")) {
                if (value.equals("robocup")) {
                    corpus = "robocup-clang";
                } else if (value.equals("geofunql")) {
                    corpus = "geo-funql";
                }else {
                    System.err.println("Value " + value + " for corpus not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-iteration")) {
                try {
                    iteration = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value + " for iteration not recognizable!");
                    help();
                    System.exit(1);
                }
            } else {
                System.err.println("Option " + option + " not recoginizable!");
                help();
                System.exit(1);
            }
        }
    }

    private static double computeMRLLM(Param parameter, Production prod) {
        double prob = 1.0;
        ArrayList<Production> children = prod.getChildren();
        for (int cIndex = 0; cIndex < children.size(); cIndex++) {
            Production child = children.get(cIndex);
            prob *= computeMRLLMProb(parameter, prod.getGProduction(), cIndex, child.getGProduction()) * computeMRLLM(parameter, child);
        }
        return prob;
    }

    private static double computeMRLLMProb(Param parameter, GProduction prod, int cIndex, GProduction child) {
        if (child.isConstant()) {
            return 1.0;
        }
        return parameter.getMRLMProb(prod, cIndex, child);
    }

    public static void main(String args[]) throws IOException, ClassNotFoundException  {

        readArgs(args);


        dispArgs();

        for (int FOLD = FOLD_START; FOLD <= FOLD_END; FOLD++) {

            long bTime = System.currentTimeMillis();

            System.err.println("Test..RUN " + RUN + ", FOLD " + FOLD);

            FileInputStream infile;
            String filename = "param/" + corpus + "/ham-" + gram + "/" + RUN + "." + FOLD + ".param." + Utility.norm(iteration);

            System.err.println("Reading the Param from: " + filename);

            infile = new FileInputStream(filename);

            ObjectInputStream ois = new ObjectInputStream(infile);
            Param parameter = (Param) ois.readObject();

            infile.close();

            Model model = parameter.getModel();

            ArrayList<Example> examples = model.getTrainExamples();

            System.err.println("There are totally " + examples.size() + " examples.");

            if (testMode.equals("parse")) {

                int total = examples.size();
                int maxCorr = 0;
                int bestCorr = 0;

                int pred = 0;

                for (int i = 0; i < examples.size(); i++) {

                    TestExample example = ((TrainExample) examples.get(i)).toTestExample();
                    if (example.getExampleId() != 24) {
                        continue;
                    }

                    System.err.println("Example: " + example.getExampleId());

                    ParseTable table;

                    if (gram.equals("unigram")) {
//                        table = ParseTable.createNUGTable(example, parameter);
                        table = ParseTable.createUGTable(example, parameter);
                    } else if (gram.equals("bigram")) {
                        table = ParseTable.createBGTable(example, parameter);
                    } else if (gram.equals("mixbigram")) {
                        if (!gram.equals("mixbigram")) {
                            System.err.println("Invalid gram:" + gram);
                            System.exit(1);
                        }
                        table = ParseTable.createMBGTable(example, parameter);
                    } else /*if(gram.equals("newmixbigram"))*/ {
                        if (!gram.equals("newmixbigram")) {
                            System.err.println("Invalid gram:" + gram);
                            System.exit(1);
                        }
                        table = ParseTable.createMBGTable(example, parameter);
                    }

                    table.setCorpus(corpus);

                    //table.parseMax();

                    int K = 10;

                    table.parseMax(K);

                    ArrayList<Production> roots = table.restoreMaxMRL_Kbest();

//                    Hashtable kbestTable = table.getKbestTable();
//                    System.err.println(kbestTable);

                    if (!table.ignore()) {
                        pred++;
                        boolean corrFlag = false;

                        for (int rank = 0; rank < roots.size(); rank++) {
                            ArrayList<Production> results = roots.get(rank).getAllChildren();
                            Utility.assignSeq(results);
                            System.err.println();

                            TrainExample ex = new TrainExample(9999, example.getNL(), example.getWords(), example.getMRL(), results);
                            TrainTable tt = TrainTable.createUGTable(ex, parameter);

                            tt.computeBeta();
                            System.err.println(tt.getGlobalBeta() * computeMRLLM(parameter, ex.getRootProduction()));

                            UGMaxTable uTable = MaxTable.createUGTable(ex, parameter);
                            uTable.alignMax(1);

                            /////////////////////////////////////////

                            TrainExample ex_corr = (TrainExample) examples.get(i);
                            tt = TrainTable.createUGTable(ex_corr, parameter);

                            tt.computeBeta();
                            System.err.println(tt.getGlobalBeta() * computeMRLLM(parameter, ex_corr.getRootProduction()));

                            uTable = MaxTable.createUGTable(ex_corr, parameter);
                            uTable.alignMax(1);

                            System.exit(1);

                            {
                                boolean answer = example.verifyCorrect(results);

                                System.out.println("NL:\t" + example.getNL());
                                System.out.println("Predicted MRL:\t" + Utility.toFunql(results));
                                System.out.println("Correct MRL:\t" + example.getMRL());
                                System.out.println("Answer:\t" + answer);
                                System.out.println();

                                if (answer) {
                                    if (rank == 0) {
                                        maxCorr++;
                                    }
                                    corrFlag = true;
                                }
                            }
                        }

                        if (corrFlag) {
                            bestCorr++;
                        }
                    } else {

                        System.out.println("NL:\t" + example.getNL());
                        System.out.println("Predicted MRL:\tFAILED...");
                        System.out.println("Correct MRL:\t" + example.getMRL());
                        System.out.println("Answer:\tfalse");
                        System.out.println();

                    }

                    if (true) {
                        continue;
                    }

                    /*
                                         if (!table.ignore()) {

                        ArrayList<Production> results = table.getGlobalMaxP();

                        if (corpus.equals("geolambda")) {
                            Expr exp = LambdaToProductions.toExpr(results.get(1));
                            LambdaCalculusProcessor.reGroup(exp);

                            String mrl = LambdaCalculusProcessor.preprocess(example.getMRL());

                            Expr result = ExpressionParser.parse(mrl);
                            result = result.stripParens();
                            LambdaCalculusProcessor.reGroup(result);

                            System.out.println("NL:\t" + example.getNL());
                            System.out.println("Predicted MRL:\t" + exp);
                            System.out.println("Correct MRL:\t" + result);
                            System.out.println("Answer:\t" + example.verifyLambdaCorrect(results));
                            System.out.println();
                            if (exp.alphaEquivalent(result)) {
                                maxCorr++;
                            }
                            pred++;
                        } else {
                            boolean answer = example.verifyCorrect(results);

                            System.out.println("NL:\t" + example.getNL());
                            System.out.println("Predicted MRL:\t" + Utility.toFunql(results));
                            System.out.println("Correct MRL:\t" + example.getMRL());
                            System.out.println("Answer:\t" + answer);
                            System.out.println();

                            if (answer) {
                                maxCorr++;
                            }
                            pred++;
                        }
                                         } else {

                        System.out.println("NL:\t" + example.getNL());
                        System.out.println("Predicted MRL:\tFAILED...");
                        System.out.println("Correct MRL:\t" + example.getMRL());
                        System.out.println("Answer:\tfalse");
                        System.out.println();
                                         }

                                         if (true) {
                        continue;
                                         }
                     */
                }

                System.err.println("-Summary-");
                System.err.println("Total        :\t" + total);
                System.err.println("Parsed       :\t" + pred);
                System.err.println("Max   Correct:\t" + maxCorr);
                System.err.println("Best Can Do:\t" + bestCorr);

            }

            long eTime = System.currentTimeMillis();
            System.err.println("Total Testing Time:\t" + (eTime - bTime) / 1000.0);
        }
    }

}
