/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


package hybridtree.eval;

import java.io.IOException;
import java.util.ArrayList;

import java.io.*;

import hybridtree.core.*;
import hybridtree.core.viterbi.*;
import hybridtree.data.*;
import hybridtree.util.Utility;



/**
 *
 * @author Wei LU
 * @version 1.0
 */
public class Viterbi {

    private static int RUN = 0;

    private static String corpus = "x";

    @SuppressWarnings("unused")
	private static boolean verboseMode = false;

    private static int iteration = 99;
    
    private static String gram = "x";

    private static int FOLD_START = 0;

    private static int FOLD_END = 9;

    private static String lang = "x";

    private static int numTotal = -1;

    private static int numTrain = -1;


    private static void help() {
        System.err.println("Useage: java Train [options]");
        System.err.println("\toptions:");
        System.err.println(
                "\t\t corpus=robocup/geofunql/geolambda (must be specified)");
        System.err.println("\t\t run=0-9 (default:0)");
        System.err.println("\t\t folds=e.g.:3-8 (default:0-9)");
        System.err.println("\t\t iteration=10,20,.. (default:100)");
        System.err.println("\t\t gram=unigram/bigram (default:bigram)");
    }

    private static void dispArgs() {
        System.out.println("run:\t" + RUN);
        System.out.println("folds:" + FOLD_START + "~" + FOLD_END);
        System.out.println("corpus:\t" + corpus);
        System.out.println("gram:\t" + gram);
    }

    private static void readArgs(String args[]) {
        if (args.length % 2 != 0) {
            System.err.println("Invalid Argument: " + args[args.length - 1]);
            help();
            System.exit(1);
        }

        for (int i = 0; i < args.length; i += 2) {
            String option = args[i];
            String value = args[i + 1];

            if (option.equals("-iteration")) {
                try {
                    iteration = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for iteration not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-corpus")) {
                if (value.equals("robocup")) {
                    corpus = "robocup-clang";
                } else if (value.equals("geofunql")) {
                    corpus = "geo-funql";
                } else {
                    System.err.println("Value " + value +
                                       " for corpus not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-verboseMode")) {
                if (value.equals("true")) {
                    verboseMode = true;
                } else if (value.equals("false")) {
                    verboseMode = false;
                } else {
                    System.err.println("Value " + value +
                                       " for verboseMode not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-run")) {
                try {
                    RUN = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for run not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-folds")) {
                try {
                    int index = value.indexOf("-");
                    FOLD_START = Integer.parseInt(value.substring(0, index));
                    FOLD_END = Integer.parseInt(value.substring(index + 1));
                } catch (Exception e) {
                    System.err.println("Value " + value +
                                       " for folds not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-numTotal")) {
                try {
                    numTotal = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for numTotal not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-numTrain")) {
                try {
                    numTrain = Integer.parseInt(value);
                } catch (NumberFormatException e) {
                    System.err.println("Value " + value +
                                       " for numTrain not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-lang")) {
                if (value.equals("en")) {
                    lang = "en";
                } else if (value.equals("es")) {
                    lang = "es";
                } else if (value.equals("ja")) {
                    lang = "ja";
                } else if (value.equals("tr")) {
                    lang = "tr";
                } else {
                    System.err.println("Value " + value +
                                       " for lang not recognizable!");
                    help();
                    System.exit(1);
                }
            } else if (option.equals("-gram")) {
                if (value.equals("unigram")) {
                    gram = "unigram";
                } else if (value.equals("bigram")) {
                    gram = "bigram";
                } else if (value.equals("mixbigram")) {
                    gram = "mixbigram";
                } else if (value.equals("newmixbigram")) {
                    gram = "newmixbigram";
                } else {
                    System.err.println("Value " + value +
                                       " for gram not recognizable!");
                    help();
                    System.exit(1);
                }
            } else {
                System.err.println("Option " + option + " not recoginizable!");
                help();
                System.exit(1);
            }
        }
    }


    public static void main(String args[]) throws IOException,
            ClassNotFoundException {

        readArgs(args);

        dispArgs();

        System.err.println("Started.");

        for (int FOLD = FOLD_START; FOLD <= FOLD_END; FOLD++) {

            String prefix = lang + "/";
            prefix += Utility.norm(numTotal);
            prefix += '-';
            prefix += Utility.norm(numTrain);
            prefix += '/';

            String filename = "param/" + corpus + "/ham-" + gram + "/" + prefix +
                              RUN + "." + FOLD + ".param." +
                              Utility.norm(iteration);

            FileInputStream infile = new FileInputStream(filename);
            ObjectInputStream ois = new ObjectInputStream(infile);

            File file = new File("result/" + corpus + "/align/ham-" + gram +
                                 "/" + prefix + RUN + "." + FOLD +
                                 ".result");
            FileOutputStream outfile = new FileOutputStream(file);

            Param parameter = (Param) ois.readObject();

            Model model = parameter.getModel();

            int K = 1;

            ArrayList<Example> train_examples = model.getTrainExamples();

            long bTime = System.currentTimeMillis();

            for (int i = 0; i < train_examples.size(); i++) {
                TrainExample train_example = (TrainExample) train_examples.get(
                        i);

                if (train_example.getExampleId() < 0) {
                    continue;
                }
                MaxTable table;

                if (gram.equals("unigram")) {
                    table = MaxTable.createUGTable(train_example, parameter);
                } else {
                    table = MaxTable.createBGTable(train_example, parameter);
                }

                table.alignMax(K);

                if (table.ignored()) {
                    System.err.println("FAILED:\t");
                    System.err.println(train_example.getNL());
                } else {
                    System.err.println("OK:\t");
                    System.out.println(train_example.getNL());

                    CandidateSet candidates = table.getTopCandidates();

                    for (int k = 0; k < candidates.size(); k++) {
                        Candidate candidate = candidates.getCandidateAt(k);
                        System.out.println("probability:" + candidate.getProb());
                        System.out.println(candidate.getRootNode().toString());
                        System.out.println();
                    }
                }
            }

            outfile.close();

            long eTime = System.currentTimeMillis();
            System.err.println((double) (eTime - bTime) / 1000.0 + " sec.");
        }
    }

}
