/**
 *  The Hybrid Tree Framework for Semantic Parsing and Language Generation
 *
 *  Copyright (C) 2009  LU, Wei (luwei@statnlp.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


package hybridtree.util;

import java.util.StringTokenizer;
import java.util.ArrayList;

import hybridtree.data.GProduction;
import hybridtree.data.Node;
import hybridtree.data.NonterminalNode;
import hybridtree.data.Production;


/**
 * @author LU, Wei
 * @version 1.0
 */

public class Utility {

    /**
     * Verify the correctness of Productions
     * @param prods ArrayList
     * @return boolean
     */
    public boolean verifyCorrect(ArrayList<Production> prods1, ArrayList<Production> prods2) {
        Production prod_correct = prods1.get(0);
        Production prod_predict = prods2.get(0);
        return verifyCorrectHelper(prod_correct, prod_predict);
    }

    /**
     * Verify the correctness of two Productions
     * @param prod_correct Production
     * @param prod_predict Production
     * @return boolean
     */
    private boolean verifyCorrectHelper(Production prod_correct, Production prod_predict) {
        if (!prod_correct.getGProduction().equals(prod_predict.getGProduction())) {
            return false;
        }

        GProduction gp = prod_correct.getGProduction();

        ArrayList<Production> children_correct = prod_correct.getChildren();
        ArrayList<Production> children_predict = prod_predict.getChildren();

        boolean result = true;

        if (gp.isCommutative()) {
            Production child_correct_0 = children_correct.get(0);
            Production child_correct_1 = children_correct.get(1);

            Production child_predict_0 = children_predict.get(0);
            Production child_predict_1 = children_predict.get(1);

            result &= this.verifyCorrectHelper(child_correct_0, child_predict_0) && this.verifyCorrectHelper(child_correct_1, child_predict_1) ||
                    this.verifyCorrectHelper(child_correct_0, child_predict_1) && this.verifyCorrectHelper(child_correct_1, child_predict_0);
        } else {
            for (int i = 0; i < children_correct.size(); i++) {
                Production child_correct = children_correct.get(i);
                Production child_predict = children_predict.get(i);
                result &= verifyCorrectHelper(child_correct, child_predict);
            }
        }
        return result;
    }

    public static String norm(int num) {
        if (num < 10) {
            return "0000" + num;
        } else if (num < 100) {
            return "000" + num;
        } else if (num < 1000) {
            return "00" + num;
        } else if (num < 10000) {
            return "0" + num;
        } else {
            return "" + num;
        }
    }

    /**
     * To check whether a string is an identifier
     * @param word the string form to check
     * @return whether the string form is an identifier
     */
    public static boolean isIdent(String word) {
        return Character.isUpperCase(word.charAt(0));
    }


    /**
     * To check whether a string is a U-number
     * @param word the string form to check
     * @return whether the string form is a U-number
     */
    public static boolean isUnum(String word) {
        try {
            int unum = Integer.parseInt(word);
            if (unum >= 0) {
                return true;
            } else {
                return false;
            }
        } catch (NumberFormatException e) {
            return false;
        }
    }

    /**
     * To check whether a string is a number
     * @param word the string form to check
     * @return whether the string form is a number
     */
    public static boolean isNum(String word) {
        try {
            Double.parseDouble(word);
            return true;
        } catch (NumberFormatException e) {
            return false;
        }
    }

    /**
     * Conversion of a string form to a list of possible double numbers
     * @param word input string representation
     * @return a collection of double numbers, in string form
     */
    public static ArrayList<String> toNum(String word) {
        if (!isNum(word)) {
            return null;
        }

        ArrayList<String> results = new ArrayList<String>();

        StringTokenizer st = new StringTokenizer(word, ".");
        if (st.countTokens() == 2) {
            String num_i = st.nextToken();
            String fac_i = st.nextToken();

            if (fac_i.length() > 2) {
                results.add(word);
                return results;
            } else {
                if (fac_i.length() == 2) {
                    results.add(word);
                    if (fac_i.charAt(1) != '0') {
                        return results;
                    } else {
                        results.add(num_i + "." + fac_i.charAt(0));
                        if (fac_i.charAt(0) != '0') {
                            return results;
                        } else {
                            results.add(num_i);
                            return results;
                        }
                    }
                } else if (fac_i.length() == 1) {
                    results.add(word);
                    results.add(word + "0");
                    if (fac_i.charAt(0) != '0') {
                        return results;
                    } else {
                        results.add(num_i);
                        return results;
                    }
                } else {
                    return null;
                }
            }
        } else {
            results.add(word);
            results.add(word + ".0");
            results.add(word + ".00");
            return results;
        }
    }

    public static String toFunql(ArrayList<Production> pv) {
        StringBuffer sb = new StringBuffer();

        toFunqlHelper(pv.get(0), sb);

        return sb.toString();
    }

    public static void assignSeq(ArrayList<Production> prods) {
        for (int i = 0; i < prods.size(); i++) {
            Production prod = prods.get(i);
            prod.setProdSeq(i);
        }
    }

    private static void toFunqlHelper(Production prod, StringBuffer sb) {
        ArrayList<Node> rhs_nodes = prod.getRHS();

        boolean startQuote = false;
        boolean firstToken = false;

        ArrayList<Production> children = prod.getChildren();
        int cIndex = 0;
        for (int i = 0; i < rhs_nodes.size(); i++) {
            Node rhs_node = rhs_nodes.get(i);
            if (rhs_node instanceof NonterminalNode) {
                Production child = children.get(cIndex++);
                toFunqlHelper(child, sb);
            } else {
                if (rhs_node.getName().equals("'")) {
                    sb.append(rhs_node.getName());
                    startQuote = !startQuote;
                    if (startQuote == true) {
                        firstToken = true;
                    }
                } else {
                    if (startQuote && !firstToken) {
                        sb.append(' ');
                    }
                    if (firstToken) {
                        firstToken = false;
                    }
                    sb.append(rhs_node.getName());
                }
            }
        }
    }

}
